/*
 * driver:
 *    This is the driver for the TpcB benchmark. It runs on a separate
 * node from the SUT, and drives the SUT via non-transactional TRPC calls.
 * It collects all sorts of timing statistics, which are the whole point
 * of writing a benchmark, silly.
 *
 */

#include <objdce/statics.H>
#include <String.h>
#include <List.h>
#include <string.h>
#include <iostream.h>
#include <ctype.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <time.h>
#include <stdlib.h>
#include <unistd.h>
#include <objdce/mutex.H>
#include <objdce/condition.H>
#include <objdce/binding.H>
#include <objdce/error.H>
#include <objdce/login.H>
#include <hyperint.H>
#include <tpcfiles_cdl.H>
extern "C"
{
#include <extsut.h>
}
#include <driver.H>		/* Keep last because curses redefines stuff */

#ifdef __hpux
extern "C" double drand48();
extern "C" void srand48(long);
#endif

Boolean Driver::allDone;

Driver::Driver(const int sNo)

{
  FunctionName("Driver::Driver");
  int inc;
  char* sName;
  
  context = new CurrentLoginContext();
  if (context->LogIn(LoginContext::PACKET_AUTHENTICATION) == 
            LoginContext::FAILURE)
    {
      cerr << "Couldn't establish login context" << endl;
      exit (-2);
    }
  serverNo = sNo;
  if (sNo < 0)			// Then use all servers
    {
      useAllSuts = true;
      ScanParams(0);		// Don\'t care about "Params" object
      allSuts = new Binding*[GetNumServers()];
      for (inc = 0; inc < GetNumServers(); inc++)
	{
	  sName = GetTPCServer(inc);
	  allSuts[inc] = new Binding(sName, RPCClient(ExternalSUT,1,0), 
				     context);
	  if (!allSuts[inc]->IsValid())
	    {
	      cerr << "Couldn't establish binding to server" << endl;
	      exit(-1);
	    }
	}
    }
  else				// Use only one
    {
      useAllSuts = false;
      ScanParams(sNo);
      allSuts = new Binding*[1];
      sName = GetTPCServer(sNo);
      allSuts[0] = new Binding(sName, RPCClient(ExternalSUT,1,0), 
			       context);
      if (!allSuts[0]->IsValid())
	{
	  cerr << "Couldn't establish binding to server" << endl;
	  exit(-1);
	}
    }
  countLatch = new MutexLock();
  count = 0;
  logging = false;
  painter = nil;
  paintThread = nil;
  numThreads = 0;
  
  unlink("/usr/tmp/tpcb.log");
  logFile = -1;
  command = '\0';
}

Driver::~Driver()

{
  CleanUp();
}

void Driver::SetDone(Boolean d)

{
  allDone = d;
}

void Driver::CleanUp()

{
  int inc;

  cout << endl;
  cout << "Terminating threads:" << endl;
  for (inc = 0; inc < numThreads; inc++)
    {
      threadList[inc]->Kill();
      sleep(1);
      delete threadList[inc];
      delete threadObjs[inc];
      cout << "." << flush;
    }
  if (paintThread != nil)
    {
      paintThread->Kill();
      sleep(1);
      delete paintThread;
      delete painter;
      cout << "." << flush;
    }
  cout << endl;
  if (logFile > 0)
    close(logFile);
  delete countLatch;
  for (inc = 0; inc < GetNumServers(); inc++)
    delete allSuts[inc];
  delete [] allSuts;
  delete context;
}

void Driver::UserScreen(const int rThreads, const int wThreads,
		        Boolean limited)

{
  FunctionName("Driver::UserScreen");
  int inc;
  char readChar;

  if (limited)
    ScanRandoms(1);
  else
    ScanRandoms(rThreads + wThreads);
  if (!limited)
    {
      screen = initscr();
      cbreak();
      noecho();
      nonl();
      leaveok(screen, true);
      painter = new PaintThread(this);
      paintThread = new AsyncThread(*painter);
      for (inc = 0; inc < rThreads; inc++)
	SpawnThread(true);
      for (inc = 0; inc < wThreads; inc++)
	SpawnThread(false);
      while (!allDone)
	{
	  readChar = getchar();
	  command = readChar;
	}
      for (inc = 0; inc < numThreads; inc++)
	{
	  threadList[inc]->Wait();
	  delete threadList[inc];
	  delete threadObjs[inc];
	}
      paintThread->Wait();
      delete paintThread;
      delete painter;
      endwin();
    }
  else
    {
      if (rThreads > 0)
	SpawnThread(true, rThreads);
      else
	SpawnThread(false, wThreads);
      threadList[0]->Wait();
    }
}

void Driver::Update(LogItem* item)

{
    countLatch->Lock();
    if (logging)
	write(logFile, (char*)item, sizeof(LogItem));
    if (item->status == LogCommitted)
	count++;
    countLatch->Unlock();
}

void Driver::SpawnThread(Boolean readOnly, int limit)

{
  if (numThreads < MAX_THREADS)
    {
      threadObjs[numThreads] = new RunThread(this, allSuts, useAllSuts, 
					     numThreads, readOnly, limit);
      threadList[numThreads] = new AsyncThread(*(threadObjs[numThreads]));
      numThreads++;
    }
}

void Driver::KillThread()

{
    if (numThreads > 0)
    {
	numThreads--;
	threadList[numThreads]->Kill();
	delete threadList[numThreads];
	delete threadObjs[numThreads];
    }
}

void Driver::ToggleLog()

{
    LogItem lItem;
    timeb curTime;

    if (logging)
    {
	logging = false;
	ftime(&curTime);
	lItem.time = curTime.time;
	lItem.millitm = curTime.millitm;
	lItem.status = LogStop;
	write(logFile, (char*)&lItem, sizeof(LogItem));
    	close(logFile);
	logFile = -1;
    }
    else
    {
	logFile = open("/usr/tmp/tpcb.log", O_WRONLY | O_CREAT | O_APPEND,
		       S_IRUSR | S_IWUSR | S_IRGRP | S_IWGRP | S_IROTH);
	ftime(&curTime);
	lItem.time = curTime.time;
	lItem.millitm = curTime.millitm;
	lItem.status = LogStart;
	write(logFile, (char*)&lItem, sizeof(LogItem));
	logging = true;
    }
}

char Driver::Command()

{
    char cmdChar = command;
    command = '\0';
    return cmdChar;
}

RunThread::RunThread(Driver* d, Binding** bs, Boolean useAll,
		     int n, Boolean ro, int l)

{
  driver = d;
  suts = bs;
  useAllSuts = useAll;
  readOnly = ro;
  limit = l;
  threadNo = n;
}

RunThread::~RunThread()

{
}

void* RunThread::Run()

{
  FunctionName("RunThread::Run");
  timeb startTime;
  timeb stopTime;
  unsigned long accountID;
  unsigned long tellerID;
  unsigned long branchID;
  idl_hyper_int delta;
  idl_hyper_int newBalance;
  short didOK;
  error_status_t err;
  LogItem lItem;
  Account account;
  AccountBase accountBase;
  int runCount = 0;
  double randVal;
  double discrim;
  HyperInt bigDelta;
  Binding* thisSvr;
  
  while (!Driver::AllDone())
    {
      DoKill();
      if (readOnly)
	{
	  randVal = GetRandom(threadNo);
	  accountID = (unsigned long)((double)(TotalParams->numAccounts) * randVal);
	  account.accountID(accountID);
	  if (useAllSuts)
	    thisSvr = suts[FindAccount(accountID)];
	  else
	    thisSvr = suts[0];
	  ftime(&startTime);
	  err = Read(*thisSvr, account, &didOK);
	  ftime(&stopTime);
	  account = accountBase;
	}
      else
	{
	  randVal = GetRandom(threadNo);
	  tellerID = (unsigned long)((double)TotalParams->numTellers * randVal);
	  branchID = tellerID / 10;
	  discrim = GetRandom(threadNo);
	  randVal = GetRandom(threadNo);
	  if (discrim < 0.85)
	    accountID = (unsigned long)((double)TotalParams->numAccounts * randVal);
	  else
	    accountID = (unsigned long)(((double)Params->numAccounts * randVal) +
					Params->startAccount);
	  randVal = GetRandom(threadNo);
	  bigDelta = (signed long)((20000000.0 * randVal) - 10000000.0);
	  delta = bigDelta;
	  if (useAllSuts)
	    thisSvr = suts[FindAccount(accountID)];
	  else
	    thisSvr = suts[0];
	  ftime(&startTime);
	  err = Process(*thisSvr, accountID, tellerID, branchID, delta, 
			&newBalance, &didOK);
	  ftime(&stopTime);
	}
      if ((err == rpc_s_ok) && (didOK))
	lItem.status = LogCommitted;
      else
	lItem.status = LogFailed;
      lItem.time = stopTime.time - startTime.time;
      if (stopTime.millitm > startTime.millitm)
	{
	  lItem.millitm = stopTime.millitm - startTime.millitm;
	}
      else
	{
	  lItem.millitm = 1000 - (startTime.millitm - stopTime.millitm);
	  lItem.time--;
	}
      driver->Update(&lItem);
      if ((limit > 0) && (++runCount > limit))
	Driver::SetDone(true);
      else if (limit > 0)
	{
	  if (lItem.status == LogCommitted)
	    cerr << ".";
	  else
	    cerr << "F";
	}
    }
  return nil;
}

PaintThread::PaintThread(Driver* d)

{
    driver = d;
    paintLatch = new MutexLock();
}

PaintThread::~PaintThread()

{
    delete paintLatch;
}

void* PaintThread::Run()

{
    FunctionName("PaintThread::Run");
    WINDOW* headerWin;
    WINDOW* countWin;
    WINDOW* threadWin;
    float tpsB;
    time_t startupTime;
    time_t currentTime;
    time_t elapsedTime;
    char command;
    char* headerStr = "TPC Benchmark B";
    int headerLen = strlen(headerStr);
    char* footerStr = 
	"r, w - spawn read, write thread; k - kill thread; l - toggle log; q - quit";
    int footerLen = strlen(footerStr);
    int partWidth = (COLS - 15) / 2;
    unsigned long count;
    timespec_t sleepTime;
    sleepTime.tv_sec = 1;
    sleepTime.tv_nsec = 0;

    headerWin = newwin(3, 18, 2, (COLS - headerLen + 2) / 2);
    threadWin = newwin(2, partWidth, 5, 5);
    countWin = newwin(3, partWidth, 5, partWidth + 10);
    mvwaddstr(headerWin, 1, 1, headerStr);
    box(headerWin, 0, 0);
    mvaddstr(LINES - 1, (COLS - footerLen) / 2, footerStr);
    startupTime = time(nil);
    refresh();
    wrefresh(headerWin);
    while (!Driver::AllDone())
    {
	DoKill();
	wclear(threadWin);
	mvwprintw(threadWin, 0, 0, "%i threads active", driver->NumThreads());
	if (driver->Logging())
	    mvwaddstr(threadWin, 1, 0, "Logging enabled");
	wclear(countWin);
	currentTime = time(nil);
	elapsedTime = currentTime - startupTime;
	count = driver->Count();
	tpsB = (float)count / (float)elapsedTime;
    	mvwprintw(countWin, 0, 0, "%lu transactions completed", count);
	mvwprintw(countWin, 1, 0, "%li:%2li elapsed", (long)elapsedTime / 60,
		  (long)elapsedTime % 60);
	mvwprintw(countWin, 2, 0, "Approximately %.1f tpsB", tpsB);
	wrefresh(countWin);
	wrefresh(threadWin);
	SleepMe(1, 0);
	command = driver->Command();
	tolower(command);
	switch(command)
	{
	case 'r':
	    driver->SpawnThread(true);
	    break;
	case 'w':
	    driver->SpawnThread(false);
	    break;
	case 'k':
	    driver->KillThread();
    	    break;
	case 'l':
	    driver->ToggleLog();
	    break;
	case 'q':
	    Driver::SetDone(true);
	    break;
	default:
	    break;
	}
    }
    return nil;
}

static void PrintUsage()

{
  cout << "Usage: tpcdriver <server #> <# read threads> <# write threads>";
  cout << endl;
  cout << "       tpcdriver <server #> -r <# read ops>" << endl;
  cout << "       tpcdriver <server #> -w <# write ops>" << endl;
  cout << "           if \"server\" is \"-a\", contact all servers" << endl;
}

int main (int argc, char** argv)

{
  Driver* driver;
  int readThreads;
  int writeThreads;
  int serverNo;
  
  if ((argc == 4) && !strcmp(argv[2], "-r"))
    {
      if (!strcmp(argv[1], "-a"))
	serverNo = -1;
      else
	serverNo = atoi(argv[1]);
      driver = new Driver(serverNo);
      readThreads = atoi(argv[3]);
      driver->UserScreen(readThreads, 0, true);
      delete driver;
      return 0;
    }
  else if ((argc == 4) && !strcmp(argv[2], "-w"))
    {
      if (!strcmp(argv[1], "-a"))
	serverNo = -1;
      else
	serverNo = atoi(argv[1]);
      driver = new Driver(serverNo);
      writeThreads = atoi(argv[3]);
      driver->UserScreen(0, writeThreads, true);
      delete driver;
      return 0;
    }
  else if (argc == 4)
    {
      if (!strcmp(argv[1], "-a"))
	serverNo = -1;
      else
	serverNo = atoi(argv[1]);
      driver = new Driver(serverNo);
      readThreads = atoi(argv[2]);
      writeThreads = atoi(argv[3]);
      driver->UserScreen(readThreads, writeThreads, false);
      delete driver;
      return 0;
    }
  else
    {
      PrintUsage();
      return -1;
    }
}

