/*
 * tpcinit:
 *    This program initializes the TPC databases, first creating the files,
 * then filling them with data.
 */

#include <objdce/statics.H>
#include <params.H>
#include <stdlib.h>
#include <string.h>
#include <iostream.h>
#include <ctype.h>
#include <objtran/transaction.H>
#include <objdce/abortreason.H>
#include <objtran/tranclient.H>
#include <objtran/tranbinding.H>
#include <hyperint.H>
#include <tpcfiles_cdl.H>
#include <tpcfiles_cdl_sfs.H>

static const int BATCH_SIZE = 250;

class Creator : public Signaled, public Signaler

{
public:
    Creator();
    virtual ~Creator();
    void Create(const char* server, const char* volume, char fileType);
    void Fill(const char* server, char fileType, 
	      unsigned long startAt, unsigned long numRecs);
    void Zero(const char* server, char fileType);
    virtual Signaled::SigStatus HandleSignal(const int);
private:
    TranClient* client;
    SFSAccountFile* accounts;
    SFSTellerFile* tellers;
    SFSBranchFile* branches;
    SFSHistoryFile* history;
};
    
Creator::Creator()

{
  client = new TranClient();
  CatchSignal(SIGINT, *this);
  CatchSignal(SIGTERM, *this);
  accounts = nil;
  tellers = nil;
  branches = nil;
  history = nil;
}

Creator::~Creator()

{
    if (accounts != nil)
	delete accounts;
    if (tellers != nil)
	delete tellers;
    if (branches != nil)
	delete branches;
    if (history != nil)
	delete history;
    delete client;
}

Signaled::SigStatus Creator::HandleSignal(const int number)

{
  cerr << "Closing SFS files..." << endl;
  if ((number == SIGINT) || (number == SIGTERM))
    {
      if (accounts != nil)
	accounts->CloseFile();
      if (tellers != nil)
	tellers->CloseFile();
      if (branches != nil)
	branches->CloseFile();
      if (history != nil)
	history->CloseFile();
      if (accounts != nil)
	delete accounts;
      if (tellers != nil)
	delete tellers;
      if (branches != nil)
	delete branches;
      if (history != nil)
	delete history;
      return Signaled::TERMINATE;
    }
  else
    return Signaled::NOT_HANDLED;
}

void Creator::Create(const char* server, const char* volume, char fileType)

{
    Transaction* createTran;

    switch(fileType)
    {
    case 'a':
	cout << "Creating tpcAccounts..." << endl;
	createTran = new NewTransaction();
	accounts = new SFSAccountFile();
	accounts->CreateFile(server, "tpcAccounts", volume, Params->numAccounts);
	createTran->Commit();
	if (createTran->Aborted())
	  DefaultReport.Report(HighMsg, createTran->GetAbortReason());
	delete createTran;
	delete accounts;
	accounts = nil;
	break;
    case 't':
	cout << "Creating tpcTellers..." << endl;
	createTran = new NewTransaction();
	tellers = new SFSTellerFile();
	tellers->CreateFile(server, "tpcTellers", volume, Params->numTellers);
	tellers->OpenFile(server, "tpcTellers", 
			  CDL_EXCLUSIVE_OPEN | CDL_NONTRANSACTIONAL_OPEN);
	tellers->DeactivateIndex(TellerFile::BRANCH);
	tellers->CloseFile();
	createTran->Commit();
	if (createTran->Aborted())
	  DefaultReport.Report(HighMsg, createTran->GetAbortReason());
	delete createTran;
	delete tellers;
	tellers = nil;
	break;
    case 'b':
	cout << "Creating tpcBranches..." << endl;
	createTran = new NewTransaction();
	branches = new SFSBranchFile();
	branches->CreateFile(server, "tpcBranches", volume, Params->numBranches);
	createTran->Commit();
	if (createTran->Aborted())
	  DefaultReport.Report(HighMsg, createTran->GetAbortReason());
	delete createTran;
	delete branches;
	branches = nil;
	break;
    case 'h':
	cout << "Creating tpcHistory..." << endl;
	createTran = new NewTransaction();
	history = new SFSHistoryFile();
	history->CreateFile(server, "tpcHistory", volume, 2000);
	createTran->Commit();
	if (createTran->Aborted())
	  DefaultReport.Report(HighMsg, createTran->GetAbortReason());
	delete createTran;
	delete history;
	history = nil;
	break;
    default:
	cerr << "Invalid file type." << endl;
	break;
    }
    cout << "Done." << endl;
}

void Creator::Fill(const char* server, char fileType, 
	           unsigned long startAt, unsigned long numRecs)

{
  unsigned long curRec;
  unsigned long curBranch;
  Account* accountVec;
  Teller teller;
  Branch branch;
  History hist;
  int inc;
  unsigned long branchStart;
  unsigned long branchEnd;
  Transaction* writeTran;
  String msg;
  char tellerName[255];
  unsigned long lastRec;
  unsigned long thisRec;
  HyperInt initialVal;
  
  tolower(fileType);
  switch(fileType)
    {
    case 'a':
      cout << "Writing " << numRecs << " account records..." << endl;
      accounts = new SFSAccountFile();
      writeTran = new NewTransaction();
      accounts->OpenFile(server, "tpcAccounts");
      writeTran->Commit();
      delete writeTran;
      
      initialVal = (long)InitialAccount;
      initialVal *= (HyperInt)100L;			// Counting pennies here!
      accountVec = new Account[BATCH_SIZE];
      curRec = startAt;
      lastRec = startAt + numRecs;
      for (curBranch = Params->startBranch; 
	   curBranch <= Params->endBranch; 
	   curBranch++)
	{
	  branchStart = BranchSize * curBranch;
	  branchEnd = branchStart + BranchSize;
	  for (; (curRec >= branchStart) && (curRec < branchEnd) &&
	       (curRec < lastRec); )
	    {
	      writeTran = new NewTransaction(240);
	      accounts->StartBatch();
	      thisRec = curRec;
	      for (inc = 0; inc < BATCH_SIZE; inc++)
		{
		  if (!writeTran->Aborted())
		    {
		      accountVec[inc].accountID(thisRec++);
		      accountVec[inc].branchID(curBranch);
		      accountVec[inc].balance(initialVal);
		      accounts->Insert(accountVec[inc]);
		    }
		  else
		    break;
		}
	      if (!writeTran->Aborted())
		accounts->EndBatch();
	      writeTran->Commit();
	      if (writeTran->Committed())
		curRec += BATCH_SIZE;
	      if (writeTran->Aborted())
		{
		  DefaultReport.Report(HighMsg, writeTran->GetAbortReason());
		  break;
		}
	      else if ((curRec % 1000) == 0)
		cout << curRec << "..." << endl;
	      else if ((curRec % BATCH_SIZE) == 0)
		cout << "." << flush;
	      delete writeTran;
	    }
	}
      delete [] accountVec;
      accounts->CloseFile();
      delete accounts;
      accounts = nil;
      break;
    case 't':
      cout << "Writing " << Params->numTellers << " teller records..." << endl;
      tellers = new SFSTellerFile();
      writeTran = new NewTransaction();
      tellers->OpenFile(server, "tpcTellers");
      writeTran->Commit();
      delete writeTran;

      initialVal = (long)InitialTeller;
      initialVal *= (HyperInt)100L;
      teller.balance(initialVal);
      for (curBranch = Params->startBranch; 
	   curBranch <= Params->endBranch; 
	   curBranch++)
	{
	  branchStart = 10 * curBranch;
	  branchEnd = branchStart + 10;
	  teller.branchID(curBranch);
	  for (curRec = branchStart; curRec < branchEnd; curRec++)
	    {
	      writeTran = new NewTransaction();
	      sprintf(tellerName, "Citibank teller number %li", curRec);
	      teller.fill(tellerName);
	      teller.tellerID(curRec);
	      tellers->Insert(teller);
	      writeTran->Commit();
	      if (writeTran->Aborted())
		{
		  DefaultReport.Report(HighMsg, writeTran->GetAbortReason());
		  break;
		}
	      delete writeTran;
	    }
	}
      tellers->CloseFile();
      delete tellers;
      tellers = nil;
      break;
    case 'b':
      cout << "Writing " << Params->numBranches << " branch records..." << endl;
      branches = new SFSBranchFile();
      writeTran = new NewTransaction();
      branches->OpenFile(server, "tpcBranches");
      writeTran->Commit();
      delete writeTran;

      initialVal = (long)InitialBranch;
      initialVal *= (HyperInt)100L;
      branch.balance(initialVal);
      branch.fill("Citibank, 111 Wall St., New York, NY");
      for (curBranch = Params->startBranch; 
	   curBranch <= Params->endBranch; 
	   curBranch++)
	{
	  writeTran = new NewTransaction();
	  branch.branchID(curBranch);
	  branches->Insert(branch);
	  writeTran->Commit();
	  if (writeTran->Aborted())
	    {
	      DefaultReport.Report(HighMsg, writeTran->GetAbortReason());
	      break;
	    }
	  delete writeTran;
	}
      branches->CloseFile();
      delete branches;
      branches = nil;
      break;
    case 'h':
      cout << "Writing a history record..." << endl;
      hist.fill("Test history record");
      history = new SFSHistoryFile();
      writeTran = new NewTransaction();
      history->OpenFile(server, "tpcHistory");
      writeTran->Commit();
      delete writeTran;
      writeTran = new NewTransaction();
      history->Insert(hist);
      writeTran->Commit();
      if (writeTran->Aborted())
	{
	  DefaultReport.Report(HighMsg, writeTran->GetAbortReason());
	  break;
	}
      delete writeTran;
      history->CloseFile();
      delete history;
      history = nil;
      break;
    default:
      cerr << "Wrong file type." << endl;
      break;
    }
  cout << "Done." << endl;
}

void Creator::Zero(const char* server, char fileType)

{
  unsigned long curRec;
  unsigned long curCurRec;
  Transaction* zeroTran;
  Account* accountVec;
  Account* accountKeyVec;
  Branch branch;
  Branch branchKey;
  Teller teller;
  Teller tellerKey;
  String msg;
  HyperInt initialVal;
  int inc;

  tolower(fileType);
  switch(fileType)
    {
    case 'a':
      cout << "Zeroing account records..." << endl;
      accounts = new SFSAccountFile();
      zeroTran = new NewTransaction();
      accounts->OpenFile(server, "tpcAccounts");
      zeroTran->Commit();
      delete zeroTran;

      initialVal = (long)InitialAccount;
      initialVal *= (HyperInt)100L;
      accountVec = new Account[BATCH_SIZE];
      accountKeyVec = new Account[BATCH_SIZE];
      curRec = Params->startAccount;
      while (curRec <= Params->endAccount)
	{
	  zeroTran = new NewTransaction();
	  accounts->StartBatch();
	  curCurRec = curRec;
	  for (inc = 0; inc < BATCH_SIZE; inc++)
	    {
	      if (!zeroTran->Aborted())
		{
		  accountKeyVec[inc].accountID(curCurRec++);
		  accounts->Modify_balance(AccountFile::PRIMARY, 
					   accountKeyVec[inc],
					   initialVal, accountVec[inc]);
		}
	      else
		break;
	    }
	  if (!zeroTran->Aborted())
	    accounts->EndBatch();
	  zeroTran->Commit();
	  if (zeroTran->Aborted())
	    {
	      delete zeroTran;
	      curCurRec = curRec;
	      for (inc = 0; inc < BATCH_SIZE; inc++)
		{
		  zeroTran = new NewTransaction();
		  accountKeyVec[0].accountID(curCurRec++);
		  accounts->Modify_balance(AccountFile::PRIMARY,
					   accountKeyVec[0],
					   initialVal, accountVec[0]);
		  zeroTran->Commit();
		  if (zeroTran->Aborted())
		    cout << "F" << flush;
		  delete zeroTran;
		}
	      curRec += BATCH_SIZE;
	    }
	  else
	    {
	      curRec += BATCH_SIZE;
	      delete zeroTran;
	    }
	  if ((curRec % 1000) == 0)
	    cout << curRec << "." << endl;
	  else
	    cout << "." << flush;
	}
      cout << "\ndone." << endl;
      accounts->CloseFile();
      delete [] accountVec;
      delete [] accountKeyVec;
      accounts = nil;
      break;
    case 't':
      cout << "Zeroing teller records..." << endl;
      tellers = new SFSTellerFile();
      zeroTran = new NewTransaction();
      tellers->OpenFile(server, "tpcTellers");
      zeroTran->Commit();
      delete zeroTran;

      initialVal = (long)InitialTeller;
      initialVal *= (HyperInt)100L;
      for (curRec = Params->startTeller;
	   curRec <= Params->endTeller;
	   curRec++)
	{
	  zeroTran = new NewTransaction();
	  tellerKey.tellerID(curRec);
	  tellers->Modify_balance(TellerFile::PRIMARY, tellerKey,
				  initialVal, teller);
	  zeroTran->Commit();
	  if (zeroTran->Aborted())
	    cout << "F" << flush;
	  else
	    cout << "." << flush;
	  delete zeroTran;
	}
      cout << "\ndone." << endl;
      tellers->CloseFile();
      delete tellers;
      tellers = nil;
      break;
    case 'b':
      cout << "Zeroing branch records..." << endl;
      branches = new SFSBranchFile();
      zeroTran = new NewTransaction();
      branches->OpenFile(server, "tpcBranches");
      zeroTran->Commit();
      delete zeroTran;

      initialVal = (long)InitialBranch;
      initialVal *= (HyperInt)100L;
      for (curRec = Params->startBranch;
	   curRec <= Params->endBranch;
	   curRec++)
	{
	  zeroTran = new NewTransaction();
	  branchKey.branchID(curRec);
	  branches->Modify_balance(BranchFile::PRIMARY, branchKey,
				   initialVal, branch);
	  zeroTran->Commit();
	  if (zeroTran->Aborted())
	    cout << "F" << flush;
	  else
	    cout << "." << flush;
	  delete zeroTran;
	}
      cout << "\ndone." << endl;
      branches->CloseFile();
      delete branches;
      branches = nil;
      break;
    default:
      cerr << "Wrong file type." << endl;
      break;
    }
}

static void PrintUsage()

{
    cout << "Usage: tpcinit c <server_no> <volumeName> : Create data file";
    cout << endl;
    cout << "       tpcinit w <server_no> <type> <startAt> : Write data into files";
    cout << endl;
    cout << "       tpcinit z <server_no> <type> : Re-initialize data" << endl;
    cout << "       where type = 'a', 't', 'b', or 'h'" << endl;
    cout << "       if type == 'a', use this format:" << endl;
    cout << "	    tpcinit w <server> a <startAt> <numRecs>" << endl;
}

int main (int argc, char** argv)

{
    Creator* obj;
    unsigned long startAt;
    unsigned long numRecs;
    int serverNo;
    char* serverName;

    if ((argc < 4) || (argc > 6))
      {
	PrintUsage();
	return -1;
      }
    serverNo = atoi(argv[2]);
    ScanParams(serverNo);
    serverName = GetServer(serverNo);
    if ((argc == 4) && !strcmp(argv[1], "c"))
    {
	obj = new Creator();
	obj->Create(serverName, argv[3], 'a');
	obj->Create(serverName, argv[3], 't');
	obj->Create(serverName, argv[3], 'b');
	obj->Create(serverName, argv[3], 'h');
	delete obj;
	return 0;
    }
    else if ((argc == 5) && !strcmp(argv[1], "c"))
    {
	obj = new Creator();
	obj->Create(serverName, argv[3], argv[4][0]);
	delete obj;
	return 0;
    }
    else if ((argc == 5) && !strcmp(argv[1], "w"))
    {
	obj = new Creator();
	startAt = (unsigned long)atol(argv[4]);
	obj->Fill(serverName, argv[3][0], startAt, 0);
	delete obj;
	return 0;
    }
    else if ((argc == 6) && !strcmp(argv[1], "w"))
    {
	obj = new Creator();
	startAt = (unsigned long)atol(argv[4]);
	numRecs = (unsigned long)atol(argv[5]);
	obj->Fill(serverName, argv[3][0], startAt, numRecs);
	delete obj;
	return 0;
    }
    else if ((argc == 4) && !strcmp(argv[1], "z"))
      {
	obj = new Creator();
	obj->Zero(serverName, argv[3][0]);
	delete obj;
	return 0;
      }
    else    
    {
	PrintUsage();
	return -1;
    }
}
