/*
 * smalldriver
 *
 *    This is a less-complicated version of the TPC driver. It can be used to 
 * test how many independent DCE processes can be active at once. It is
 * single-threaded, and simulates a real user by randomly sleeping between
 * transactions.
 *
 * smalldriver.C,v 1.1 1994/04/28 21:37:58 gjb Exp
 */

#include <objdce/statics.H>
#include <String.h>
#include <iostream.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <time.h>
#include <fstream.h>
#include <params.H>
#include <objdce/binding.H>
#include <objdce/error.H>
#include <objdce/login.H>
#include <objdce/transignals.H>
#include <hyperint.H>
#include <List.h>
#include <tpcfiles_cdl.H>
extern "C"
{
#include <dce/rpc.h>
#include <dce/utc.h>
#include <extsut.h>
}

#ifdef __hpux
extern "C" double drand48();
extern "C" void srand48(long);
#endif

#define NANO 1000000000
#define MICRO 1000000
#define MILLI 1000
#define HUNDREDTH 100

static ofstream outFile;

/*
 * TranStamp
 *    A timestamp for a transaction.
 */

class TranStamp

{
public:
  TranStamp();
  TranStamp(const TranStamp&);
  TranStamp& operator = (const TranStamp&);
  int operator == (const TranStamp&) const;
  void Start();
  void Stop(Boolean OK = true);
  void Elapsed(utc_t* out);
  Boolean Successful();
private:
  utc_t startTime;
  utc_t stopTime;
  Boolean success;
};

TranStamp::TranStamp()

{
  success = false;
}

TranStamp::TranStamp(const TranStamp& stamp)

{
  startTime = stamp.startTime;
  stopTime = stamp.stopTime;
  success = stamp.success;
}

TranStamp& TranStamp::operator = (const TranStamp& stamp)

{
  startTime = stamp.startTime;
  stopTime = stamp.stopTime;
  success = stamp.success;
  return *this;
}

int TranStamp::operator == (const TranStamp& stamp) const

{
  utc_cmptype startTimes;
  utc_cmptype stopTimes;

  utc_cmpmidtime(&startTimes, (utc_t*)&stamp.startTime, (utc_t*)&startTime);
  utc_cmpmidtime(&stopTimes, (utc_t*)&stamp.stopTime, (utc_t*)&stopTime);
  return ((startTimes == utc_equalTo) && (stopTimes == utc_equalTo));
}

void TranStamp::Start()

{
  utc_gettime(&startTime);
}

void TranStamp::Stop(Boolean OK)

{
  utc_gettime(&stopTime);
  success = OK;
}

void TranStamp::Elapsed(utc_t* out)

{
  utc_subtime(out, &stopTime, &startTime);
}

Boolean TranStamp::Successful()

{
  return success;
}

/*
 * SmallDriver
 *    This is the actual driver.
 */

class SmallDriver : public Signaled, public Signaler

{
public:
  SmallDriver(const String& principal, const String& password,
	      const String& serverName, long avgTran, long avgSleep,
	      double writePercent);
  virtual ~SmallDriver();
  void Run();
  void PrintStats(ostream& out);
  virtual Signaled::SigStatus HandleSignal(const int);
private:
  Binding* sutBinding;
  LoginContext* context;
  TranStamp loginTime;
  TranStamp bindTime;
  List<TranStamp> rpcReadTimes;
  List<TranStamp> rpcWriteTimes;
  long numRPCs;
  long avgTran;
  long avgSleep;
  double writePercent;
};

SmallDriver::SmallDriver(const String& principal, const String& password,
			 const String& sName, long tran, long sleep,
		         double wp)

{
  ScanParams(0);
  CatchSignal(SIGINT, *this);
  CatchSignal(SIGTERM, *this);
  loginTime.Start();
  PasswordLoginContext* passContext;
  passContext = new PasswordLoginContext();
  if (passContext->LogIn(principal, password, LoginContext::CONNECT) ==
      LoginContext::FAILURE)
    {
      cerr << "Couldn't log in as principal " << principal << endl;
      exit (-2);
    }
  context = passContext;
  loginTime.Stop(true);
  bindTime.Start();
  sutBinding = new Binding(sName, RPCClient(ExternalSUT,1,0), context);
  if (!sutBinding->IsValid())
    {
      cerr << "Couldn't establish binding to server" << endl;
      exit(-1);
    }
  bindTime.Stop(true);
  srand48((long)time(NULL));
  numRPCs = 0;
  avgTran = tran;
  avgSleep = sleep;
  writePercent = wp;
}

SmallDriver::~SmallDriver()

{
  delete sutBinding;
  delete context;
}

Signaled::SigStatus SmallDriver::HandleSignal(const int sig)

{
  if ((sig == SIGINT) || (sig == SIGTERM))
    {
      delete context;
      delete sutBinding;
      PrintStats(outFile);
      outFile.close();
      return Signaled::TERMINATE;
    }
  else
    return Signaled::NOT_HANDLED;
}

static long ComputeCurve(long center)

{
  double randVal;
  double range;
  double bottom;

  bottom = (double)center / 2.0;
  range = (double)center;
  randVal = range * drand48();
  if (randVal < 0.0)
    randVal = 0.0;
  return (long)(randVal + bottom);
}

static void PrintTime(utc_t* utc, String& out)

{
  reltimespec_t relTime;
  timespec_t inacc;
  unsigned long seconds;
  long hundredths;
  char tmpNum[64];
  
  utc_binreltime(&relTime, &inacc, utc);
  seconds = relTime.tv_sec;
  hundredths = relTime.tv_nsec / (NANO / HUNDREDTH);
  sprintf(tmpNum, "%lu", seconds);
  out = tmpNum;
  out += ".";
  if (hundredths < 10)
    out += "0";
  sprintf(tmpNum, "%i", hundredths);
  out += tmpNum;
}

void SmallDriver::Run()

{
  double randVal;
  long numTran;
  long inc;
  TranStamp rpcTime;
  unsigned long accountID;
  unsigned long tellerID;
  unsigned long branchID;
  HyperInt bigDelta;
  idl_hyper_int delta;
  idl_hyper_int newBalance;
  error_status_t err;
  short didOK;
  Account account;
  long sleepTime;

  numTran = ComputeCurve(avgTran);
  for (inc = 0; inc < numTran; inc++)
    {
      didOK = 1;
      randVal = drand48();
      if (randVal <= writePercent)
	{
	  randVal = drand48();
	  tellerID = (unsigned long)((double)TotalParams->numTellers * randVal);
	  branchID = tellerID / 10;
	  randVal = drand48();
	  accountID = (unsigned long)((double)TotalParams->numAccounts * randVal);
	  randVal = drand48();
	  bigDelta = (signed long)((20000000.0 * randVal) - 10000000.0);
	  delta = bigDelta;
	  rpcTime.Start();
	  err = Process(*sutBinding, accountID, tellerID, branchID, 
			delta, &newBalance, &didOK);
	  if ((err == rpc_s_ok) && didOK)
	    rpcTime.Stop(true);
	  else
	    rpcTime.Stop(false);
	  rpcWriteTimes.put(rpcTime);
	}
      else
	{
	  randVal = drand48();
	  accountID = (unsigned long)((double)TotalParams->numAccounts * randVal);
	  account.accountID(accountID);
	  rpcTime.Start();
	  err = Read(*sutBinding, account, &didOK);
	  if ((err == rpc_s_ok) && didOK)
	    rpcTime.Stop(true);
	  else
	    rpcTime.Stop(false);
	  rpcReadTimes.put(rpcTime);
	}	
      sleepTime = ComputeCurve(avgSleep);
      sleep((unsigned)sleepTime);
      numRPCs++;
    }
}  

void SmallDriver::PrintStats(ostream& out)

{
  utc_t oneTime;
  utc_t totalTime;
  utc_t firstTime;
  utc_t lastTime;
  long count;
  long failureCount;
  double ratio;
  String timeStr;
  TranStamp oneStamp;
  timespec zeroInacc;
  reltimespec zeroTime;
  
  zeroTime.tv_sec = 0;
  zeroTime.tv_nsec = 0;
  zeroInacc.tv_sec = 0;
  zeroInacc.tv_nsec = 0;

  loginTime.Elapsed(&oneTime);
  PrintTime(&oneTime, timeStr);
  out << "Initial authentication: " << timeStr << endl;
  bindTime.Elapsed(&oneTime);
  PrintTime(&oneTime, timeStr);
  out << "Initial binding: " << timeStr << endl;

  Listiter<TranStamp> readIter(rpcReadTimes);
  count = 0;
  failureCount = 0;
  utc_mkbinreltime(&totalTime, &zeroTime, &zeroInacc);
  while (readIter.next(oneStamp))
    {
      if (count == 0)
	oneStamp.Elapsed(&firstTime);
      count++;
      oneStamp.Elapsed(&oneTime);
      utc_addtime(&totalTime, &totalTime, &oneTime);
      if (!oneStamp.Successful())
	failureCount++;
    }
  oneStamp.Elapsed(&lastTime);
  out << endl;
  out << count << " read RPCs" << endl;
  out << failureCount << " failed." << endl;
  PrintTime(&totalTime, timeStr);
  out << "Total: " << timeStr << endl;
  PrintTime(&firstTime, timeStr);
  out << "First call: " << timeStr << endl;
  PrintTime(&lastTime, timeStr);
  out << "Last call: " << timeStr << endl;
  ratio = 1.0 / (double)count;
  utc_mulftime(&oneTime, &totalTime, ratio);
  PrintTime(&oneTime, timeStr);
  out << "Average: " << timeStr << endl;

  Listiter<TranStamp> writeIter(rpcWriteTimes);
  count = 0;
  failureCount = 0;
  utc_mkbinreltime(&totalTime, &zeroTime, &zeroInacc);
  while (readIter.next(oneStamp))
    {
      if (count == 0)
	oneStamp.Elapsed(&firstTime);
      count++;
      oneStamp.Elapsed(&oneTime);
      utc_addtime(&totalTime, &totalTime, &oneTime);
      if (oneStamp.Successful())
	failureCount++;
    }
  oneStamp.Elapsed(&lastTime);
  out << endl;
  out << count << " write RPCs" << endl;
  out << failureCount << " failed." << endl;
  PrintTime(&totalTime, timeStr);
  out << "Total: " << timeStr << endl;
  PrintTime(&firstTime, timeStr);
  out << "First call: " << timeStr << endl;
  PrintTime(&lastTime, timeStr);
  out << "Last call: " << timeStr << endl;
  ratio = 1.0 / (double)count;
  utc_mulftime(&oneTime, &totalTime, ratio);
  PrintTime(&oneTime, timeStr);
  out << "Average: " << timeStr << endl;
}

static void PrintUsage()

{
  cerr << "Usage: smalldriver <principal> <password> <server name>" << endl;
  cerr << "         <num transactions> <sleep time> <write percentage>" << endl;
  cerr << "         <output file>" << endl;
}

int main (int argc, char** argv)

{
  String principal;
  String password;
  String server;
  String outFileName;
  long numTran;
  long sleepTime;
  double percent;
  long numPercent;
  Boolean wrong = false;
  SmallDriver* driver;

  if (argc != 8)
    {
      PrintUsage();
      return -1;
    }
  principal = argv[1];
  if (principal == "")
    wrong = true;
  password = argv[2];
  if (password == "")
    wrong = true;
  server = argv[3];
  if (server == "")
    wrong = true;
  numTran = atol(argv[4]);
  if (numTran <= 0)
    wrong = true;
  sleepTime = atol(argv[5]);
  numPercent = atol(argv[6]);
  if ((numPercent < 0) || (numPercent > 100))
    wrong = true;
  percent = (double)numPercent / 100.0;
  outFileName = argv[7];
  if (outFileName == "")
    wrong = true;
  if (wrong)
    {
      PrintUsage();
      return -1;
    }
  
  outFile.open(outFileName);
  if (outFile.fail())
    {
      cerr << "Error opening output file." << endl;
      return -1;
    }
  driver = new SmallDriver(principal, password, server, 
			   numTran, sleepTime, percent);
  driver->Run();
  driver->PrintStats(outFile);
  outFile.close();
  delete driver;
  return 0;
}

     
