/*
 * Sut:
 *    This is the system under test: the program that performs the actual
 * transaction processing. Here is where we try and see which ways
 * of processing transactions are efficient and which aren't.
 *
 */

#include <objdce/statics.H>
#include <stdio.h>
#include <iostream.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/timeb.h>
#include <sut.H>
#include <objtran/transaction.H>
#include <objdce/abortreason.H>
#include <objdce/error.H>
#include <objdce/stats.H>
#include <objtran/tranbinding.H>
#include <objdce/login.H>
#include <hyperint.H>
#include <tpcfiles_cdl.H>
#include <tpcfiles_cdl_sfs.H>
extern "C"
{
#include <sut.h>
#include <extsut.h>
}
#include <params.H>

static const int ReadTimeout = 10;
static const int ProcessTimeout = 10;

/* Global */

Sut* Sut::myself;

error_status_t Process(handle_t handle, unsigned long accountID,
		       unsigned long tellerID, unsigned long branchID,
		       idl_hyper_int delta, idl_hyper_int* newBalance,
		       idl_short_int* committed)
{
  
  Sut::Instance()->Process(handle, accountID, tellerID, branchID, 
			   delta, newBalance, committed);
  return 0;
}

void UpdateAccount(trpc_handle_t handle, unsigned long accountID, 
		   idl_hyper_int delta, idl_hyper_int* newBal)
{
  Sut::Instance()->UpdateAccount(handle, accountID, delta, newBal);
}

void UpdateTeller(trpc_handle_t handle, unsigned long tellerID, 
		  idl_hyper_int delta, idl_hyper_int* newBal)
{
  Sut::Instance()->UpdateTeller(handle, tellerID, delta, newBal);
}

void UpdateBranch(trpc_handle_t handle, unsigned long branchID, 
		  idl_hyper_int delta, idl_hyper_int* newBal)
{
  Sut::Instance()->UpdateBranch(handle, branchID, delta, newBal);
}

error_status_t Read(handle_t handle, AccountBase* rec,
	  idl_short_int* committed)
{
    Sut::Instance()->Read(handle, rec, committed);
    return 0;
}

error_status_t ReadBranch(handle_t handle, BranchBase* rec,
			  idl_short_int* committed)
{
    Sut::Instance()->ReadBranch(handle, rec, committed);
    return 0;
}

error_status_t ReadTeller(handle_t handle, TellerBase* rec,
			  idl_short_int* committed)
{
    Sut::Instance()->ReadTeller(handle, rec, committed);
    return 0;
}

Sut::Sut(int serverNo)

{
  FunctionName("Sut::Sut");
  Transaction* openTran;
  char* otherDCEName;
  Boolean stillOK = true;
  DCEuuid objUUID;		// Create new object uuid
  char userInput[255];
  char* prinName;

  ScanParams(serverNo);
  char* myDCEName = GetTPCServer(serverNo);
  char* sfsName = GetServer(serverNo);
  myNo = serverNo;
  myself = this;
  prinName = getenv("ENCINA_PRINCIPAL");
  if (prinName == NULL)
  {
    cerr << "ENCINA_PRINCIPAL must be set" << endl;
    exit (-1);
  }

  cout << "Login as DCE principal " << prinName << "..." << endl;
  context = new ServerLoginContext();
  if (context->LogIn(prinName, KeyTable, ACLDir,
		     LoginContext::PACKET_AUTHENTICATION) == 
           LoginContext::FAILURE)
    {
      cerr << "Couldn't obtain login context" << endl;
      exit(-2);
    }
  cout << "Registering server with name " << myDCEName << "..." << flush;
  server = new TranServer(RPCServer(SUT,2,0), RPCVector(SUT,2,0),
			  myDCEName, "Transactional TPC-B server", UDP,
			  context, (List_of_p<TranComponent>*)0,
			  NULL, &objUUID);
  server->AddInterface(RPCServer(ExternalSUT,1,0), myDCEName, 
		       "TPC-B DCE Server", NULL, NULL, &objUUID);
  cout << "done." << endl;

  otherServers = new TranBinding*[GetNumServers()];
  if (GetNumServers() > 1)
    {
      cout << "Hit <Return> when the other TPC servers are up: " << flush;
      gets(userInput);
    }
  for (int inc = 0; (inc < GetNumServers()) && stillOK; inc++)
    {
      if (inc != myNo)
	{
	  cout << "Looking up server " << inc << "..." << endl;
	  otherDCEName = GetTPCServer(inc);
	  otherServers[inc] = new TranBinding(otherDCEName, RPCClient(SUT,2,0),
					      context);
	  if (!(otherServers[inc]->IsValid()))
	    stillOK = false;
	}
    }

  if (stillOK)
    {
      openTran = new NewTransaction();
      accounts = new SFSAccountFile();
      accounts->OpenFile(sfsName, "tpcAccounts");
      if (!openTran->Aborted())
	{
	  tellers = new SFSTellerFile();
	  tellers->OpenFile(sfsName, "tpcTellers");
	}
      if (!openTran->Aborted())
	{
	  branches = new SFSBranchFile();
	  branches->OpenFile(sfsName, "tpcBranches");
	}
      if (!openTran->Aborted())
	{
	  history = new SFSHistoryFile();
	  history->OpenFile(sfsName, "tpcHistory");
	}
      openTran->Commit();
      if (openTran->Aborted())
	stillOK = false;
      delete openTran;
    }
  if (!stillOK)
    cerr << "Couldn't start TPC server. Exiting." << endl;
  else
    {
      cout << "Listening..." << endl;
      server->StartServer(MAX_THREADS);
      cout << "Shutting down..." << endl;
    }
}

Sut::~Sut()

{
  cout << "Cleaning up..." << endl;
  InterfaceStats statistics;
  server->GetStatistics(statistics);
  cout << "RPC calls in:  " << statistics.CallsIn() << endl;
  cout << "RPC calls out: " << statistics.CallsOut() << endl;
  cout << "Packets in:    " << statistics.PacketsIn() << endl;
  cout << "Packets out:   " << statistics.PacketsOut() << endl;
  accounts->CloseFile();
  delete accounts;
  tellers->CloseFile();
  delete tellers;
  branches->CloseFile();
  delete branches;
  history->CloseFile();
  delete history;
  delete server;
  delete context;
}

Sut* Sut::Instance()

{
  return myself;
}

void Sut::Process(handle_t handle, unsigned long accountID, 
		  unsigned long tellerID,
		  unsigned long branchID, idl_hyper_int delta, 
	          idl_hyper_int* newBalance, short* committed)
{
  FunctionName("Sut::Process");
  char errMsg[255];
  Account account;
  Account accountKey;
  Teller teller;
  Teller tellerKey;
  Branch branch;
  Branch branchKey;
  History hist;
  NewTransaction mainTran(ProcessTimeout);
  int whichServer;
  idl_hyper_int retVal;
  idl_hyper_int newBal;
  timeb curTime;

  if (!server->GenericRPCServer::IsClntAuthorized(handle))
    {
      mainTran.Abort("Client is not authorized", -1, 0);
    }
  if (!mainTran.Aborted())
    {
      // Update account file
      accounts->StartBatch();
      whichServer = FindAccount(accountID);
      if (whichServer == myNo)
	{						// Do it here
	  accountKey.accountID(accountID);
	  accounts->Increment_balance(AccountFile::PRIMARY, accountKey,
				      delta, account, true);
	}
      else
	{						// Do it there
	  (*SUT_v2_0_client_epv.UpdateAccount)(*(otherServers[whichServer]),
					       accountID, delta, &newBal);
	  retVal = newBal;
	}
    }
  if (!mainTran.Aborted())
    {
      // Update teller file
      whichServer = FindTeller(tellerID);
      if (whichServer == myNo)
	{
	  tellerKey.tellerID(tellerID);
	  tellers->Increment_balance(TellerFile::PRIMARY, tellerKey, 
				     delta, teller, true);
	}
      else
	{
	  (*SUT_v2_0_client_epv.UpdateTeller)(*(otherServers[whichServer]),
					      tellerID, delta, &newBal);
	}
    }
  if (!mainTran.Aborted())
    {
      // Update branch file
      whichServer = FindBranch(branchID);
      if (whichServer == myNo)
	{
	  branchKey.branchID(branchID);
	  branches->Increment_balance(BranchFile::PRIMARY, branchKey, 
				      delta, branch, true);
	}
      else
	{
	  (*SUT_v2_0_client_epv.UpdateBranch)(*(otherServers[whichServer]),
					      branchID, delta, &newBal);
	}
    }
  if (!mainTran.Aborted())
    {
      // Update history file
      hist.accountID(accountID);
      hist.tellerID(tellerID);
      hist.branchID(branchID);
      hist.delta(delta);
      hist.fill("This is a history record");
      ftime(&curTime);
      hist.seconds(curTime.time);
      hist.milliseconds(curTime.millitm);
      history->Insert(hist);
    }
  
  if (!mainTran.Aborted())
    accounts->EndBatch();

  mainTran.Commit();
  if (mainTran.Committed())
    {
      DefaultReport.VReport(LowMsg, "Succeeded: account %li\n", accountID);
      DefaultReport.VReport(LowMsg, "Succeeded: teller %li\n", tellerID);
      DefaultReport.VReport(LowMsg, "Succeeded: branch %li\n", branchID);
      retVal = account.balance();
      *committed = true;
    }
  else
    {
      DefaultReport.VReport(NormalMsg, "Failed: teller %li\n", tellerID);
      DefaultReport.VReport(NormalMsg, "Failed: account %li\n", accountID);
      DefaultReport.VReport(NormalMsg, "Failed: branch %li\n", branchID);
      retVal.high = retVal.low = 0;
      *committed = false;
    }
  *newBalance = retVal;
}

void Sut::UpdateAccount(trpc_handle_t handle, unsigned long accountID, 
			idl_hyper_int delta, idl_hyper_int* newBal)

{
  Account account;
  Account accountKey;
  ExistingTransaction tran;

  if (!server->IsClntAuthorized(handle))
    {
      tran.Abort("Client is not authorized", -1, 0);
    }
  else
    {
      accountKey.accountID(accountID);
      accounts->Increment_balance(AccountFile::PRIMARY, accountKey, 
				  delta, account, true);
      *newBal = account.balance();
      if (tran.Aborted())
	DefaultReport.VReport(NormalMsg, "Failed: account %li\n", accountID);
      else
	DefaultReport.VReport(LowMsg, "Succeeded: account %li\n", accountID);
    }
}

void Sut::UpdateTeller(trpc_handle_t handle, unsigned long tellerID, 
		       idl_hyper_int delta, idl_hyper_int* newBal)

{
  Teller teller;
  Teller tellerKey;
  ExistingTransaction tran;

  if (!server->IsClntAuthorized(handle))
    {
      tran.Abort("Client is not authorized", -1, 0);
    }
  else
    {
      tellerKey.tellerID(tellerID);
      tellers->Increment_balance(TellerFile::PRIMARY, tellerKey, 
				 delta, teller, true);
      *newBal = teller.balance();
      if (tran.Aborted())
	DefaultReport.VReport(NormalMsg, "Failed: teller %li\n", tellerID);
      else
	DefaultReport.VReport(LowMsg, "Succeeded: teller %li\n", tellerID);
    }
}

void Sut::UpdateBranch(trpc_handle_t handle, unsigned long branchID, 
		       idl_hyper_int delta, idl_hyper_int* newBal)

{
  Branch branch;
  Branch branchKey;
  ExistingTransaction tran;

  if (!server->IsClntAuthorized(handle))
    {
      tran.Abort("Client is not authorized", -1, 0);
    }
  else
    {
      branchKey.branchID(branchID);
      branches->Increment_balance(BranchFile::PRIMARY, branchKey, 
				  delta, branch, true);
      *newBal = branch.balance();
      if (tran.Aborted())
	DefaultReport.VReport(NormalMsg, "Failed: branch %li\n", branchID);
      else
	DefaultReport.VReport(LowMsg, "Succeeded: branch %li\n", branchID);
    }
}

void Sut::Read(handle_t handle, AccountBase* rec, short* committed)

{
    FunctionName("Sut::Read");
    AccountBase* tmpBase;
    Account account;
    Account accountKey;
    char errMsg[255];
    NewTransaction mainTran(ReadTimeout);

    if (!server->GenericRPCServer::IsClntAuthorized(handle))
      mainTran.Abort("Client not authorized", -1, 0);
    else
      {
	account = *rec;
	accountKey.accountID(account.accountID());
	accounts->ReadUnique(READ, AccountFile::PRIMARY, accountKey, account);
      }
    mainTran.Commit();

    if (mainTran.Committed())
    {
      DefaultReport.VReport(LowMsg, "ReadAccount %li\n", 
			    account.accountID());
      *committed = 1;
    }
    else
    {
      DefaultReport.VReport(LowMsg, 
			    "Failed to read account %li\n", 
			    account.accountID());
      *committed = 0;
    }
    tmpBase = account;
    *rec = *tmpBase;
}

void Sut::ReadBranch(handle_t handle, BranchBase* rec, short* committed)

{
    FunctionName("Sut::ReadBranch");
    Branch branch;
    Branch branchKey;
    char errMsg[255];
    BranchBase* tmpBase;
    NewTransaction mainTran(ReadTimeout);

    if (!server->GenericRPCServer::IsClntAuthorized(handle))
      mainTran.Abort("Client not authorized", -1, 0);
    else
      {
	branch = *rec;
	branchKey.branchID(branch.branchID());
	branches->ReadUnique(READ, BranchFile::PRIMARY, branchKey, branch);
      }
    mainTran.Commit();

    if (mainTran.Committed())
    {
      DefaultReport.VReport(LowMsg, "Read branch %li\n", branch.branchID());
      *committed = 1;
    }
    else
      {
	DefaultReport.VReport(LowMsg, "Failed to read branch %li\n", 
			       branch.branchID());
	*committed = 0;
      }
    tmpBase = branch;
    *rec = *tmpBase;
}

void Sut::ReadTeller(handle_t handle, TellerBase* rec, short* committed)

{
    FunctionName("Sut::ReadTeller");
    Teller teller;  
    Teller tellerKey;
    char errMsg[255];
    TellerBase* tmpBase;
    NewTransaction mainTran(ReadTimeout);

    if (!server->GenericRPCServer::IsClntAuthorized(handle))
      mainTran.Abort("Client not authorized", -1, 0);
    else
      {
	teller = *rec;
	tellerKey.tellerID(teller.tellerID());
	tellers->ReadUnique(READ, TellerFile::PRIMARY, tellerKey, teller);
      }
    mainTran.Commit();

    if (mainTran.Committed())
      {
	DefaultReport.VReport(LowMsg, "Read teller %li\b", teller.tellerID());
	*committed = 1;
      }
    else
      {
	DefaultReport.VReport(LowMsg, "Failed to read teller %li\n",
			      teller.tellerID());
	*committed = 0;
    }
    tmpBase = teller;
    *rec = *tmpBase;
}

static void PrintUsage()

{
    cout << "Usage: tpcsut <server number>" << endl;
}

int main(int argc, char** argv)

{
  Sut* sut;
  
  if (argc == 2)
    {
      int serverNo = atoi(argv[1]);
      sut = new Sut(serverNo);
      delete sut;
      return 0;
    }
  else
    {
      PrintUsage();
      return -1;
    }
}

