/*
 * RFSDataFile
 * 
 *    This class talks to the SFS server and deals with RFS records.
 *
 * rfsdatafile.C,v 1.2 1994/04/07 19:28:52 gjb Exp
 */

#include <rfsdatafile.H>
#include <objdce_config.h>
#include <objtran/transaction.H>
#include <objdce/abortreason.H>
#include <objdce/dcetime.H>
#include <objdce/error.H>
#include <Regex.h>
#include <stdlib.h>
#include <sys/wait.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/stat.h>

/*
 * Internal constants
 */

static const char* const DefaultEditor = "vi";
static const char* const RFSSaveFile = "./Dead_RFS";

static const char* const id_title = "Bug ID";
static const char* const application_title = "Application";
static const char* const summary_title = "Summary";
static const char* const requestor_title = "Requestor";
static const char* const priority_title = "Priority";
static const char* const person_title = "Person-Days";
static const char* const request_title = "Request Date";
static const char* const estimated_title = "Estimated Completion Date";
static const char* const actual_title = "Actual Completion Date";
static const char* const release_title = "Incorporated in Release";
static const char* const description_title = "Description";
static const char* const benefit_title = "Benefit";
static const int OperationTimeout = 30;

RFSDataFile::RFSDataFile(const String& dbName,
			 const String& fileName)

{
  CDLStatus status;
  String abtReason;

  NewTransaction openTran;
  status = rfsFile.OpenFile(dbName, fileName);
  openTran.Commit();

  if (openTran.Aborted())
    {
      openTran.GetAbortReason()->Print(abtReason);
      printf("Couldn't open RFS file\n%s\n", (const char*)abtReason);
    }

  char* envVal = getenv("EDITOR");
  if (envVal == (char*)0)
    {
      editorName = DefaultEditor;
    }
  else
    {
      editorName = envVal;
    }
/***
#ifdef sun
  tzsetwall();
#endif
***/
}

RFSDataFile::~RFSDataFile()

{
  rfsFile.CloseFile();
}

Boolean RFSDataFile::DisplayAllItems(FILE* outFile,
				     const DisplayType t,
				     Boolean openOnly)
{
  Boolean retVal = true;
  RFSItem curItem;
  RFSItem priorityKey;
  unsigned long displayCount = 0;

  NewTransaction allTran;
  if (openOnly)
    {
      priorityKey.priority(0);
      rfsFile.Select(RFSDatabase::PRIORITY, &priorityKey, (RFSItem*)0,
		     FORWARDS);
    }
  else
    {
      rfsFile.Select(RFSDatabase::PRIORITY, (RFSItem*)0, (RFSItem*)0,
		     FORWARDS);
    }
  while ((!allTran.Aborted()) && 
	 (rfsFile.ReadCurrent(BRACKET, NEXT, curItem) == CDL_SUCCESS))
    {
      DisplayItem(outFile, curItem, t, false);
      displayCount++;
    }
  allTran.Commit();
  if (allTran.Aborted())
    {
      retVal = false;
      String abtReason;
      allTran.GetAbortReason()->Print(abtReason);
      printf("Error in DisplayAllItems:\n%s\n", (const char*)abtReason);
    }
  else
    {
      printf("Dumped %lu records.\n", displayCount);
    }
  return retVal;
}

Boolean RFSDataFile::DisplayByID(FILE* outFile, const unsigned long id,
				 const DisplayType t)

{
  Boolean retVal = true;
  RFSItem curItem;
  RFSItem key;
  CDLStatus status;

  key.id(id);
  NewTransaction allTran(OperationTimeout);
  status = rfsFile.ReadUnique(READ, RFSDatabase::PRIMARY, key, curItem);
  allTran.Commit();
  if (status == CDL_RECORD_NOT_FOUND)
    {
      printf("No matching records found.\n");
      retVal = false;
    }
  else if (allTran.Aborted())
    {
      retVal = false;
      String abtReason;
      allTran.GetAbortReason()->Print(abtReason);
      printf("Error in DisplayByID:\n%s\n", (const char*)abtReason);
    }
  else
    {
      DisplayItem(outFile, curItem, t, false);
    }
  return retVal;
}

Boolean RFSDataFile::DisplayByApplication(FILE* outFile, 
					  const String& appName,
					  const DisplayType t)

{
  Boolean retVal = true;
  RFSItem curItem;
  Regex searchFor(appName, Regex::case_insensitive);
  Substrinfo foundIt;
  Subex tmpSubex;

  NewTransaction allTran;
  rfsFile.Select(RFSDatabase::PRIMARY, (RFSItem*)0, (RFSItem*)0, FORWARDS);
  while ((!allTran.Aborted()) && 
	 (rfsFile.ReadCurrent(BRACKET, NEXT, curItem) == CDL_SUCCESS))
    {
      foundIt = searchFor.match(curItem.application(), tmpSubex);
      if ((curItem.priority() >= 0) && foundIt)
	{
	  DisplayItem(outFile, curItem, t, false);
	}
    }
  allTran.Commit();
  if (allTran.Aborted())
    {
      retVal = false;
      String abtReason;
      allTran.GetAbortReason()->Print(abtReason);
      printf("Error in DisplayByApplication:\n%s\n", (const char*)abtReason);
    }
  return retVal;
}

Boolean RFSDataFile::DisplayByPriority(FILE* outFile, 
				       const long priority,
				       const DisplayType t)

{
  Boolean retVal = true;
  RFSItem curItem;
  RFSItem key;

  key.priority(priority);
  NewTransaction allTran;
  rfsFile.Select(RFSDatabase::PRIORITY, &key, &key, FORWARDS);
  while ((!allTran.Aborted()) && 
	 (rfsFile.ReadCurrent(BRACKET, NEXT, curItem) == CDL_SUCCESS))
    {
      DisplayItem(outFile, curItem, t, false);
    }
  allTran.Commit();
  if (allTran.Aborted())
    {
      retVal = false;
      String abtReason;
      allTran.GetAbortReason()->Print(abtReason);
      printf("Error in DisplayByPriority:\n%s\n", (const char*)abtReason);
    }
  return retVal;
}

Boolean RFSDataFile::DisplayByEstimatedDate(FILE* outFile, 
					    const long requestDate,
					    const DisplayType t)

{
  Boolean retVal = true;
  RFSItem curItem;
  RFSItem key;

  key.estimatedCompletionDate(requestDate);
  NewTransaction allTran;
  rfsFile.Select(RFSDatabase::ESTIMATEDDATE, (RFSItem*)0, &key, FORWARDS);
  while ((!allTran.Aborted()) && 
	 (rfsFile.ReadCurrent(BRACKET, NEXT, curItem) == CDL_SUCCESS))
    {
      if (curItem.priority() >= 0)
	{
	  DisplayItem(outFile, curItem, t, false);
	}
    }
  allTran.Commit();
  if (allTran.Aborted())
    {
      retVal = false;
      String abtReason;
      allTran.GetAbortReason()->Print(abtReason);
      printf("Error in DisplayByEstimatedDate:\n%s\n", (const char*)abtReason);
    }
  return retVal;
}

Boolean RFSDataFile::Insert()

{
  RFSItem newItem;
  RFSItem lastItem;
  unsigned long newID;
  Boolean retVal = true;
  time_t now;
  Boolean firstTime = false;

  if (EditBlankItem(newItem))
    {
      now = time((time_t*)0);
      newItem.requestDate((long)now);
      NewTransaction insertTran(OperationTimeout);
      rfsFile.Select(RFSDatabase::PRIMARY, (RFSItem*)0, (RFSItem*)0,
		     BACKWARDS);
      if (!insertTran.Aborted())
	{
	  NewTransaction lastTran(OperationTimeout);
	  if (rfsFile.ReadCurrent(READ, PREVIOUS, lastItem) == CDL_END_OF_RANGE)
	    {
	      firstTime = true;
	    }
	  lastTran.Commit();
	}
      if (!insertTran.Aborted())
	{
	  if (!firstTime)
	    {
	      newID = lastItem.id() + 1;
	    }
	  else
	    {
	      newID = 1;
	    }
	  newItem.id(newID);
	  rfsFile.Insert(newItem);
	}
      insertTran.Commit();
      if (insertTran.Aborted())
	{
	  retVal = false;
	  String abtReason;
	  insertTran.GetAbortReason()->Print(abtReason);
	  printf("Error inserting a new record:\n%s\n",
		 (const char*)abtReason);
	  SaveItem(RFSSaveFile, newItem);
	}
      return retVal;
    }
  else
    return false;
}

Boolean RFSDataFile::Delete(const unsigned long id)

{
  RFSItem deadKey;
  Boolean retVal = true;

  deadKey.id(id);
  NewTransaction deleteTran(OperationTimeout);
  rfsFile.DeleteUnique(RFSDatabase::PRIMARY, deadKey);
  deleteTran.Commit();
  if (deleteTran.Aborted())
    {
      retVal = false;
      String abtReason;
      deleteTran.GetAbortReason()->Print(abtReason);
      printf("Error deleting record:\n%s\n",
	     (const char*)abtReason);
    }
  return retVal;
}

Boolean RFSDataFile::Close(const unsigned long id)

{
  RFSItem closedKey;
  RFSItem closedItem;
  Boolean retVal = true;
  String abtReason;
  time_t now;

  closedKey.id(id);
  NewTransaction readTran(OperationTimeout);
  rfsFile.ReadUnique(READ, RFSDatabase::PRIMARY, 
		     closedKey, closedItem);
  if (!readTran.Aborted())
    {
      if (closedItem.priority() == ClosedRFSPriority)
	{
	  readTran.Abort("The RFS is already closed.", 2);
	}
      else
	{
	  now = time((time_t*)0);
	  rfsFile.Modify_actualCompletionDate(RFSDatabase::PRIMARY,
					      closedKey, now, closedItem);
	  if (!readTran.Aborted())
	    {
	      rfsFile.Modify_priority(RFSDatabase::PRIMARY,
				      closedKey, ClosedRFSPriority,
				      closedItem);
	    }
	}
    }
  readTran.Commit();
  if (readTran.Aborted())
    {
      retVal = false;
      readTran.GetAbortReason()->Print(abtReason);
      printf("Error updating record:\n%s\n", (const char*)abtReason);
    }
  return retVal;
}

Boolean RFSDataFile::ReOpen(const unsigned long id,
			    const long newPriority)
{
  RFSItem closedKey;
  RFSItem closedItem;
  Boolean retVal = true;
  String abtReason;

  closedKey.id(id);
  NewTransaction readTran(OperationTimeout);
  rfsFile.ReadUnique(READ, RFSDatabase::PRIMARY, 
		     closedKey, closedItem);
  if (!readTran.Aborted())
    {
      if (closedItem.priority() != ClosedRFSPriority)
	{
	  readTran.Abort("The RFS is already open.", 2);
	}
      else
	{
	  rfsFile.Modify_actualCompletionDate(RFSDatabase::PRIMARY,
					      closedKey, -1, closedItem);
	  if (!readTran.Aborted())
	    {
	      rfsFile.Modify_priority(RFSDatabase::PRIMARY,
				      closedKey, newPriority,
				      closedItem);
	    }
	}
    }
  readTran.Commit();
  if (readTran.Aborted())
    {
      retVal = false;
      readTran.GetAbortReason()->Print(abtReason);
      printf("Error updating record:\n%s\n", (const char*)abtReason);
    }
  return retVal;
}

Boolean RFSDataFile::Edit(const unsigned long id)

{
  RFSItem editItem;
  RFSItem itemKey;
  Boolean retVal = true;
  String abtReason;

  NewTransaction longTran;
  itemKey.id(id);
  NewTransaction readTran(OperationTimeout);
  rfsFile.ReadUnique(READ, RFSDatabase::PRIMARY, itemKey, editItem);
  readTran.Commit();
  if (readTran.Aborted())
    {
      retVal = false;
      readTran.GetAbortReason()->Print(abtReason);
      printf("Couldn't read record for editing\n%s\n",
	     (const char*)abtReason);
      longTran.Abort("Couldn't read record for editing", 1);
    }
  if (!longTran.Aborted())
    {
      if (!EditExistingItem(editItem))
	{
	  retVal = false;
	  longTran.Abort("The record was not changed.", 2);
	}
    }
  if (!longTran.Aborted())
    {
      NewTransaction updateTran(OperationTimeout);
      rfsFile.UpdateUnique(RFSDatabase::PRIMARY, itemKey, editItem);
      updateTran.Commit();
      if (updateTran.Aborted())
	{
	  retVal = false;
	  updateTran.GetAbortReason()->Print(abtReason);
	  printf("Couldn't replace record\n%s\n",
		 (const char*)abtReason);
	  longTran.Abort("Couldn't replace record", 1);
	  SaveItem(RFSSaveFile, editItem);
	}
    }
  longTran.Commit();
  return retVal;
}

Boolean RFSDataFile::DisplayItem(FILE* outFile, const RFSItem& item,
				 const DisplayType dType,
				 Boolean fromUser)
{
  if (dType == SUMMARIZE)
    {
      fprintf(outFile, "%06lu %li %s %s\n", item.id(),
	      item.priority(), (const char*)item.application(),
	      (const char*)item.summary());
      return true;
    }
  else
    {
      String tmpDate;
      if (!fromUser)
	{
	  fprintf(outFile, "%s: %06lu\n", id_title, item.id());
	}
      fprintf(outFile, "%s: %s\n", application_title,
	      (const char*)item.application());
      fprintf(outFile, "%s: %s\n", summary_title,
	      (const char*)item.summary());
      fprintf(outFile, "%s: %s\n", requestor_title,
	      (const char*)item.requestor());
      fprintf(outFile, "%s: %li\n", priority_title,
	      item.priority());
      fprintf(outFile, "%s: %li\n", person_title,
	      item.personDays());
      if (!fromUser)
	{
	  tmpDate = DateToString(item.requestDate());
	  fprintf(outFile, "%s: %s\n", request_title,
		  (const char*)tmpDate);
	}
      tmpDate = DateToString(item.estimatedCompletionDate());
      fprintf(outFile, "%s: %s\n", estimated_title,
	      (const char*)tmpDate);
      if (!fromUser)
	{
	  tmpDate = DateToString(item.actualCompletionDate());
	  fprintf(outFile, "%s: %s\n", actual_title,
		  (const char*)tmpDate);
	}
      fprintf(outFile, "%s: %s\n", release_title,
	      (const char*)item.releaseID());
      fprintf(outFile, "%s:\n%s", description_title,
	      (const char*)item.description());
      fprintf(outFile, "%s:\n%s", benefit_title,
	      (const char*)item.benefit());

      fprintf(outFile, "\n\n");

      return true;
    }
}

Boolean RFSDataFile::Restore(FILE* inFile)

{
  RFSItem newItem;
  Boolean retVal = true;
  String abtReason;
  unsigned long restoreCount = 0;

  printf("Restoring...\n");
  lineBuf[0] = '\0';
  firstOne = true;
  while (!feof(inFile) && retVal)
    {
      if (ReadItem(inFile, newItem, false))
	{
	  NewTransaction insertTran(OperationTimeout);
	  rfsFile.Insert(newItem);
	  insertTran.Commit();
	  if (insertTran.Aborted())
	    {
	      retVal = false;
	      insertTran.GetAbortReason()->Print(abtReason);
	      printf("Error inserting records into new file:\n%s\n",
		     (const char*)abtReason);
	    }
	  else
	    {
	      printf(".");
	      fflush(stdout);
	      restoreCount++;
	    }
	  firstOne = false;
	}
      else
	{
	  retVal = false;
	  printf("Error scanning records for restore.\n");
	}
    }
  if (!retVal)
    {
      printf("Restore failed.\n");
    }
  else
    {
      printf("Restored %lu records.\n", restoreCount);
    }
  return retVal;
}

Boolean RFSDataFile::SaveItem(const String& fileName, 
			      const RFSItem& item)
{
  FILE* deadFile;
  Boolean retVal = true;

  deadFile = fopen(fileName, "w");
  if (deadFile == (FILE*)0)
    {
      printf("Couldn't open save file \"%s\"\n", (const char*)fileName);
      retVal = false;
    }
  else
    {
      if (DisplayItem(deadFile, item, DISPLAY, true))
	{
	  printf("A copy of the new item was saved in \"%s\"\n", 
		 (const char*)fileName);
	}
      else
	{
	  printf("A copy of the item could not be saved.\n");
	  retVal = false;
	}
      fclose(deadFile);
    }
  return retVal;
}

Boolean RFSDataFile::ReadLine(FILE* inFile)

{
  Boolean ok;

  if (ok = (fgets(lineBuf, LineBufSize, inFile) != (char*)0))
    {
      ObjtranReport.VReport(ErrorLevels::TraceMsg, "ReadLine: %s", lineBuf);
    }
  return ok;
}

Boolean RFSDataFile::ReadLineAndSkipBlanks(FILE* inFile)

{
  Boolean ok = true;

  if (ok = (fgets(lineBuf, LineBufSize, inFile) != (char*)0))
    {
      ObjtranReport.VReport(ErrorLevels::TraceMsg, "ReadLineAndSkip: %s", lineBuf);
    }
  while (ok && (strcspn(lineBuf, "\n\r") == 0))
    {
      if (ok = (fgets(lineBuf, LineBufSize, inFile) != (char*)0))
	{
	  ObjtranReport.VReport(ErrorLevels::TraceMsg, "ReadLineAndSkip: %s", lineBuf);
	}
    }
  return ok;
}  

Boolean RFSDataFile::ScanLine(const String& title, String& out)

{
  char* linePtr;
  int textLen;
  Boolean ok;

  if (strncmp(lineBuf, title, title.length()) ||
      ((linePtr = strchr(lineBuf, ':')) == (char*)0))
    {
      printf("No value found for field \"%s\"\n", (const char*)title);
      ok = false;
    }
  else
    {
      while ((*linePtr == ' ') || (*linePtr == ':'))
	linePtr++;
      textLen = strcspn(linePtr, "\n\r");
      out.assign(linePtr, textLen);
      ObjtranReport.VReport(ErrorLevels::TraceMsg, "Field %s: found \"%s\"\n", (const char*)title,
	     (const char*)out);
      ok = true;
    }
  return ok;
}

Boolean RFSDataFile::ScanMultiLine(FILE* inFile, const String& title, 
				   const String& nextTitle, String& out)

{
  char* linePtr;
  int textLen;
  Boolean ok = true;
  Boolean notAtEnd = true;

  if (strncmp(lineBuf, title, title.length()) ||
      ((linePtr = strchr(lineBuf, ':')) == (char*)0))
    {
      printf("No value found for field \"%s\"\n", (const char*)title);
      ok = false;
    }
  else
    {
      while ((*linePtr == ' ') || (*linePtr == ':'))
	linePtr++;
      textLen = strcspn(linePtr, "\n\r");
      if (textLen > 0)
	{
	  out = linePtr;
	  ObjtranReport.VReport(ErrorLevels::TraceMsg, "Field %s: First line : \"%s\"\n", 
		 (const char*)title, linePtr);
	}
      else
	{
	  out = "";
	  ObjtranReport.VReport(ErrorLevels::TraceMsg, "Field %s: First line: blank\n", (const char*)title);
	}
    }

  notAtEnd = ReadLine(inFile);
  while (notAtEnd && (strncmp(lineBuf, nextTitle, nextTitle.length())))
    {
      out += lineBuf;
      ObjtranReport.VReport(ErrorLevels::TraceMsg, "Field %s: Line: \"%s\"\n", 
	     (const char*)title, lineBuf);
      notAtEnd = ReadLine(inFile);
    }
  return ok;
}

Boolean RFSDataFile::ScanLine(const String& title, long& out)

{
  char* linePtr;
  Boolean ok = true;

  if (strncmp(lineBuf, title, title.length()) ||
      ((linePtr = strchr(lineBuf, ':')) == (char*)0))
    {
      printf("No value found for field \"%s\"\n", (const char*)title);
      ok = false;
    }
  else
    {
      while ((*linePtr == ' ') || (*linePtr == ':'))
	linePtr++;
      if (sscanf(linePtr, "%li", &out) != 1)
	{
	  printf("No integer value found for field \"%s\"\n",
		 (const char*)title);
	  ok = false;
	}
      else
	{
	  ObjtranReport.VReport(ErrorLevels::TraceMsg, "Field %s: \"%li\"\n", (const char*)title, out);
	}
    }
  return ok;
}

Boolean RFSDataFile::ScanLine(const String& title, unsigned long& out)

{
  char* linePtr;
  Boolean ok = true;

  if (strncmp(lineBuf, title, title.length()) ||
      ((linePtr = strchr(lineBuf, ':')) == (char*)0))
    {
      printf("No value found for field \"%s\"\n", (const char*)title);
      ok = false;
    }
  else
    {
      while ((*linePtr == ' ') || (*linePtr == ':'))
	linePtr++;
      if (sscanf(linePtr, "%lu", &out) != 1)
	{
	  printf("No unsigned integer value found for field \"%s\"\n",
		 (const char*)title);
	  ok = false;
	}
      else
	{
	  ObjtranReport.VReport(ErrorLevels::TraceMsg, "Field %s: \"%lu\"\n", (const char*)title, out);
	}
    }
  return ok;
}

Boolean RFSDataFile::ScanDateLine(const String& title, long& out)

{
  char* linePtr;
  int textLen;
  String dateStr;
  Boolean ok = true;

  if (strncmp(lineBuf, title, title.length()) ||
      ((linePtr = strchr(lineBuf, ':')) == (char*)0))
    {
      printf("No value found for field \"%s\"\n", (const char*)title);
      ok = false;
    }
  else
    {
      while ((*linePtr == ' ') || (*linePtr == ':'))
	linePtr++;
      textLen = strcspn(linePtr, "\n\r");
      dateStr.assign(linePtr, textLen);
      out = StringToDate(dateStr);
      if (out == 0)
	{
	  printf("Invalid date format for field \"%s\"\n",
		 (const char*)title);
	  ok = false;
	}
      else
	{
	  ObjtranReport.VReport(ErrorLevels::TraceMsg, "Field %s: \"%s\"\n", (const char*)title, 
		 (const char*)dateStr);
	}
    }
  return ok;
}

Boolean RFSDataFile::ReadItem(FILE* inFile, 
			      RFSItem& newItem, 
			      Boolean fromUser)
{
  char lineBuf[256];
  Boolean ok = true;
  String lineStr;
  long tmpLong;
  unsigned long tmpULong;

  if (firstOne)		// Set when restoring first record only
    {
      ok = ReadLineAndSkipBlanks(inFile);
    }
  if (ok && !fromUser)
    {
      if (ok = ScanLine(id_title, tmpULong))
	{
	  newItem.id(tmpULong);
	  ok = ReadLineAndSkipBlanks(inFile);
	}
    }
  if (ok)
    {
      if (ok = ScanLine(application_title, lineStr))
	{
	  newItem.application(lineStr);
	  ok = ReadLineAndSkipBlanks(inFile);
	}
    }
  if (ok)
    {
      if (ok = ScanLine(summary_title, lineStr))
	{
	  newItem.summary(lineStr);
	  ok = ReadLineAndSkipBlanks(inFile);
	}
    }
  if (ok)
    {
      if (ok = ScanLine(requestor_title, lineStr))
	{
	  newItem.requestor(lineStr);
	  ok = ReadLineAndSkipBlanks(inFile);
	}
    }
  if (ok)
    {
      if (ok = ScanLine(priority_title, tmpLong))
	{
	  newItem.priority(tmpLong);
	  ok = ReadLineAndSkipBlanks(inFile);
	}
    }
  if (ok)
    {
      if (ok = ScanLine(person_title, tmpLong))
	{
	  newItem.personDays(tmpLong);
	  ok = ReadLineAndSkipBlanks(inFile);
	}
    }
  if (ok && !fromUser)
    {
      if (ok = ScanDateLine(request_title, tmpLong))
	{
	  newItem.requestDate(tmpLong);
	  ok = ReadLineAndSkipBlanks(inFile);
	}
    }
  if (ok)
    {
      if (ok = ScanDateLine(estimated_title, tmpLong))
	{
	  newItem.estimatedCompletionDate(tmpLong);
	  ok = ReadLineAndSkipBlanks(inFile);
	}
    }
  if (ok && !fromUser)
    {
      if (ok = ScanDateLine(actual_title, tmpLong))
	{
	  newItem.actualCompletionDate(tmpLong);
	  ok = ReadLineAndSkipBlanks(inFile);
	}
    }
  if (ok)
    {
      if (ok = ScanLine(release_title, lineStr))
	{
	  newItem.releaseID(lineStr);
	  ok = ReadLineAndSkipBlanks(inFile);
	}
    }
  if (ok)
    {
      if (ok = ScanMultiLine(inFile, description_title, 
			     benefit_title, lineStr))
	{
	  newItem.description((const unsigned char*)((const char*)lineStr), 
			      lineStr.length() + 1);
	  // Don't read next line after calling ScanMultiLine
	}
    }
  if (ok)
    {
      if (ok = ScanMultiLine(inFile, benefit_title, 
			     id_title, lineStr))
	{
	  newItem.benefit((const unsigned char*)((const char*)lineStr), 
			  lineStr.length() + 1);
	}
    }

  return ok;
}

Boolean RFSDataFile::WriteTemplate(FILE* outFile)

{
  fprintf(outFile, "%s:\n", application_title);
  fprintf(outFile, "%s:\n", summary_title);
  fprintf(outFile, "%s:\n", requestor_title);
  fprintf(outFile, "%s:\n", priority_title);
  fprintf(outFile, "%s:\n", person_title);
  fprintf(outFile, "%s:\n", estimated_title);
  fprintf(outFile, "%s:\n", release_title);
  fprintf(outFile, "%s:\n", description_title);
  fprintf(outFile, "%s:\n", benefit_title);
  return true;
}

Boolean RFSDataFile::EditBlankItem(RFSItem& newItem)

{
  String tmpFileName;
  FILE* tmpFile;
  Boolean retVal = true;
  Boolean formatOK = false;
  char continueChar;
  char sillyBuf[64];
  struct stat statBuf;
  time_t origMTime;

  tmpFileName = tempnam("/usr/tmp", "RFSData");
  tmpFile = fopen(tmpFileName, "w");
  if (tmpFile == (FILE*)0)
    {
      printf("Couldn't open a temporary file for editing\n");
      retVal = false;
    }
  else
    {
      WriteTemplate(tmpFile);
      fclose(tmpFile);
      stat(tmpFileName, &statBuf);
      origMTime = statBuf.st_mtime;
      while (retVal && !formatOK)
	{
	  if (!LaunchEditor(tmpFileName))
	    {
	      printf("Couldn't launch the editor\n");
	      retVal = false;
	    }
	  else
	    {
	      stat(tmpFileName, &statBuf);
	      if (statBuf.st_mtime == origMTime)
		{
		  retVal = false; 	// File wasn't changed
		}
	      else
		{
		  tmpFile = fopen(tmpFileName, "r");
		  lineBuf[0] = '\0';
		  firstOne = true;
		  formatOK = ReadItem(tmpFile, newItem);
		  fclose(tmpFile);
		  if (!formatOK)
		    {
		      printf("\nErrors in input file format.\n");
		      printf("Hit <Enter> to return to the editor, or \"q\" to quit.\n");
		      gets(sillyBuf);
		      continueChar = sillyBuf[0];
		      if (continueChar == 'q')
			{
			  retVal = false;
			}
		    }
		}
	    }
	}
      unlink(tmpFileName);
    }
  return retVal;
}

Boolean RFSDataFile::EditExistingItem(RFSItem& newItem)

{
  String tmpFileName;
  FILE* tmpFile;
  Boolean retVal = true;
  Boolean formatOK = false;
  char continueChar;
  char sillyBuf[64];
  struct stat statBuf;
  time_t origMTime;

  tmpFileName = tempnam("/usr/tmp", "RFSData");
  tmpFile = fopen(tmpFileName, "w");
  if (tmpFile == (FILE*)0)
    {
      printf("Couldn't open a temporary file for editing\n");
      retVal = false;
    }
  else
    {
      DisplayItem(tmpFile, newItem, DISPLAY, true);
      fclose(tmpFile);
      stat(tmpFileName, &statBuf);
      origMTime = statBuf.st_mtime;
      while (retVal && !formatOK)
	{
	  if (!LaunchEditor(tmpFileName))
	    {
	      printf("Couldn't launch the editor\n");
	      retVal = false;
	    }
	  else
	    {
	      stat(tmpFileName, &statBuf);
	      if (statBuf.st_mtime == origMTime)
		{
		  retVal = false;	// File didn't change
		}
	      else
		{
		  tmpFile = fopen(tmpFileName, "r");
		  lineBuf[0] = '\0';
		  firstOne = true;
		  formatOK = ReadItem(tmpFile, newItem);
		  fclose(tmpFile);
		  if (!formatOK)
		    {
		      printf("\nErrors in input file format.\n");
		      printf("Hit <Enter> to return to the editor, or \"q\" to quit.\n");
		      gets(sillyBuf);
		      continueChar = sillyBuf[0];
		      if (continueChar == 'q')
			{
			  retVal = false;
			}
		    }
		}
	    }
	}
      unlink(tmpFileName);
    }
  return retVal;
}

Boolean RFSDataFile::LaunchEditor(const String& fileName)

{
  Boolean retVal = true;
#if HAVE_UNION_WAIT
  union wait waitStatus;
#else
  int waitStatus;
#endif

  if (vfork() == 0)
    {
      if (execlp(editorName, editorName, 
		(const char*)fileName, (char*)0) < 0)
	{
	  perror("Error launching editor");
	  exit(1);
	}
    }
  else
    {
      wait(&waitStatus);
      if ((WIFEXITED(waitStatus)) && 
	  (WEXITSTATUS(waitStatus) != 0))
	{
	  retVal = false;
	}
    }
  return retVal;
}

long RFSDataFile::StringToDate(const String& date)

{
  time_t retVal = (time_t)0;
  tm timeTm;
  char timeBuf[64];
  int bufPos;
  int bufLen;
  
  if (date.length() == 0)
    {
      return 0;
    }
  else
    {
      strncpy(timeBuf, date, 64);
      timeBuf[63] = '\0';
      bufPos = 0;
      bufLen = strlen(timeBuf);
      while ((bufPos < bufLen) && (timeBuf[bufPos] == ' '))
	{
	  bufPos++;
	}
      if (bufPos < bufLen)
	{
	  int month;
	  int day;
	  int year;
	  memset((char*)&timeTm, '\0', sizeof(tm));
/*	  strptime(&(timeBuf[bufPos]), "%D", &timeTm); */
	  sscanf(&(timeBuf[bufPos]), "%i/%i/%i", &month, &day, &year);
	  month = month - 1;
	  timeTm.tm_mon = month;
	  timeTm.tm_mday = day;
	  if (year > 100)
	    {
	      year -= 1900;
	    }
	  timeTm.tm_year = year;
#if HAVE_TIMELOCAL
	  retVal = timelocal(&timeTm);
#else
	  retVal = mktime(&timeTm);
#endif
	}
      return (long)retVal;
    }
}

String RFSDataFile::DateToString(const long date)

{
  tm timeStruct;
  long nanos = 0;
  char timeBuf[256];

  if (date <= 0)
    {	
      strcpy(timeBuf, "None set.");
    }
  else
    {
      timeStruct = *(localtime((time_t*)&date));
      strftime(timeBuf, 256, "%m/%d/%y", &timeStruct);
    }
  return String(timeBuf);
}

