/*
 * requestclient.C
 *
 *    This client demonstrates how to send YADDL objects from a server and
 * get them back, then does the same thing with arrays of objects.
 *
 * requestclient.C,v 1.1 1993/11/27 01:54:38 gjb Exp
 */

#include <objdce/statics.H>
#include <objdce/binding.H>
#include <requestee.h>
#include <request_cdl.H>
#include <iostream.h>

/*
 * Global variables
 */

Binding* svrBinding;			// The server to which we'll talk

static void TryOneRequest()

{
  RequestRec req;
  error_status_t rpcErr;
  String dceMsg;

  // Put some stuff in the record.

  req.requestType(RequestCredit);
  req.numAccounts(2);
  req.accounts(0, "000001");
  req.balances(0, 100.00);
  req.accounts(1, "000121");
  req.balances(1, 1989.12);

  // Send one request. The RPC function takes a RequestRecBase*, and we
  // have a RequestRec. We can just pass a RequestRec because it has a
  // conversion operator.
  cout << "Sending one request: " << endl;
  cout << req;
  rpcErr = SendOneRequest(*svrBinding, req);
  if (Binding::CheckErrorType(rpcErr, dceMsg) != Binding::ET_SUCCESS)
    {
      cout << "RPC error: " << dceMsg << endl;
    }
  else
    {
      cout << "Successfully sent one request." << endl;
    }

  // Get a request from the server. Again, we just use the conversion
  // operator on RequestRec.
  rpcErr = GetOneRequest(*svrBinding, req);
  if (Binding::CheckErrorType(rpcErr, dceMsg) != Binding::ET_SUCCESS)
    {
      cout << "RPC error: " << dceMsg << endl;
    }
  else
    {
      cout << "Got a request:" << endl;
      cout << req;
    }

  // Send a request to the server, then get back another one.
  // We'll use the conversion operator one more time for this.
  cout << "Sending one request, and getting one back. Sending:" << endl;
  cout << req;
  rpcErr = ExchangeOneRequest(*svrBinding, req);
  if (Binding::CheckErrorType(rpcErr, dceMsg) != Binding::ET_SUCCESS)
    {
      cout << "RPC error: " << dceMsg << endl;
    }
  else
    {
      cout << "Sent one request, and got one back:" << endl;
      cout << req;
    }
}
  
static void TryMultipleRequests()

{
  RequestRec req;
  RequestRecBaseArray reqArray;
  RequestRecBaseArray outReqArray;
  List<RequestRec> reqList;
  Listiter<RequestRec> reqListIter(reqList);
  error_status_t rpcErr;
  String dceMsg;
  int inc;
  char tmpAccount[10];

  // Fill in an array of requests to send to the server
  req.requestType(RequestDebit);
  req.numAccounts(1);
  for (inc = 0; inc < 6; inc++)
    {
      sprintf(tmpAccount, "%6i", inc);
      req.accounts(0, tmpAccount);
      req.balances(0, 101.11);
      reqList.put(req);
    }

  // Send the list of requests to the server. Use ClientListToArray to put
  // the requests into a RequestRecBaseArray for RPC transmission. This
  // function allocates storage for the array, so we free it when we're
  // done using ClientArrayFree.
  cout << "Sending " << reqList.length() << " requests" << endl;
  RequestRec::ClientListToArray(reqList, reqArray);
  rpcErr = SendSomeRequests(*svrBinding, &reqArray);
  if (Binding::CheckErrorType(rpcErr, dceMsg) != Binding::ET_SUCCESS)
    {
      cout << "RPC error: " << dceMsg << endl;
    }
  else
    {
      cout << "Sent " << inc << " requests to the server." << endl;
    }
  // Free up the storage for our array
  RequestRec::ClientArrayFree(reqArray);

  // Empty the list.
  reqListIter.reset();
  while (reqListIter.remove_next());

  // Get a list of requests back from the server. Copy the elements back
  // to a list of objects using ArrayToList. The server allocates the memory,
  // but we must free it on this side using ClientArrayFree.
  rpcErr = GetSomeRequests(*svrBinding, &reqArray);
  if (Binding::CheckErrorType(rpcErr, dceMsg) != Binding::ET_SUCCESS)
    {
      cout << "RPC error: " << dceMsg << endl;
    }
  else
    {
      RequestRec::ArrayToList(reqArray, reqList);
      cout << "In server array: " << reqArray.size << " requests." << endl;
      cout << "Got " << reqList.length() << " requests from server:" << endl;
      reqListIter.reset();
      while (reqListIter.next(req))
	{
	  cout << req << endl;
	}
    }
  RequestRec::ClientArrayFree(reqArray);
  reqListIter.reset();
  while (reqListIter.remove_next());

  // Fill in the array again
  req.requestType(RequestCredit);
  req.numAccounts(1);
  for (inc = 0; inc < 2; inc++)
    {
      sprintf(tmpAccount, "%6i", inc);
      req.accounts(0, tmpAccount);
      req.balances(0, 957.12);
      reqList.put(req);
    }
  
  // Send a list of requests to the server, and get another list back
  // We're going to allocate an array of records, and send it to the server,
  // then the server's going to send another list back. We can't
  // allocate the array on this side and allow the server free it, since
  // DCE doesn't support this. So, we send one array which we allocate,
  // and another that the server allocates.
  RequestRec::ClientListToArray(req, reqArray);
  rpcErr = ExchangeSomeRequests(*svrBinding, &reqArray, &outReqArray);
  if (Binding::CheckErrorType(rpcErr, dceMsg) != Binding::ET_SUCCESS)
    {
      cout << "RPC error: " << dceMsg << endl;
    }
  else
    {
      // Rather than create two lists, empty the one list before copying
      // items from the "out" array
      reqListIter.reset();
      while (reqListIter.remove_next());
      RequestRec::ArrayToList(outReqArray, reqList);
      cout << "Sent " << inc << " requests to the server." << endl;
      cout << "Got " << reqList.length() << " requests back." << endl;
      reqListIter.reset();
      // Empty the list while we iterate through the elements
      while (reqListIter.remove_next(req))
	{
	  cout << req << endl;
	}
    }
  // Now, free up both arrays
  RequestRec::ClientArrayFree(reqArray);
  RequestRec::ClientArrayFree(outReqArray);
}

static void PrintUsage(char* execName)

{
  cerr << "Usage: " << execName << " <server name>" << endl;
}

int main (int argc, char** argv)

{
  String svrName;

  if (argc != 2)
    {
      PrintUsage(argv[0]);
      return 1;
    }
  
  svrName = argv[1];
  cout << "Binding to " << svrName << "..." << endl;
  svrBinding = new Binding(svrName, RPCClient(Requestee,1,0),
			   LoginContext::CALL);
  if (!svrBinding->IsValid())
    {
      cout << "Could not find server." << endl;
      delete svrBinding;
      return 1;
    }

  TryOneRequest();
  TryMultipleRequests();

  cout << "Done." << endl;
  delete svrBinding;
  return 0;
}
