/*
 * svrtest_funcs.C
 *
 *    These functions implement our TRPC/RPC server.
 *
 * svrtest_funcs.C,v 1.2 1994/04/07 19:29:21 gjb Exp
 */

#include <svrtest_cdl_sfs.H>
#include <nontransvrtest.h>
extern "C"
{
#include <transvrtest.h>
}
#include <objtran/transaction.H>
#include <objtran/tranbinding.H>
#include <objtran/transerver.H>
#include <objdce/abortreason.H>
#include <dce/rpcsts.h>

extern TranServer* MainServer;
extern SFSSimpleFile* MainFile;

/*
 * Non-transactional interfaces:
 *    NonTranInsert does the same thing as TranInsert. However, it must
 * first create a trpc_handle_t and start a transaction before calling
 * the TranInsert function.
 *    The "NonTranFind" use context handles and the ability of SFS to save
 * its context to let the user non-transactionally browse through all the
 * records that match the given key.
 */

error_status_t NonTranInsert(handle_t handle,
			     SimpleRecordBase* records,
			     unsigned long* dbErr)
{
  trpc_handle_t tranHandle;

  trpc_CreateBinding(handle, TRAN_APPL_ID_NULL,
		     TRAN_ADDRESS_NULL, &tranHandle);
  NewTransaction tran;
  
  TranInsert(tranHandle, records);
  tran.Commit();
  if (tran.Aborted())
    {
      *dbErr = (error_status_t)tran.GetAbortReason()->GetCode();
    }
  else
    {
      *dbErr = 0;
    }
  trpc_FreeBinding(tranHandle);
  return 0;
}

error_status_t NonTranFindBegin(handle_t handle,
				long lKey,
				SimpleRecordContext* context,
				unsigned long* dbErr)
{
  SimpleRecord key;
  CDLStatus cdlErr = CDL_SUCCESS;

  if (!MainServer->IsClntAuthorized(handle))
    {
      return rpc_s_binding_has_no_auth;
    }

  NewTransaction tran;

  key.key(lKey);
  cdlErr = MainFile->Select(SimpleFile::PRIMARY, &key, &key, FORWARDS);
  if (cdlErr == CDL_SUCCESS)
    {
      cdlErr = MainFile->SaveContext(*context);
    }
  tran.Commit();
  if (cdlErr != CDL_SUCCESS)
    {
      cout << *(tran.GetAbortReason()) << endl;
      *context = (SimpleRecordContext)0;
    }
  *dbErr = (unsigned long)cdlErr;
  return 0;
}

error_status_t NonTranFindNext(SimpleRecordContext* context,
			       SimpleRecordBase* next,
			       unsigned long* dbErr)
{
  SimpleRecord outRec;
  NewTransaction tran;
  CDLStatus cdlErr = CDL_SUCCESS;

  cdlErr = MainFile->RestoreContext(*context);
  if (cdlErr == CDL_SUCCESS)
    {
      cdlErr = MainFile->ReadCurrent(READ, NEXT, outRec);
    }
  if (cdlErr == CDL_SUCCESS)
    {
      cdlErr = MainFile->SaveContext(*context);
    }
  tran.Commit();
  if (cdlErr != CDL_SUCCESS)
    {
      cout << *(tran.GetAbortReason()) << endl;
      *context = (SimpleRecordContext)0;
    }
  else
    {
      outRec.Copy(*next);
    }
  *dbErr = (unsigned long)cdlErr;
  return 0;
}

error_status_t NonTranFindDone(SimpleRecordContext* context,
			       unsigned long* dbErr)

{
  NewTransaction tran;
  CDLStatus cdlErr = CDL_SUCCESS;

  // Restoring the context frees it up
  cdlErr = MainFile->RestoreContext(*context);
  tran.Commit();
  if (cdlErr != CDL_SUCCESS)
    {
      cout << *(tran.GetAbortReason()) << endl;
    }
  *context = (SimpleRecordContext)0;
  *dbErr = (unsigned long)cdlErr;
  return 0;
}

void SimpleRecordContext_rundown(SimpleRecordContext context)

{
  NewTransaction tran;
  CDLStatus cdlErr;

  cdlErr = MainFile->RestoreContext(context);
  tran.Commit();
  if (cdlErr != CDL_SUCCESS)
    {
      cout << *(tran.GetAbortReason()) << endl;
    }
}

/*
 * Transactional routines
 *
 *    These do the real work, including checking for proper authorization.
 */

void TranInsert(trpc_handle_t handle,
		SimpleRecordBase* records)

{
  SimpleRecord record = *records;
  ExistingTransaction tran;

#ifndef WK_ENDPOINTS
  if (!MainServer->IsClntAuthorized(handle))
    {
      tran.Abort("Client is not authorized", rpc_s_binding_has_no_auth, 0);
      return;
    }
#endif
  List<SimpleRecord> recList;
  recList.put(record);
  MainFile->InsertList(recList);
}

void TranFind(trpc_handle_t handle, long keyVal,
	      SimpleRecordBaseArray* records)
{
  List<SimpleRecord> recordList;
  SimpleRecord key;
  ExistingTransaction tran;

#ifndef WK_ENDPOINTS
  if (!MainServer->IsClntAuthorized(handle))
    {
      tran.Abort("Client is not authorized", rpc_s_binding_has_no_auth, 0);
      // OK to do if the list is empty -- in fact, a good idea
      SimpleRecord::ServerListToArray(recordList, *records);
      return;
    }
#endif
  key.key(keyVal);
  MainFile->ReadAll(READ, SimpleFile::PRIMARY, key,
		    FORWARDS, recordList);
  SimpleRecord::ServerListToArray(recordList, *records);
}

void TranFindData(trpc_handle_t handle, unsigned char* data,
		  SimpleRecordBase* records)
{
  SimpleRecord outRec;
  SimpleRecord key;
  ExistingTransaction tran;

#ifndef WK_ENDPOINTS
  if (!MainServer->IsClntAuthorized(handle))
    {
      tran.Abort("Client is not authorized", rpc_s_binding_has_no_auth, 0);
      return;
    }
#endif
  key.data((const char*)data);
  MainFile->ReadUnique(READ, SimpleFile::DATA,
		       key, outRec);
  outRec.Copy(*records);
}

void TranListData(trpc_handle_t handle,
		  SimpleRecordBaseArray* records)
{
  List<SimpleRecord> recordList;
  ExistingTransaction tran;

#ifndef WK_ENDPOINTS
  if (!MainServer->IsClntAuthorized(handle))
    {
      tran.Abort("Client is not authorized", rpc_s_binding_has_no_auth, 0);
      // OK to do if the list is empty -- in fact, a good idea
      SimpleRecord::ServerListToArray(recordList, *records);
      return;
    }
#endif
  MainFile->ReadAll(READ, SimpleFile::DATA, (SimpleRecord*)0,
		    (SimpleRecord*)0, FORWARDS, recordList);	
  SimpleRecord::ServerListToArray(recordList, *records);
}

			
