/*
 * SFSFileBase
 *
 *    Here, we do most of the work associated with SFS, including managing
 * a list of OFDs and a table of transaction IDs and batched SFS requests.
 * Much of this code was copied verbatim from SFSFile and its related classes
 * in objtran. Eventually, this code will replace the objtran code.
 *
 * sfsfilebase.C,v 1.17 1994/04/07 19:27:13 gjb Exp
 */

    /*********************************************************************
    Copyright (C) 1993-1994 by Citibank, N.A., all rights reserved.

    Permission to use, copy, modify and distribute this software and its 
    documentation for any purpose and without fee, except possible 
    redistribution costs, is hereby granted, provided that (1) the above 
    copyright notice appears in all copies and (2) both that copyright-
    notice and this permission-notice appear in supporting documentation and 
    (3) the name of Citibank, N.A. is not used in advertising or publicity 
    pertaining to distribution of the software without specific and written 
    prior permission of Citibank, N.A.

    The Software is provided "as is."  Citibank, N.A. disclaims all 
    warranties, express or implied, including but not limited to warranties 
    of suitability, merchantability and fitness for any purpose.  In no 
    event shall Citibank, N.A. be responsible for any loss of use, data or 
    profit or for any other special, indirect, incidental or consequential 
    damages, neither in an action of contract nor in an action of tort, 
    arising out of or in connection with the use or performance of this 
    software.
    *********************************************************************/


extern "C" {
#include <pthread.h>
}
#include <objdce/binding.H>
#include <cdl/sfsfilebase.H>
#include <cdl/cdldefs.H>
#include <objdce/error.H>
#include <objdce/aclclient.H>
#include <objdce/statics.H>
extern "C" {
#include <threadTid/threadTid.h>
}

#ifdef _AIX
#include <List.c>
#include <set.c>

#pragma define (Const_listiter<SFSTranInfo>)
#pragma define (List<SFSTranInfo>)
#pragma define (Listiter<SFSTranInfo>)
#pragma define (Set<SFSTranInfo>)
#pragma define (lnnk_ATTLC<SFSTranInfo>)
#endif

/* 
 * Constants for timeout values in seconds: used by Open 
 */

static const unsigned long SFSNormalTimeout = 30;
static const unsigned long SFSLongTimeout = 240;
static const unsigned long SFSBytesPerPage = 4096;
static const unsigned long SFSNullOfd = 0xfffffffe;

/*
 * More constants 
 */

#define RecNameAttribute "CDL_RECORD_NAME"
static const unsigned int MaxAttrNameLen = 1024;
static const int NumRetries = 5;

/* 
 * Static members
 */

MutexLock* SFSFileBase::batchLatch = 0;
Set<SFSBatchQueue>* SFSFileBase::batchQueues = 0;
pthread_once_t SFSFileBase::initedBatches = pthread_once_init;

/*
 * SFSTranInfo operations
 */

SFSTranInfo::SFSTranInfo()

{
  tid = TRAN_TID_NULL;
  ofd = SFSNullOfd;
}

SFSTranInfo::SFSTranInfo(const Transaction& tran)

{
  tid = tran;
  ofd = SFSNullOfd;
}

SFSTranInfo::SFSTranInfo(const SFSTranInfo& obj)

{
  tid = obj.tid;
  ofd = obj.ofd;
}

SFSTranInfo& SFSTranInfo::operator = (const SFSTranInfo& obj)

{
  tid = obj.tid;
  ofd = obj.ofd;
  return *this;
}

int SFSTranInfo::operator == (const SFSTranInfo& obj) const

{
  return tran_TidEqual(tid, obj.tid);
}

Set_or_Bag_hashval Set<SFSTranInfo>::hash(const SFSTranInfo& obj)

{
  return (Set_or_Bag_hashval)tran_TidHash(obj.tid);
}

/* 
 * SFSFileBase operations
 */

/*
 * Constructor
 *    Do one-time static member initialization via pthread_once and other
 * functions, and initialize whatever else we have.
 */

SFSFileBase::SFSFileBase() 
    : CDLFileBase(), TransactionMember(), Terminated()

{
  ObjtranFirstInit();
  CDLFirstInit();
  opened = false;
  if (pthread_once(&initedBatches, FirstTimeInit) != 0)
    {
      printf("Error in static member initialization in %s at line %i\n",
	     __FILE__, __LINE__);
      exit(1);
    }
}

/*
 * FirstTimeInit
 *    This static function initializes the static members of this class. 
 * It should be called once and only once by pthread_once
 */

void SFSFileBase::FirstTimeInit()

{
  CDLReport.Report(ErrorLevels::TraceMsg, "Initializing batch queues\n");
  batchLatch = new MutexLock();
  batchQueues = new Set<SFSBatchQueue>();
}

/*
 * Destructor
 *    Call HandleTermination to cause an orderly shutdown.
 */

SFSFileBase::~SFSFileBase()

{
  HandleTermination();
}

/*
 * HandleTermination
 *    Close our OFDs if there are any still open.
 */

void SFSFileBase::HandleTermination()

{
   Boolean iWasOpen;

   ofdLatch.Lock();
   iWasOpen = opened;
   ofdLatch.Unlock();
   if (iWasOpen)
     {
       ICloseFile();
     }
}

/*
 * NoAbort
 *    If the error message passed is non-fatal, we should not abort the
 * transaction.
 */

Boolean SFSFileBase::NoAbort(CDLStatus err)

{
  Boolean retVal;
  switch (err)
    {
    case CDL_END_OF_RANGE:		// Don\'t abort on these errors
    case CDL_DIRTY_READ:
      CDLReport.Report(ErrorLevels::TraceMsg,
		       "Received either END_OF_RANGE or DIRTY_READ Error. Not Aborting\n");
      retVal = true;
      break;
    default:
      retVal = false;
      break;
    }
  return retVal;
}

/*
 * RetryError
 *    Return true if the error message indicates the operation should
 * be retried.
 */

Boolean SFSFileBase::RetryError(const sfs_status_t err)

{
  Boolean retVal;

  switch (err)
    {
    case SFS_OPERATION_TIMED_OUT:
      CDLReport.Report(ErrorLevels::TraceMsg,
		       "Received SFS_OPERATION_TIMED_OUT. Not Aborting\n");
      retVal = true;
      break;
    case SFS_COMMUNICATION_ERROR:
      CDLReport.Report(ErrorLevels::TraceMsg,
		       "Received SFS_COMMUNICATION_ERROR. Not Aborting\n");
      retVal = true;
      break;
    default:
      retVal = false;
      break;
    }
  return retVal;
}

/*
 * CheckError
 *    Check whether an SFS error occurred. If so, output it using the
 * proper message in the message catalog, and abort the current transaction
 * if there is an active transaction.
 */

CDLStatus SFSFileBase::CheckError(const int msgNum, const sfs_status_t err,
				  Transaction& tran, ErrorLevels::ReportingLevels level)
{
  if (err != SFS_SUCCESS)
    {
      CDLStatus retVal = ConvertStatus(err);
      if (NoAbort(retVal))
	{
	return retVal;					    // A non-fatal error
	}
      else
	{
	char	statusBuf[ENCINA_MAX_STATUS_STRING_SIZE];

	encina_StatusToString(err, ENCINA_MAX_STATUS_STRING_SIZE, statusBuf);
	CDLReport.VReportMsg(level, msgNum, statusBuf);   // Print out error message
	if (threadTid_Lookup() != TRAN_TID_NULL)
	    {
	    tran.Abort(statusBuf, (const int)retVal, CDLId);	// Abort Transaction
	    }
	return retVal;
	}
    }
  else
    return CDL_SUCCESS;
}

CDLStatus SFSFileBase::ThrowError(const int msgNum, CDLStatus err,
				  Transaction& tran, ErrorLevels::ReportingLevels level)
{
  CDLReport.ReportMsg(level, msgNum);
  if (threadTid_Lookup() != TRAN_TID_NULL)
    tran.Abort("Internal CDL error", (const int)err, CDLId);
  return CDL_SUCCESS;
}

/*
 * ConvertStatus
 *    Convert an sfs_status_t to a CDLStatus.
 */

CDLStatus SFSFileBase::ConvertStatus(const sfs_status_t err)

{	
  CDLStatus ret;
  switch(err)
    {
    case SFS_SUCCESS:
      ret = CDL_SUCCESS;
      break;
    case SFS_CRP_UNDEFINED:
      ret = CDL_CRP_UNDEFINED;
      break;
    case SFS_DISALLOWED_ON_PRIMARY_INDEX:
      ret = CDL_DISALLOWED_ON_PRIMARY_INDEX;
      break;
    case SFS_DUPLICATE_IN_UNIQUE_INDEX:
      ret = CDL_DUPLICATE_IN_UNIQUE_INDEX;
      break;
    case SFS_END_OF_KEY_RANGE:
    case SFS_CURRENT_UNDEFINED:
      ret = CDL_END_OF_RANGE;
      break;
    case SFS_FILE_NAME_EXISTS:
      ret = CDL_FILE_NAME_EXISTS;
      break;
    case SFS_INDEX_NOT_ACTIVE:
      ret = CDL_INDEX_NOT_ACTIVE;
      break;
    case SFS_INDEX_NOT_UNIQUE:
      ret = CDL_INDEX_NOT_UNIQUE;
      break;
    case SFS_INSUFFICIENT_AUTHORITY:
    case SFS_INSUFFICIENT_AUTHENTICATION_LEVEL:
    case SFS_INSUFFICIENT_FILE_ACCESS_RIGHTS:
    case SFS_INSUFFICIENT_FILE_SYSTEM_ACCESS_RIGHTS:
      ret = CDL_PERMISSION_DENIED;
      break;
    case SFS_NO_SUCH_FILE:
      ret = CDL_FILE_NOT_FOUND;
      break;
    case SFS_NO_SUCH_FILE_SYSTEM:
    case SFS_NO_SUCH_VOLUME:
      ret = CDL_FILE_SYSTEM_NOT_FOUND;
      break;
    case SFS_NO_SUCH_KEY_VALUE:
      ret = CDL_KEY_NOT_FOUND;
      break;
    case SFS_OPERATION_TIMED_OUT:
      ret = CDL_TIMED_OUT;
      break;
    case SFS_USER_TRANSACTION_REQUIRED:
      ret = CDL_TRANSACTION_REQUIRED;
      break;
    case SFS_VOLUME_IO_ERROR:
      ret = CDL_IO_ERROR;
      break;
    case SFS_VOLUME_SPACE_EXHAUSTED:
    case SFS_AREA_SPACE_EXHAUSTED:
      ret = CDL_VOLUME_FULL;
      break;
    case SFS_COMMUNICATION_ERROR:
      ret = CDL_COMM_ERROR;
      break;
    case SFS_READ_FROM_MODIFIED_FILE:
      ret = CDL_DIRTY_READ;
      break;
    case SFS_MAX_NUMBER_RECORDS_EXCEEDED:
      ret = CDL_FILE_TOO_LARGE;
      break;
    case SFS_SLOT_OCCUPIED:
      ret = CDL_SLOT_OCCUPIED;
      break;
    case SFS_FILE_DAMAGED:
      ret = CDL_FILE_DAMAGED;
      break;
    default:
      ret = CDL_SFS_ERROR;
      break;
    }
  return ret;
}

/* 
 * CheckOpen
 *    Check if the file is open or not open
 */

CDLStatus SFSFileBase::CheckOpen(Boolean shouldBeOpen, Transaction&)

{
  ofdLatch.Lock();
  Boolean iWasOpened = opened;
  ofdLatch.Unlock();

  if (shouldBeOpen && !iWasOpened)
    {
      CDLReport.ReportMsg(ErrorLevels::ErrorMsg, 11000);
      return CDL_FILE_NOT_OPEN;
    }
  else if (!shouldBeOpen && iWasOpened)
    {
      CDLReport.ReportMsg(ErrorLevels::ErrorMsg, 11001);
      return CDL_FILE_ALREADY_OPEN;
    }
  else
    return CDL_SUCCESS;
}

/* 
 * GetOfd
 *    This important function manages the OFD pool. It checks the hash table
 * of open OFDs for one that matches the specified transaction. If it finds
 * one, it uses that. If not, it gets one from the list of free OFDs. If
 * there are no free OFDs, it opens a new one. Note that the transaction
 * in question is always a top-level transaction, or the top-level parent
 * of the transaction given.
 */

CDLStatus SFSFileBase::GetOfd(Transaction& tran, sfs_ofd_t& ret)

{
  CDLStatus err = CDL_SUCCESS;
  sfs_status_t status;
  tran_tid_t tid = tran;
  tran_tid_t topTid;
  int isTopLevel;
  SFSTranInfo tranKey;

  if (openFlags & (CDL_NONTRANSACTIONAL_OPEN | CDL_EXCLUSIVE_OPEN))
    {
      CDLReport.Report(ErrorLevels::InfoMsg, "Setting OFD to masterOfd\n");
      ret = masterOfd;
      return CDL_SUCCESS;
    }
  if (tran.Aborted())
    {
      CDLReport.Report(ErrorLevels::TraceMsg, "Setting OFD to SFSNullOfd\n");
      ret = SFSNullOfd;
      return CDL_TRANSACTION_ABORTED;
    }

  tran_TidIsTopLevel(tid, &isTopLevel);
  if (isTopLevel)
    topTid = tid;
  else
    tran_TidTopAncestor(tid, &topTid);
  ExistingTransaction topTran(topTid);
  tranKey.tid = topTid;
  ofdLatch.Lock();
  const SFSTranInfo* tInfo = activeTrans.contains(tranKey);
  if (tInfo != (SFSTranInfo*)0)		// There was an ofd for this tid
    {
      ret = tInfo->ofd;
    }
  else if (freeOfds)			// There was a free ofd
    {
      topTran.Register(this, REGISTER_FINISHED);  // Let me know when done
      ret = freeOfds.get();
      tranKey.ofd = ret;
      activeTrans.insert(tranKey);
      CDLReport.VReport(ErrorLevels::TraceMsg, 
			"TID %lu: Associating with open OFD number %lu\n",
			(unsigned long)topTid,
			(unsigned long)ret);
    }
  else					// Open a new ofd
    {
      topTran.Register(this, REGISTER_FINISHED);
      status = sfs_OpenFileByOfd(masterOfd, timeout, isolation,
				 timeout, false, &ret);
      ofdLatch.Unlock();
      err = CheckError(10015, status, tran, ErrorLevels::ErrorMsg);
      ofdLatch.Lock();
      if (err != CDL_SUCCESS)
	ret = SFSNullOfd;
      else
	{
	  tranKey.ofd = ret;
	  activeTrans.insert(tranKey);
	  CDLReport.VReport(ErrorLevels::InfoMsg,
			    "TID %lu: Associating with new OFD number %lu\n", 
			    (unsigned long)topTid,
			    (unsigned long)ret);
	}
    }
  ofdLatch.Unlock();
  return err;
}

/*
 * FinishedCallback
 *    This function is inherited from the TransactionMember class, and is
 * called when the transaction is done. Here, we put the ofd back on the list
 * of free OFDs, and remove this transaction's entry from the batch table,
 * if there is one.
 */

void SFSFileBase::FinishedCallback(const Transaction& tran)

{
  tran_tid_t tid = tran;
  SFSTranInfo tranKey(tran);
  SFSBatchQueue batchKey(tran);

  ofdLatch.Lock();
  const SFSTranInfo* foundTran = activeTrans.contains(tranKey);
  if (foundTran != (SFSTranInfo*)0)
    {
      freeOfds.put(foundTran->ofd);
      activeTrans.remove(tranKey);
      CDLReport.VReport(ErrorLevels::TraceMsg,
			"Freeing OFD number %lu\n", (unsigned long)(foundTran->ofd));
    }
  ofdLatch.Unlock();
  batchLatch->Lock();
  const SFSBatchQueue* foundBatch = batchQueues->contains(batchKey);
  if (foundBatch != (SFSBatchQueue*)0)
    {
      CDLReport.VReport(ErrorLevels::TraceMsg, "TID %lu: Deleting batch queue\n", tid);
      batchQueues->remove(batchKey);
    }
  batchLatch->Unlock();
}
	
/*
 * IsBatching
 *    Return whether batching has been turned on for this transaction.
 */

Boolean SFSFileBase::IsBatching(const Transaction& tran)

{
  SFSBatchQueue testKey(tran);
  Boolean retVal;

  batchLatch->Lock();
  const SFSBatchQueue* foundBatch = batchQueues->contains(testKey);
  if (foundBatch == (SFSBatchQueue*)0)
    retVal = false;
  else
    retVal = (foundBatch->batching > 0);
  batchLatch->Unlock();
  return retVal;
}

/*
 * AddBatch
 *    Add an object to the batch queue. This object contains all the info
 * SFS needs to process the request later.
 */

Boolean SFSFileBase::AddBatch(const Transaction& tran, const SFSBatchBase* batch)

{
  SFSBatchQueue testKey(tran);
  Boolean retVal;

  batchLatch->Lock();
  const SFSBatchQueue* foundBatch = batchQueues->contains(testKey);
  if (foundBatch == (SFSBatchQueue*)0)
    retVal = false;
  else
    {
     CDLReport.VReport(ErrorLevels::TraceMsg,
		       "TID %lu: Adding item to batch queue\n",
		       (tran_tid_t)tran);
     SFSBatchQueue* nonConstBatch = (SFSBatchQueue*)foundBatch;
     nonConstBatch->queue.put(batch);
     retVal = true;
    }
  batchLatch->Unlock();
  return retVal;
}

/*
 * GetDirection, GetMode
 *    Convert from the CDL type to the corresponding SFS type.
 */

sfs_selector_t SFSFileBase::GetDirection(CDLRecordDir iDir) const

{
  sfs_selector_t retVal;
  switch(iDir)
    {
    case NEXT:
      retVal = sfs_next;
      break;
    case PREVIOUS:
      retVal = sfs_previous;
      break;
    case CURRENT:
      retVal = sfs_current;
      break;
    }
  return retVal;
}

sfs_lockMode_t SFSFileBase::GetMode(CDLReadLock cLock) const

{
  sfs_lockMode_t retVal;
  switch (cLock)
    {
    case READ:
    case LAZY:
      retVal = sfs_readLock;
      break;
    case WRITE:
      retVal = sfs_writeLock;
      break;
    case BRACKET:
      retVal = sfs_bracketReadLock;
      break;
    case NONE:
      retVal = sfs_noLock;
      break;
    }
  return retVal;
}
      
/*
 * IOpenFile
 *    Open the file by opening using the master ofd. Future accesses will
 * use copies of this ofd. By default, files are opened for shared, transactional,
 * serializable access with 30-second timeouts and no duplicate checking.
 * Flags can request exclusive access, non-transactional access (in which
 * case data is written to disk when each operation completed) and longer
 * timeouts.
 */

CDLStatus SFSFileBase::IOpenFile(const String& dbName, const String& fileName,
				 const String& typeName, unsigned short flags,
			         Boolean noNameCheck)
{
  ExistingTransaction tran;
  sfs_ofdSpec_t spec;
  sfs_status_t err = SFS_SUCCESS;
  CDLStatus ret;
  String realName;
  char recNameBuf[MaxAttrNameLen];
  unsigned int realAttrNameLen;
  int retries = 0;

  realName = dbName;
  realName += "/";
  realName += fileName;
  const char* fName = realName;

  ACLclient acl;
  openFlags = flags;
  ret = CheckOpen(false, tran);
  if (ret != CDL_SUCCESS)
    return ret;
  
  String perm;
  spec.authority = 0;
  for(int b=0; b<6; b++)
    {
      switch (b)
	{
          case 0:
	  perm = "A";
             if(acl.IsAuthorized(realName, 0, perm))
               spec.authority |= SFS_ADMINISTER_FILE;
             break;
          case 1:
             perm = "D";
             if(acl.IsAuthorized(realName, 0, perm))
               spec.authority |= SFS_DELETE_FILE;
             break;
          case 2:
             perm = "I";
             if(acl.IsAuthorized(realName, 0, perm))
               spec.authority |= SFS_INSERT_FILE;
             break;
          case 3:
             perm = "Q";
             if(acl.IsAuthorized(realName, 0, perm))
               spec.authority |= SFS_INQUIRE_FILE;
             break;
          case 4:
             perm = "R";
             if(acl.IsAuthorized(realName, 0, perm))
               spec.authority |= SFS_READ_FILE;
             break;
          case 5:
             perm = "U";
             if(acl.IsAuthorized(realName, 0, perm))
               spec.authority |= SFS_UPDATE_FILE;
             break;
	  default: 
             break;
        } 
    }

        
  if (openFlags & CDL_EXCLUSIVE_OPEN)
    {
      spec.accessMode = sfs_exclusiveAccess;
      spec.authority |= SFS_EXCLUSIVE_FILE;
    }
  else
    spec.accessMode = sfs_sharedAccess;
  if (openFlags & CDL_NONTRANSACTIONAL_OPEN)
    {
      spec.consistency = sfs_nonTransactional;
      isolation = sfs_nonTranOperational;
      spec.operationalForce = true;
    }
  else
    {
      spec.consistency = sfs_transactional;
      isolation = sfs_serializability;
      spec.operationalForce = false;
    }
  spec.isolationLevel = isolation;
  spec.autoClose = false;
  spec.duplicates = sfs_noDetection;
  if (openFlags & CDL_LONG_TIMEOUTS)
    timeout = SFSLongTimeout;
  else
    timeout = SFSNormalTimeout;
  spec.operationTimeout = timeout;
  do
    err = sfs_OpenFile((char*)fName, &spec, timeout, &masterOfd);
  while (RetryError(err) && (retries++ < NumRetries) && !tran.Aborted());
  ret = CheckError(10000, err, tran, ErrorLevels::InfoMsg);
  if (ret == CDL_SUCCESS)
    {
      CDLReport.VReport(ErrorLevels::InfoMsg, "Opened master file ofd for file %s\n", fName);
      ofdLatch.Lock();
      opened = true;
      freeOfds.put(masterOfd);
      ofdLatch.Unlock();
      if (!noNameCheck)
	{				// Now get the record name attribute
	  err = sfs_GetAttribute(masterOfd, RecNameAttribute, MaxAttrNameLen,
				 (void*)recNameBuf, &realAttrNameLen);
	  if (err == SFS_NO_SUCH_ATTRIBUTE)		
	    CDLReport.ReportMsg(ErrorLevels::ErrorMsg, 11004);
	  else
	    {
	      ret = CheckError(10020, err, tran, ErrorLevels::ErrorMsg);
	      if ((ret == CDL_SUCCESS) &&
		  strcmp(typeName, recNameBuf))	// Type name different
		CDLReport.VReportMsg(ErrorLevels::ErrorMsg, 11003, recNameBuf,
				     (const char*)typeName);
	    }
	}
    }
  return ret;
}
  
/*
 * ICloseFile
 *    Close all the open OFDs. First set the "open" flag to false so no more
 * SFS operations will be begun, then wait until all the OFDs are free. If
 * this doesn't happen in a little while, go ahead anyway. Then close all the
 * free OFDs.
 */

CDLStatus SFSFileBase::ICloseFile()

{
  CDLStatus status;
  int waitCount = 0;
  int freeLength;
  sfs_ofd_t ofd;
  ExistingTransaction tran;
  Boolean aFailure = false;
  sfs_status_t err;
  timespec interval;
  interval.tv_sec = 5;
  interval.tv_nsec = 0;
 
  if ((status = CheckOpen(true, tran)) != CDL_SUCCESS)
    return status;
  ofdLatch.Lock();
  opened = false;			// No more new SFS ops will be allowed
  while ((waitCount < 10) && (activeTrans.size() > 0))
    {
      CDLReport.Report(ErrorLevels::InfoMsg, 
		       "Waiting for transactions to finish in ICloseFile\n");
      waitCount++;
      pthread_delay_np(&interval);	// Wait for trans to end
    }
  if (waitCount >= 10)
    CDLReport.Report(ErrorLevels::InfoMsg, "Giving up and closing OFDs\n");		// Giving up!
  freeLength = freeOfds.length();
  Listiter<sfs_ofd_t> freeIter(freeOfds);
  while (freeIter.remove_next(ofd))			// While OFDs in list
    {
      err = sfs_CloseOfd(ofd);
      if (!CheckError(10001, err, tran, ErrorLevels::InfoMsg))
	aFailure = true;
      CDLReport.VReport(ErrorLevels::InfoMsg,
			"Closing OFD number %lu\n", (unsigned long)ofd);
    }
  // The master OFD is now in the free list, so don't close it here
  ofdLatch.Unlock();
  
  if ((waitCount >= 10) || aFailure)
    status = CDL_FILES_LEFT_OPEN;
  else
    status = CDL_SUCCESS;
  return status;
}

/*
 * ICreate
 *    Create a new file. The information passed is an SFSBaseFileSpec
 * object created by the CDL-generated class. 
 */
 
CDLStatus SFSFileBase::ICreate(SFSBaseFileSpec& spec, 
			       const String& dbName, const String& fileName,
			       const String& volName, unsigned long numBytes)
{
  CDLStatus ret;
  ExistingTransaction tran;
  sfs_fileSpec_t* iSpec = spec;
  unsigned long pages;
  sfs_status_t err;
  sfs_ofd_t ofd;
  char recNameBuffer[MaxAttrNameLen + 1];
  String realFileName;
  int retries = 0;

  if ((ret = CheckOpen(false, tran)) != CDL_SUCCESS)
    return ret;
  const char* vName = volName;
  realFileName = dbName;
  realFileName += '/';
  realFileName += fileName;
  const char* fName = realFileName;
  pages = numBytes / SFSBytesPerPage;
  iSpec->storageSpec.volumeName = (char*)vName;
  iSpec->storageSpec.allocated = pages;
  iSpec->storageSpec.utilized = 0;
  CDLReport.VReport(ErrorLevels::InfoMsg,
		    "Creating File: %s - Allocated: %ld - Volume: %s\n",
		    fName, pages, vName);
  do
    err = sfs_CreateFile((char*)fName, iSpec);
  while (RetryError(err) && (retries++ < NumRetries) && !tran.Aborted());
  ret = CheckError(10002, err, tran, ErrorLevels::InfoMsg);

  if (ret == CDL_SUCCESS)
    {
      CDLReport.VReport(ErrorLevels::InfoMsg, "Successfully created file %s\n", fName);
      ret = IOpenFile(dbName, fileName, spec.TypeName(),
		      CDL_EXCLUSIVE_OPEN | CDL_NONTRANSACTIONAL_OPEN, true);	
    }
  if (ret == CDL_SUCCESS)
    {
      ret = GetOfd(tran, ofd);
      if (ret == CDL_SUCCESS)
	{
	  strncpy(recNameBuffer, spec.TypeName(), MaxAttrNameLen);
	  recNameBuffer[MaxAttrNameLen] = '\0';
	  retries = 0;
	  do
	    err = sfs_SetAttribute(ofd, RecNameAttribute, (void*)recNameBuffer,
				   strlen(recNameBuffer) + 1);
	  while (RetryError(err) && (retries++ < NumRetries) && !tran.Aborted());
	  ret = CheckError(10019, err, tran, ErrorLevels::InfoMsg);
	}
      ICloseFile();
    }
  return ret;
}

/*
 * ISelect
 *    Select a range of records. The "start" and "end" parameters specify the
 * key range -- the keys will be selected including the keys at both ends.
 * If both keys are the same, a range will be selected containing all
 * records that match that one key. If both keys are NULL, the entire
 * file will be selected. If either key is NULL, the range will start at
 * the beginning of the file or end at the end as appropriate.
 */

CDLStatus SFSFileBase::ISelect(const String& indexName, SFSBaseKey* start,
			       SFSBaseKey* end, CDLIndexDir direction)
{
  CDLStatus ret;
  ExistingTransaction tran;
  sfs_ofd_t ofd;
  sfs_key_t* startK;
  sfs_key_t* endK;
  sfs_scanDirection_t dir;
  sfs_endPointDefinition_t startPt;
  sfs_endPointDefinition_t endPt;
  sfs_status_t err;
  int retries = 0;

  if ((ret = GetOfd(tran, ofd)) != CDL_SUCCESS)
    return ret;
  if (start == (SFSBaseKey*)0)
    {
      startPt = sfs_begin;
      startK = (sfs_key_t*)0;
    }
  else
    {
      startPt = sfs_gteqKey;
      startK = *start;
    }
  if (end == (SFSBaseKey*)0)
    {
      endPt = sfs_end;
      endK = (sfs_key_t*)0;
    }
  else
    {
      endPt = sfs_lteqKey;
      endK = *end;
    }
  if (direction == BACKWARDS)
    dir = sfs_backwardScan;
  else
    dir = sfs_forwardScan;
  const char* iName = indexName;
  CDLReport.VReport(ErrorLevels::TraceMsg, "Selecting Range for Key: %s\n", iName);
  do
    err = sfs_SelectKeyRange(ofd, (char*)iName, startPt, startK, 
			     endPt, endK, dir);
  while (RetryError(err) && (retries++ < NumRetries) && !tran.Aborted());
  ret = CheckError(10003, err, tran, ErrorLevels::InfoMsg);
  if (ret == CDL_SUCCESS)
    {
      CDLReport.VReport(ErrorLevels::TraceMsg, "TID %lu: Selected a range of records\n",
		       (tran_tid_t)tran);
    }
  return ret;
}
		    
/*
 * IInsert
 *    Insert a record into the file. From here on in, most of the functions
 * are pretty simple.
 */

CDLStatus SFSFileBase::IInsert(SFSBaseRecord& base)

{
  CDLStatus ret;
  ExistingTransaction tran;
  sfs_status_t err;
  sfs_record_t* rec;
  sfs_ofd_t ofd;
  int retries = 0;

  ret = GetOfd(tran, ofd);
  if (ret == CDL_SUCCESS)
    {
      rec = base;
      if (IsBatching(tran))
	{
	  SFSBatchInsert* ins = new SFSBatchInsert(ofd, base);
	  if (!AddBatch(tran, ins))
	    {
	      ret = CDL_BATCH_QUEUE_CORRUPTED;
	      ThrowError(10004, ret, tran, ErrorLevels::InfoMsg);
	    }
	}
      else
	{
	  do
	    {
	      err = sfs_Insert(ofd, rec);
	    }
	  while (RetryError(err) && (retries++ < NumRetries) 
		 && !tran.Aborted());

	  ret = CheckError(10004, err, tran, ErrorLevels::InfoMsg);
	  if (ret == CDL_SUCCESS)
	    {
	      CDLReport.VReport(ErrorLevels::InfoMsg, "TID %lu: Inserted a record\n", 
				(tran_tid_t)tran);
	    }
	}
    }
  return ret;
}

/*
 * IReadCurrent
 *    Read the record at the current position.
 */

CDLStatus SFSFileBase::IReadCurrent(CDLReadLock cLock, SFSBaseRecord& base,
				    CDLRecordDir iDir)
{
  CDLStatus ret;
  ExistingTransaction tran;
  sfs_status_t err;
  sfs_record_t* rec;
  sfs_ofd_t ofd;
  sfs_selector_t dir;
  sfs_lockMode_t lock;
  int retries = 0;

  ret = GetOfd(tran, ofd);
  if (ret == CDL_SUCCESS)
    {
      rec = base;
      dir = GetDirection(iDir);
      lock = GetMode(cLock);
      do
	err = sfs_Read(ofd, dir, lock, rec);
      while (RetryError(err) && (retries++ < NumRetries) && !tran.Aborted());
      ret = CheckError(10005, err, tran, ErrorLevels::InfoMsg);
      if (ret == CDL_SUCCESS)
	{
	  CDLReport.VReport(ErrorLevels::InfoMsg, "TID %lu: Read one record from the range\n",
			    (tran_tid_t)tran);
	}
    }
  return ret;
}

/*
 * IReadUnique
 *    Read one and only one record via a key in a unique index. SFS will
 * conveniently generate an error message if the index named is not unique.
 */

CDLStatus SFSFileBase::IReadUnique(CDLReadLock cLock, SFSBaseKey& baseKey,
				   SFSBaseRecord& baseRec)
{
  CDLStatus ret;
  ExistingTransaction tran;
  sfs_status_t err;
  sfs_record_t* rec;
  sfs_key_t* key;
  sfs_ofd_t ofd;
  sfs_lockMode_t lock;
  int retries = 0;

  ret = GetOfd(tran, ofd);
  if (ret == CDL_SUCCESS)
    {
      rec = baseRec;
      key = baseKey;
      lock = GetMode(cLock);
      const char* keyName = baseKey.Name();
      CDLReport.VReport(ErrorLevels::InfoMsg, "ReadUnique for Key: %s\n", keyName);
      if (IsBatching(tran))
	{
	  SFSBatchRead* bRead = new SFSBatchRead(ofd, keyName, baseKey,
						 lock, baseRec);
	  if (!AddBatch(tran, bRead))
	    {
	      ret = CDL_BATCH_QUEUE_CORRUPTED;
	      ThrowError(10005, ret, tran, ErrorLevels::InfoMsg);
	    }
	}
      else
	{
	  do
	    err = sfs_ReadByKey(ofd, (char*)keyName, key, lock, rec);
	  while (RetryError(err) && (retries++ < NumRetries) && 
		 !tran.Aborted());
	  ret = CheckError(10005, err, tran, ErrorLevels::InfoMsg);
	  if (ret == CDL_SUCCESS)
	    {
	      CDLReport.VReport(ErrorLevels::InfoMsg, "TID %lu: Read one record by key\n",
				(tran_tid_t)tran);
	    }
	}
    }
  return ret;
}

/*
 * IDeleteCurrent
 *    Delete the specified record in the range. Since the user can specify
 * whether to delete the next, previous, or current record, and since SFS
 * will only delete the current one, we have to position the range manually.
 */

CDLStatus SFSFileBase::IDeleteCurrent(CDLRecordDir iDir)

{
  CDLStatus ret;
  ExistingTransaction tran;
  sfs_status_t err;
  sfs_ofd_t ofd;
  sfs_record_t bogus;
  int retries = 0;
  Boolean firstMoveSuccess = false;
  
  bogus.format = sfs_incompleteScattered;	// Empty record for moving CRP
  bogus.buffer = (sfs_pointer_t)0;
  bogus.bufferLength = bogus.dataLength = 0;
  bogus.numExternalFields = 0;
  bogus.externalFieldsArrayP = (sfs_externalField_t*)0;
  
  ret = GetOfd(tran, ofd);
  if ((ret == CDL_SUCCESS) && (iDir != CURRENT))
    {
      do
	{
	  if (iDir == NEXT)
	    err = sfs_Read(ofd, sfs_next, sfs_noLock, &bogus);
	  else if (iDir == PREVIOUS)
	    err = sfs_Read(ofd, sfs_previous, sfs_noLock, &bogus);
	}
      while (RetryError(err) && (retries++ < NumRetries) && !tran.Aborted());
      ret = CheckError(10006, err, tran, ErrorLevels::InfoMsg);
      if (ret == CDL_SUCCESS)
	{
	  CDLReport.VReport(ErrorLevels::TraceMsg, "TID %lu: Moved the cursor for delete\n",
			    (tran_tid_t)tran);
	  firstMoveSuccess = true;
	}
    }
  retries = 0;
  if (ret == CDL_SUCCESS)
    {
      do
	err = sfs_DeleteCurrent(ofd);
      while (RetryError(err) && (retries++ < NumRetries) && !tran.Aborted());
      ret = CheckError(10006, err, tran, ErrorLevels::InfoMsg);
      if (ret == CDL_SUCCESS)
	{
	  CDLReport.VReport(ErrorLevels::InfoMsg, 
			    "TID %lu: Deleted one record from the range\n",
			    (tran_tid_t)tran);
	}
    }
  retries = 0;
  if (((ret == CDL_SUCCESS) || (firstMoveSuccess)) && (iDir != CURRENT))
    {
      do
	{
	  if (iDir == NEXT)
	    err = sfs_Read(ofd, sfs_previous, sfs_noLock, &bogus);
	  else if (iDir == PREVIOUS)
	    err = sfs_Read(ofd, sfs_next, sfs_noLock, &bogus);
	}
      while (RetryError(err) && (retries++ < NumRetries) && !tran.Aborted());
      ret = CheckError(10006, err, tran, ErrorLevels::InfoMsg);
      if (ret == CDL_SUCCESS)
	{
	  CDLReport.VReport(ErrorLevels::TraceMsg, "TID %lu: Moved the cursor back\n", 
			    (tran_tid_t)tran);
	}
      else if (ret == CDL_END_OF_RANGE)
	{
	  ret = CDL_SUCCESS;
	}
    }
  return ret;
}

/*
 * IDeleteUnique
 *    Delete the one and only one record specified by the key. SFS will
 * complain if we're not using a unique index.
 */

CDLStatus SFSFileBase::IDeleteUnique(SFSBaseKey& base)

{
  CDLStatus ret;
  ExistingTransaction tran;
  sfs_status_t err;
  sfs_key_t* key;
  sfs_ofd_t ofd;
  const char* indName = base.Name();
  int retries = 0;

  ret = GetOfd(tran, ofd);
  if (ret == CDL_SUCCESS)
    {	
      key = base;
      if (IsBatching(tran))
	{
	  SFSBatchDelete* bDel = new SFSBatchDelete(ofd, indName, base);
	  if (!AddBatch(tran, bDel))
	    {
	      ret = CDL_BATCH_QUEUE_CORRUPTED;
	      ThrowError(10006, ret, tran, ErrorLevels::InfoMsg);
	    }
	}
      else
	{
	  do
	    err = sfs_DeleteByKey(ofd, (char*)indName, key);
	  while (RetryError(err) && (retries++ < NumRetries) && 
		 !tran.Aborted());
	  ret = CheckError(10006, err, tran, ErrorLevels::InfoMsg);
	  if (ret == CDL_SUCCESS)
	    {
	      CDLReport.VReport(ErrorLevels::InfoMsg, 
				"TID %lu: Deleted one record from the range\n",
				(tran_tid_t)tran);
	    }
	}
    }
  return ret;
}

/*
 * IUpdateCurrent
 *    Update the record at the specified position in the range. We must set the
 * CRP manually if the user wants to update the next or previous record. The
 * user will get an error message if he or she tries to change the primary
 * index -- we could change it for them, but then we'd screw up the current
 * selection.
 */

CDLStatus SFSFileBase::IUpdateCurrent(SFSBaseRecord& base, CDLRecordDir iDir)

{
  CDLStatus ret;
  ExistingTransaction tran;
  sfs_status_t err;
  sfs_record_t* rec;
  sfs_ofd_t ofd;
  sfs_record_t bogus;
  int retries = 0;
  Boolean firstMoveSuccess = false;

  bogus.format = sfs_incompleteScattered;	// Empty record for moving CRP
  bogus.buffer = (sfs_pointer_t)0;
  bogus.bufferLength = bogus.dataLength = 0;
  bogus.numExternalFields = 0;
  bogus.externalFieldsArrayP = (sfs_externalField_t*)0;

  ret = GetOfd(tran, ofd);
  if ((ret == CDL_SUCCESS) && (iDir != CURRENT))
    {
      do
	{
	  if (iDir == NEXT)
	    err = sfs_Read(ofd, sfs_next, sfs_noLock, &bogus);
	  else if (iDir == PREVIOUS)
	    err = sfs_Read(ofd, sfs_previous, sfs_noLock, &bogus);
	}
      while (RetryError(err) && (retries++ < NumRetries) && !tran.Aborted());
      ret = CheckError(10007, err, tran, ErrorLevels::InfoMsg);
      if (ret == CDL_SUCCESS)
	{
	  CDLReport.VReport(ErrorLevels::TraceMsg, "TID %lu: Moved the cursor for update\n",
			    (tran_tid_t)tran);
	  firstMoveSuccess = true;
	}
    }
  retries = 0;
  if (ret == CDL_SUCCESS)
    {
      rec = base;
      do
	err = sfs_UpdateCurrent(ofd, rec);
      while (RetryError(err) && (retries++ < NumRetries) && !tran.Aborted());
      ret = CheckError(10007, err, tran, ErrorLevels::InfoMsg);
      if (ret == CDL_SUCCESS)
	{
	  CDLReport.VReport(ErrorLevels::InfoMsg, 
			    "TID %lu: deleted one record from the range\n",
			    (tran_tid_t)tran);
	}
    }
  retries = 0;
  if (((ret == CDL_SUCCESS) || (firstMoveSuccess)) && (iDir != CURRENT))
    {
      do
	{
	  if (iDir == NEXT)
	    err = sfs_Read(ofd, sfs_previous, sfs_noLock, &bogus);
	  else if (iDir == PREVIOUS)
	    err = sfs_Read(ofd, sfs_next, sfs_noLock, &bogus);
	}
      while (RetryError(err) && (retries++ < NumRetries) && !tran.Aborted());
      ret = CheckError(10006, err, tran, ErrorLevels::InfoMsg);
      if (ret == CDL_SUCCESS)
	{
	  CDLReport.VReport(ErrorLevels::TraceMsg, "TID %lu: Moved the cursor back\n",
			    (tran_tid_t)tran);
	}
      else if (ret == CDL_END_OF_RANGE)
	{
	  ret = CDL_SUCCESS;
	}
    }
  return ret;
}

/*
 * IUpdateUnique
 *    Update one and only one record specified by a key. SFS causes this
 * function to fail on non-unique indices. It also fails if the user tries
 * to modify the primary index, although we may change this by automatically
 * deleting the old record and re-inserting the new one, but that may make
 * things too complicated and confusing.
 */

CDLStatus SFSFileBase::IUpdateUnique(SFSBaseKey& baseKey, SFSBaseRecord& baseRec)

{
  CDLStatus ret;
  ExistingTransaction tran;
  sfs_status_t err;
  sfs_record_t* rec;
  sfs_key_t* key;
  sfs_ofd_t ofd;
  int retries = 0;

  ret = GetOfd(tran, ofd);
  if (ret == CDL_SUCCESS)
    {
      rec = baseRec;
      key = baseKey;
      const char* indName = baseKey.Name();
      if (IsBatching(tran))
	{
	  SFSBatchUpdate* bUpdate = new SFSBatchUpdate(ofd, indName,
						       baseKey, baseRec);
	  if (!AddBatch(tran, bUpdate))
	    {
	      ret = CDL_BATCH_QUEUE_CORRUPTED;
	      ThrowError(10007, ret, tran, ErrorLevels::InfoMsg);
	    }
	}
      else
	{
	  do
	    err = sfs_UpdateByKey(ofd, (char*)indName, key, rec);
	  while (RetryError(err) && (retries++ < NumRetries) && 
		 !tran.Aborted());
	  ret = CheckError(10007, err, tran, ErrorLevels::InfoMsg);
	  if (ret == CDL_SUCCESS)
	    {
	      CDLReport.VReport(ErrorLevels::InfoMsg, 
				"TID %lu: updated one record by key\n",
				(tran_tid_t)tran);
	    }
	}
    }
  return ret;
}

/* 
 * IModifyField
 *
 *    Modify just one field in a record, getting an increment lock if 
 * possible.
 */

CDLStatus SFSFileBase::IModifyField(SFSBaseKey& baseKey, const String& fieldName,
				    sfs_pointer_t val, sfs_pointer_t out,
				    unsigned long length)
{
  CDLStatus ret;
  ExistingTransaction tran;
  sfs_status_t err;
  sfs_key_t* key;
  sfs_ofd_t ofd;
  int retries = 0;

  ret = GetOfd(tran, ofd);
  if (ret == CDL_SUCCESS)
    {
      key = baseKey;
      const char* indName = baseKey.Name();
      const char* fName = fieldName;
      if (IsBatching(tran))
	{
	  SFSBatchModifyField* bMod = new SFSBatchModifyField(ofd, indName, 
					      baseKey, fieldName, 
					      sfs_modifyAssign, 0,
					      val, out, length);

	  if (!AddBatch(tran, bMod))
	    {
	      ret = CDL_BATCH_QUEUE_CORRUPTED;
	      ThrowError(10008, ret, tran, ErrorLevels::InfoMsg);
	    }
	}
      else
	{
	  do
	    err = sfs_ModifyFieldByKey(ofd, (char*)indName, key, (char*)fName,
				       sfs_modifyAssign, 0, val, out);
	  while (RetryError(err) && (retries++ < NumRetries) && 
		 !tran.Aborted());
	  ret = CheckError(10008, err, tran, ErrorLevels::InfoMsg);
	  if (ret == CDL_SUCCESS)
	    {
	      CDLReport.VReport(ErrorLevels::InfoMsg, 
				"TID %lu: Modified one field in one record\n",
				(tran_tid_t)tran);
	    }
	}
    }
  return ret;
}

/* 
 * IIncrementField
 *    Used only on numeric fields, this increments (or decrements) the value
 * of a field by the specified amount. Otherwise, it's just like IModifyField.
 *    If the user wants to use an increment lock for this operation rather
 * than the standard write lock, he/she/it specifies it via the 
 * "incrementLock" parameter.
 */

CDLStatus SFSFileBase::IIncrementField(SFSBaseKey& baseKey, 
                                    const String& fieldName,
				    sfs_pointer_t val, sfs_pointer_t out,
				    unsigned long length,
                                    Boolean incrementLock)
{
  CDLStatus ret;
  ExistingTransaction tran;
  sfs_status_t err;
  sfs_key_t* key;
  sfs_ofd_t ofd;
  int retries = 0;

  ret = GetOfd(tran, ofd);
  if (ret == CDL_SUCCESS)
    {
      key = baseKey;
      const char* indName = baseKey.Name();
      const char* fName = fieldName;
      unsigned int incLock = 0;
      if (incrementLock)
	{
	  incLock = 1;
	}
      CDLReport.VReport(ErrorLevels::InfoMsg, 
			"IncrementField: %s for index: %s incLock: %u\n", 
			fName, indName, incLock);
      if (IsBatching(tran))
	{
	  SFSBatchModifyField* bMod = new SFSBatchModifyField(ofd, indName, 
					      baseKey, fieldName, 
					      sfs_modifyAdd, incLock,
					      val, out, length);

	  if (!AddBatch(tran, bMod))
	    {
	      ret = CDL_BATCH_QUEUE_CORRUPTED;
	      ThrowError(10017, ret, tran, ErrorLevels::InfoMsg);
	    }
	}
      else
	{
	  do
	    {
	      err = sfs_ModifyFieldByKey(ofd, (char*)indName, key, (char*)fName,
					 sfs_modifyAdd, incLock, val, out);
	    }
	  while (RetryError(err) && (retries++ < NumRetries) && 
		 !tran.Aborted());
	  ret = CheckError(10017, err, tran, ErrorLevels::InfoMsg);
	  if (ret == CDL_SUCCESS)
	    {
	      CDLReport.VReport(ErrorLevels::InfoMsg, 
				"TID %lu: Incremented one field in one record\n",
				(tran_tid_t)tran);
	    }
	}
    }
  return ret;
}

/*
 * IGetEsn
 *    Get the entry sequence number for this record. This will only work
 * for an entry-sequenced file.
 */

CDLStatus SFSFileBase::IGetEsn(sfs_esn_t& esn)

{
  sfs_ofd_t ofd;
  ExistingTransaction tran;
  sfs_status_t err;
  CDLStatus ret;
  int retries = 0;

  ret = GetOfd(tran, ofd);
  if (ret == CDL_SUCCESS)
    {
      do
	err = sfs_GetEsn(ofd, &esn);
      while (RetryError(err) && (retries++ < NumRetries) && !tran.Aborted());
      ret = CheckError(10016, err, tran, ErrorLevels::InfoMsg);
    }
  return ret;
}

/*
 * ISaveContext
 *    Save the context from the current OFD.
 */

CDLStatus SFSFileBase::ISaveContext(void** context)

{
  sfs_ofd_t ofd;
  ExistingTransaction tran;
  sfs_status_t err;
  CDLStatus ret;
  int retries = 0;

  ret = GetOfd(tran, ofd);
  if (ret == CDL_SUCCESS)
    {
      do
	err = sfs_SaveContext(ofd, (sfs_context_t**)context);
      while (RetryError(err) && (retries++ < NumRetries) && !tran.Aborted());
      ret = CheckError(10024, err, tran, ErrorLevels::InfoMsg);
    }
  if (ret == CDL_SUCCESS)
    {
      CDLReport.VReport(ErrorLevels::InfoMsg,
			"TID %lu: Saved context from OFD %lu\n",
			(tran_tid_t)tran, ofd);
    }
  return ret;
}

/*
 * IRestoreContext
 *    Save the context from the current OFD.
 */

CDLStatus SFSFileBase::IRestoreContext(void* context)

{
  sfs_ofd_t ofd;
  ExistingTransaction tran;
  sfs_status_t err;
  CDLStatus ret;
  int retries = 0;

  ret = GetOfd(tran, ofd);
  if (ret == CDL_SUCCESS)
    {
      do
	err = sfs_RestoreContext(ofd, (sfs_context_t*)context);
      while (RetryError(err) && (retries++ < NumRetries) && !tran.Aborted());
      ret = CheckError(10025, err, tran, ErrorLevels::InfoMsg);
    }
  if (ret == CDL_SUCCESS)
    {
      CDLReport.VReport(ErrorLevels::InfoMsg,
			"TID %lu: Restored context into OFD %lu\n",
			(tran_tid_t)tran, ofd);
    }
  return ret;
}

/*
 * IActivateIndex
 *    Make the specified index active and update it according to the 
 * contents of the main file.
 */

CDLStatus SFSFileBase::IActivateIndex(SFSBaseSecondary& indexBase,
				      const String& volume,
				      SFSBaseKey* nullKey)
{
  sfs_ofd_t ofd;
  ExistingTransaction tran;
  sfs_status_t err;
  CDLStatus ret;
  int retries = 0;

  ret = GetOfd(tran, ofd);
  if (ret == CDL_SUCCESS)
    {
      sfs_secondaryIndexSpec_t* indexSpec = indexBase;
      const char* indName = indexBase.Name();
      const char* volName = volume;
      indexSpec->storageSpec.volumeName = (char*)volName;
      indexSpec->storageSpec.allocated = 1;
      indexSpec->storageSpec.utilized = 0;
      if (nullKey != (SFSBaseKey*)0)
	{
	  indexSpec->excludedKeyP = *nullKey;
	}
      CDLReport.VReport(ErrorLevels::InfoMsg,
			"AddSecondaryIndex: %s to Volume: %s\n", 
			indName, volName);
      do
	{
	  err = sfs_AddSecondaryIndex(ofd, (char*)indName, indexSpec);
	  if (err == SFS_INDEX_NAME_EXISTS)		// So just rebuild it
	    err = sfs_RebuildSecondaryIndex(ofd, (char*)indName);
	}
      while (RetryError(err) && (retries++ < NumRetries) && !tran.Aborted());
      if (nullKey != (SFSBaseKey*)0)
	{
	  indexSpec->excludedKeyP = (sfs_key_t*)0;
	}
      ret = CheckError(10009, err, tran, ErrorLevels::InfoMsg);
      if (ret == CDL_SUCCESS)
	{
	  CDLReport.VReport(ErrorLevels::InfoMsg, "TID %lu: Enabled secondary index %s\n",
			    (tran_tid_t)tran, indName);
	}
    }
  return ret;
}

/*
 * IDeactivateIndex
 *    Deactivate the specified index. It will no longer be updated when the
 * database changes.
 */

CDLStatus SFSFileBase::IDeactivateIndex(SFSBaseSecondary& indexBase)

{
  sfs_ofd_t ofd;
  ExistingTransaction tran;
  sfs_status_t err;
  CDLStatus ret;
  int retries = 0;

  ret = GetOfd(tran, ofd);
  if (ret == CDL_SUCCESS)
    {
      const char* indName = indexBase.Name();
      CDLReport.VReport(ErrorLevels::InfoMsg, "Deactivating Index: %s\n", indName);
      do
	err = sfs_DeactivateSecondaryIndex(ofd, (char*)indName, false);
      while (RetryError(err) && (retries++ < NumRetries) && !tran.Aborted());
      ret = CheckError(10010, err, tran, ErrorLevels::InfoMsg);
      if (ret == CDL_SUCCESS)
	{
	  CDLReport.VReport(ErrorLevels::InfoMsg, "TID %lu: Disabled secondary index %s\n",
			    (tran_tid_t)tran, indName);
	}
    }
  return ret;
}

/*
 * IDestroyIndex
 *    Deactivate the specified index and deallocate all its storage as well.
 */

CDLStatus SFSFileBase::IDestroyIndex(SFSBaseSecondary& indexBase)

{
  sfs_ofd_t ofd;
  ExistingTransaction tran;
  sfs_status_t err;
  CDLStatus ret;
  int retries = 0;

  ret = GetOfd(tran, ofd);
  if (ret == CDL_SUCCESS)
    {
      const char* indName = indexBase.Name();
      CDLReport.VReport(ErrorLevels::InfoMsg, "Destroying Index: %s\n", indName);
      do
	err = sfs_DeactivateSecondaryIndex(ofd, (char*)indName, true);
      while (RetryError(err) && (retries++ < NumRetries) && !tran.Aborted());
      ret = CheckError(10011, err, tran, ErrorLevels::InfoMsg);
      if (ret == CDL_SUCCESS)
	{
	  CDLReport.VReport(ErrorLevels::InfoMsg, 
			    "TID %lu: Destroyed secondary index %s\n",
			    (tran_tid_t)tran, indName);
	}
    }
  return ret;
}

/*
 * IDestroyFile
 *    Completely delete this file. Since the file will be closed by SFS
 * after it's destroyed, we must act as if we closed all the OFDs. So,
 * we need code like the code in ICloseFile. The closing code is simpler
 * here because the file is opened in exclusive mode and there's no list
 * of OFDs to worry about.
 */

CDLStatus SFSFileBase::IDestroyFile()

{
  ExistingTransaction tran;
  sfs_status_t err;
  CDLStatus status;
  int retries = 0;
  int waitCount = 0;
  timespec interval;
  interval.tv_sec = 5;
  interval.tv_nsec = 0;

  if ((status = CheckOpen(true, tran)) != CDL_SUCCESS)
    return status;
  ofdLatch.Lock();
  opened = false;
  while ((waitCount < 10) && (activeTrans.size() > 0))
    {
      CDLReport.Report(ErrorLevels::InfoMsg, "Waiting for transactions to finish in IDestroyFile\n");
      waitCount++;
      pthread_delay_np(&interval);	// Wait for trans to end
    }
  if (waitCount >= 10)
    CDLReport.Report(ErrorLevels::InfoMsg, "Giving up and closing OFDs\n");		// Giving up!

  do
    err = sfs_DestroyFile(masterOfd);
  while (RetryError(err) && (retries++ < NumRetries) && !tran.Aborted());
  ofdLatch.Unlock();
  status = CheckError(10012, err, tran, ErrorLevels::InfoMsg);
  if (status == CDL_SUCCESS)
    {
      CDLReport.VReport(ErrorLevels::InfoMsg, "TID %lu: Destroyed the file\n",
			(tran_tid_t)tran);
    }

  return status;
}

/*
 * IEmptyFile
 *    Remove all the records in this file.
 */

CDLStatus SFSFileBase::IEmptyFile()

{
  sfs_ofd_t ofd;
  ExistingTransaction tran;
  sfs_status_t err;
  CDLStatus ret;
  int retries = 0;

  ret = GetOfd(tran, ofd);
  if (ret == CDL_SUCCESS)
    {
      do
	err = sfs_EmptyFile(ofd, true);		// Deallocate disk space too
      while (RetryError(err) && (retries++ < NumRetries) && !tran.Aborted());
      ret = CheckError(10013, err, tran, ErrorLevels::InfoMsg);
      if (ret == CDL_SUCCESS)
	{
	  CDLReport.VReport(ErrorLevels::InfoMsg, 
			    "TID %lu: Emptied the file of all records\n",
			    (tran_tid_t)tran);
	}
    }
  return ret;
}

/*
 * IStartBatch
 *    Begin batching. Some future operations (particularly the "ByKey" 
 * operations) will be placed on a queue in the context of the current
 * transaction and executed later all at once.
 */

CDLStatus SFSFileBase::IStartBatch()

{
  ExistingTransaction tran;
  SFSBatchQueue testKey(tran);
  CDLStatus retVal;

  batchLatch->Lock();
  const SFSBatchQueue* foundBatch = batchQueues->contains(testKey);
  if (foundBatch == (SFSBatchQueue*)0)
    {
      CDLReport.VReport(ErrorLevels::TraceMsg, "TID %lu: Creating batch queue\n", (tran_tid_t)tran);
      testKey.batching = 1;
      batchQueues->insert(testKey);
      retVal = CDL_SUCCESS;
    }
  else
    {
      SFSBatchQueue* nonConstBatch = (SFSBatchQueue*)foundBatch;
      nonConstBatch->batching++;		// Casting here is OK
      retVal = CDL_SUCCESS;
    }
  batchLatch->Unlock();
  return retVal;
}

/*
 * IEndBatch
 *    Decrement the batch count. If it has reached zero, build the batching
 * data structure for SFS and execute all the batched operations.
 */

CDLStatus SFSFileBase::IEndBatch()

{
  ExistingTransaction tran;
  SFSBatchQueue testKey(tran);
  CDLStatus ret = CDL_SUCCESS;
  unsigned int failPosition = 0;
  sfs_batchRequest_t* requests;
  sfs_batchRequest_t* tmpRequest;
  SFSBatchBase* tmpBase;
  int numRequests;
  sfs_status_t err;
  int inc;
  int retries = 0;
  
  batchLatch->Lock();
  const SFSBatchQueue* foundBatch = batchQueues->contains(testKey);
  if (foundBatch == (SFSBatchQueue*)0)
    ret = CDL_BATCH_QUEUE_CORRUPTED;
  else
    {
      SFSBatchQueue* nonConstBatch = (SFSBatchQueue*)foundBatch;
      nonConstBatch->batching--;
      if (nonConstBatch->batching == 0)
	{
	  numRequests = nonConstBatch->queue.length();
	  requests = (sfs_batchRequest_t*)malloc(sizeof(sfs_batchRequest_t) *
						 numRequests);
	  List_of_piter<SFSBatchBase> iter(nonConstBatch->queue);
	  for (inc = 0; inc < numRequests; inc++)
	    {
	      iter.next(tmpBase);
	      tmpRequest = *tmpBase;
	      requests[inc] = *tmpRequest;
	    }
	  batchLatch->Unlock();
	  CDLReport.VReport(ErrorLevels::InfoMsg,
			    "TID %lu: Executing %i items from batch queue\n",
			    (tran_tid_t)tran, 
			    numRequests);
	  do
	    err = sfs_Batch(numRequests, requests, &failPosition);
	  while (RetryError(err) && (retries++ < NumRetries) && 
		 !tran.Aborted());
	  batchLatch->Lock();
	  /* Make sure the queue wasn't deleted by transaction abort */
	  if (batchQueues->contains(testKey) != (SFSBatchQueue*)0)
	    {
	      iter.reset();
	      while (iter.remove_next(tmpBase))
		delete tmpBase;
	    }
	  batchLatch->Unlock();
	  ret = CheckError(10014, err, tran, ErrorLevels::InfoMsg);
	  if (ret != CDL_SUCCESS)
	    CDLReport.VReportMsg(ErrorLevels::InfoMsg, 10018, failPosition);
	  batchLatch->Lock();
	  free((char*)requests);
	}
    }
  batchLatch->Unlock();
  return ret;
}
