/*
 * CDLField:
 *
 *    This file holds stuff related to a field in the main CDL file. 
 * CDLRecordField, a subclass, refers to a field that has a record in it.
 * This class will look up a pointer to a record based on its name,
 * making parsing a little easier.
 *
 * cdlfield.C,v 1.8 1994/02/01 22:32:46 gjb Exp
 */

    /*********************************************************************
    Copyright (C) 1993 by Citibank, N.A., all rights reserved.

    Permission to use, copy, modify and distribute this software and its 
    documentation for any purpose and without fee, except possible 
    redistribution costs, is hereby granted, provided that (1) the above 
    copyright notice appears in all copies and (2) both that copyright-
    notice and this permission-notice appear in supporting documentation and 
    (3) the name of Citibank, N.A. is not used in advertising or publicity 
    pertaining to distribution of the software without specific and written 
    prior permission of Citibank, N.A.

    The Software is provided "as is."  Citibank, N.A. disclaims all 
    warranties, express or implied, including but not limited to warranties 
    of suitability, merchantability and fitness for any purpose.  In no 
    event shall Citibank, N.A. be responsible for any loss of use, data or 
    profit or for any other special, indirect, incidental or consequential 
    damages, neither in an action of contract nor in an action of tort, 
    arising out of or in connection with the use or performance of this 
    software.
    *********************************************************************/


#include <cdlfield.H>
#include <cdlmainparser.H>
#include <mainparsedefs.H>
#include <parselist.H>

#ifdef _AIX
#include <List.c>

#pragma define (Const_list_of_piter<CDLField>)
#pragma define (List_of_p<CDLField>)
#pragma define (List_of_piter<CDLField>)
#endif

CDLField::CDLField() : NamedObject()

{
  mType = NONE;
  length = 0;
  arraySize = 1;
}

CDLField::CDLField(FieldType type, ModifierType mod, 
		   const String& fieldName, const int fieldLen,
		   const int aSize)
    : NamedObject(fieldName)
{
  fType = type;
  mType = mod;
  length = fieldLen;
  arraySize = aSize;
  CDLReport.VReport(ErrorLevels::TraceMsg,
		    "Creating CDLField: %s - Len: %d\n", 
		    (const char*)fieldName, fieldLen);
}

CDLField::CDLField(FieldType type, const String& fieldName) 
    : NamedObject(fieldName)
{
  CDLReport.VReport(ErrorLevels::TraceMsg, "Creating CDLField: %s\n", 
		    (const char*)fieldName);
  fType = type;
  mType = NONE;
  length = 0;
  arraySize = 1;
}

CDLField::CDLField(FieldType type, const String& fieldName, 
		   const int fieldLen, 
		   const int aSize) : NamedObject(fieldName)
{
  CDLReport.VReport(ErrorLevels::TraceMsg, "Creating CDLField: %s\n", 
		    (const char*)fieldName);
  fType = type;
  mType = NONE;
  length = fieldLen;
  arraySize = aSize;
}

CDLField::CDLField(FieldType type, const String& fieldName,
		   const String& rtName, 
		   const int aSize) : NamedObject(fieldName), 
		                      realTypeName(rtName)

{
  CDLReport.VReport(ErrorLevels::TraceMsg, 
		    "Creating CDLField: %s of type %s\n",
		    (const char*)fieldName, (const char*)realTypeName);
  fType = type;
  mType = NONE;
  length = 0;
  arraySize = aSize;
}

CDLField::CDLField(const CDLField& field) : NamedObject()

{
  this->operator = (field);
}

CDLField::~CDLField()

{
}

CDLField& CDLField::operator = (const CDLField& field)

{
  NamedObject::operator = (field);
  fType = field.fType;
  mType = field.mType;
  name = field.name;
  length = field.length;
  arraySize = field.arraySize;
  realTypeName = field.realTypeName;
  return *this;
}

CDLRecord* CDLField::Record()

{
  return (CDLRecord*)0;
}

CDLRecordField::CDLRecordField() 
    : CDLField()

{
  theRecord = (CDLRecord*)0;
}

CDLRecordField::CDLRecordField(const String& fieldName,
			       const String& recordName,
			       const int aSize)
    : CDLField(RECORD, fieldName, recordName, aSize)
{
  theRecord = (CDLRecord*)0;
   /* We don't find the pointer from the name until someone asks! */
}

CDLRecordField::CDLRecordField(const CDLRecordField& field) 
   : CDLField(field)

{
  theRecord = field.theRecord;
}

CDLRecordField& CDLRecordField::operator = (const CDLRecordField& field)

{
  this->CDLField::operator = (field);
  theRecord = field.theRecord;
  return *this;
}

CDLRecordField::~CDLRecordField()

{
}

CDLRecord* CDLRecordField::Record()

{
  String errStr;
  if (theRecord == (CDLRecord*)0)
    {
      theRecord = CDLMainGetParser()->Find(realTypeName);
      if (theRecord == (CDLRecord*)0)
	{
	  CDLReport.VGetMsg(1100, errStr, (const char*)realTypeName);
	  CDLPError(errStr);
	}
    }
  return theRecord;
}
