/*
 * CDLField:
 *
 *    This object represents a single field in the primary .cdl file. There
 * are four fields: the data type, the field name, an optional field length
 * (for array fields) and an optional data modifier (for signed or unsigned
 * integer types). This object stores those values.
 *
 * cdlfield.H,v 1.8 1994/04/07 19:25:34 gjb Exp 
 */

    /*********************************************************************
    Copyright (C) 1993 by Citibank, N.A., all rights reserved.

    Permission to use, copy, modify and distribute this software and its 
    documentation for any purpose and without fee, except possible 
    redistribution costs, is hereby granted, provided that (1) the above 
    copyright notice appears in all copies and (2) both that copyright-
    notice and this permission-notice appear in supporting documentation and 
    (3) the name of Citibank, N.A. is not used in advertising or publicity 
    pertaining to distribution of the software without specific and written 
    prior permission of Citibank, N.A.

    The Software is provided "as is."  Citibank, N.A. disclaims all 
    warranties, express or implied, including but not limited to warranties 
    of suitability, merchantability and fitness for any purpose.  In no 
    event shall Citibank, N.A. be responsible for any loss of use, data or 
    profit or for any other special, indirect, incidental or consequential 
    damages, neither in an action of contract nor in an action of tort, 
    arising out of or in connection with the use or performance of this 
    software.
    *********************************************************************/


#ifndef CDL_FIELD_H
#define CDL_FIELD_H

extern "C"
{
#include <pthread.h>
}
#include <cdl/cdldefs.H>
#include <String.h>
#include <iostream.h>
#include <namedobject.H>

class CDLRecord;

class CDLField : public NamedObject

{
public:
  enum FieldType
    { REAL, RECORD, BYTES, VARBYTES, STRING, 
      HYPER, LONG, SHORT, CHAR, ENUM, CHARRAY };
  enum ModifierType
    { SIGNED, UNSIGNED, NONE };
  CDLField();
  CDLField(FieldType, ModifierType, 
	   const String& fieldName, const int fieldLen,
	   const int aSize = 1);
  CDLField(FieldType, const String& fieldName);
  CDLField(FieldType, const String& fieldName, const int fieldLen,
	   const int aSize = 1);
  CDLField(FieldType, const String& fieldName, const String& realTypeName,
	   const int aSize = 1);
  CDLField(const CDLField&);
  virtual ~CDLField();
  CDLField& operator = (const CDLField&);
  FieldType Type() const;
  ModifierType Modifier() const;
  int Length() const;
  int ArraySize() const;
  const String& RealTypeName() const;
  virtual CDLRecord* Record();
protected:
  FieldType fType;			// The basic type of this field
  ModifierType mType;			// Signed or unsigned for some types
  int length;				// For a variable-length type
  int arraySize;			// For an array of types
  String realTypeName;			// For enums & records, real type name
};

class CDLRecordField : public CDLField

{
public:
  CDLRecordField();
  CDLRecordField(const String& fieldName, const String& recordName,
	         const int aSize);
  CDLRecordField(const CDLRecordField&);
  virtual ~CDLRecordField();
  CDLRecordField& operator = (const CDLRecordField&);
  virtual CDLRecord* Record();
private:
  CDLRecord* theRecord;
};

inline Boolean IsInteger(CDLField::FieldType type)

{ return ((type == CDLField::HYPER) || (type == CDLField::LONG) ||
	  (type == CDLField::SHORT) || (type == CDLField::CHAR)); }

inline Boolean IsReal(CDLField::FieldType type)
     
{ return (type == CDLField::REAL); }

inline Boolean IsPointer(CDLField::FieldType type)

{ return ((type == CDLField::BYTES) || (type == CDLField::VARBYTES) ||
	  (type == CDLField::STRING) || (type == CDLField::CHARRAY)); }

inline Boolean IsRecord(CDLField::FieldType type)
     
{ return (type == CDLField::RECORD); }

inline Boolean IsAtomic(CDLField::FieldType type)

{ return ((type == CDLField::HYPER) || (type == CDLField::LONG) ||
	  (type == CDLField::SHORT) || (type == CDLField::CHAR) ||
	  (type == CDLField::REAL) || (type == CDLField::ENUM)); }

inline Boolean IsByteArray(CDLField::FieldType type)
     
{ return ((type == CDLField::BYTES) || (type == CDLField::VARBYTES)); }

inline Boolean IsString(CDLField::FieldType type)
     
{ return (type == CDLField::STRING); }

inline CDLField::FieldType CDLField::Type() const

{ return fType; }

inline CDLField::ModifierType CDLField::Modifier() const

{ return mType; }

inline int CDLField::Length() const

{ return length; }

inline int CDLField::ArraySize() const

{ return arraySize; }

inline const String& CDLField::RealTypeName() const

{ return realTypeName; }
  
#endif
