/*
 * CDLGenerator:
 *
 *    This is an abstract class for generators.
 *
 * cdlgenerator.C,v 1.11 1994/04/07 19:25:39 gjb Exp
 */

    /*********************************************************************
    Copyright (C) 1993-1994 by Citibank, N.A., all rights reserved.

    Permission to use, copy, modify and distribute this software and its 
    documentation for any purpose and without fee, except possible 
    redistribution costs, is hereby granted, provided that (1) the above 
    copyright notice appears in all copies and (2) both that copyright-
    notice and this permission-notice appear in supporting documentation and 
    (3) the name of Citibank, N.A. is not used in advertising or publicity 
    pertaining to distribution of the software without specific and written 
    prior permission of Citibank, N.A.

    The Software is provided "as is."  Citibank, N.A. disclaims all 
    warranties, express or implied, including but not limited to warranties 
    of suitability, merchantability and fitness for any purpose.  In no 
    event shall Citibank, N.A. be responsible for any loss of use, data or 
    profit or for any other special, indirect, incidental or consequential 
    damages, neither in an action of contract nor in an action of tort, 
    arising out of or in connection with the use or performance of this 
    software.
    *********************************************************************/


#include <cdlgenerator.H>
#include <arglist.H>
#include <Path.h>

#ifdef _AIX
#include <List.c>

#pragma define (Const_list_of_piter<CDLGenerator>)
#pragma define (List_of_p<CDLGenerator>)
#pragma define (List_of_piter<CDLGenerator>)
#pragma define (Const_list_of_piter<CDLDataGenerator>)
#pragma define (List_of_p<CDLDataGenerator>)
#pragma define (List_of_piter<CDLDataGenerator>)
#endif

/*
 * FieldNode:
 *
 *    This is an object used internally and is added to a list built by 
 * BuildFieldList, which recursively (actually, non-recursively) descends the
 * nested records in a CDL record.
 */

FieldNode::FieldNode() : NamedObject()

{
}

FieldNode::FieldNode(String n, CDLField* p) : NamedObject(n)

{
  theField = *p;
}

FieldNode::FieldNode(const FieldNode& node)

{
  this->operator = (node);
}

FieldNode::~FieldNode()

{
}

FieldNode& FieldNode::operator = (const FieldNode& node)

{
  name = node.name;
  theField = node.theField;
  varName = node.varName;
  lengthFieldName = node.lengthFieldName;
  return *this;
}

int FieldNode::operator == (const FieldNode& node) const

{
  return varName == node.varName;
}

void FieldNode::VariableName(const String& v)

{
  varName = v;
}

void FieldNode::LengthFieldName(const String& lfName)

{
  lengthFieldName = lfName;
}

/*
 * CDLGenericGenerator 
 */

CDLGenericGenerator::CDLGenericGenerator(const ArgList& args) 
   : origArgs(args)

{
  indentLevel = 0;
}

CDLGenericGenerator::~CDLGenericGenerator()

{
}

/*
 * MaskeFileName
 *    Given a name and a suffix, take the original suffix off the name
 * and return the name with the new suffix. If "keepDir" is set, also
 * remove the directory name from the head of the file. In addition,
 * shorten the base name (but not the suffix) if the name is longer
 * than maxLen. If maxLen is 0, ignore it.
 */

void CDLGenericGenerator::MakeFileName(const String& inName, 
				       String& outName,
				       const String& suffix,
				       Boolean keepDir,
				       unsigned int maxLen)
{
  int dotIndex = inName.strrchr('.');
  String origSuffix = inName(dotIndex);
  Path inPath(inName);
  inPath.expand_tilde();
  String baseName = inPath.basename(origSuffix);
  if (keepDir)
    {
      Path dirPath = inPath.dirname();
      outName = (String)dirPath;
      outName += "/";
    }
  else
    outName = "";
  if (maxLen != 0)
    {
      if ((baseName.length() + suffix.length()) > maxLen)
	baseName.shrink(maxLen - suffix.length());
    }
  outName += baseName;
  outName += suffix;
}

/*
 * Indent
 *    Increase the indent level for this object. Other generators can indent
 * their code by outputting the "indent" string at the start of every line.
 */

void CDLGenericGenerator::Indent()

{
  indentLevel += IndentStep;
  for (int inc = 0; inc < IndentStep; inc++)
    indent += " ";
}

/* 
 * Outdent
 *    Decrease the indent level.
 */

void CDLGenericGenerator::Outdent()

{
  indentLevel -= IndentStep;
  if (indentLevel < 0)
    indentLevel = 0;
  indent = "";
  for (int inc = 0; inc < indentLevel; inc++)
    indent += " ";
}

/*
 * OpenBrace
 *    Indent to the next level and output a brace.
 */

void CDLGenericGenerator::OpenBrace(ostream& out)

{
  Indent();
  out << indent << "{" << endl;
  Indent();
}

/*
 * CloseBrace
 *    Pop back to the old indentation level and output a close brace.
 */

void CDLGenericGenerator::CloseBrace(ostream& out)

{
  Outdent();
  out << indent << "}" << endl;
  Outdent();
}

/*
 * GetSize
 *    Return the maximum size of an element of this field's type in bytes.
 * Subclasses may override this. Note that trying to get the size of 
 * field of type "record" is an error.
 */

int CDLGenericGenerator::GetSize(CDLField& field) const

{
    int retVal;
  switch (field.Type())
    {
    case CDLField::REAL:
      retVal = 8;
      break;
    case CDLField::HYPER:
      retVal = 8;
      break;
    case CDLField::LONG:
      retVal = 4;
      break;
    case CDLField::SHORT:
    case CDLField::CHAR:
      retVal = 2;
      break;
    case CDLField::BYTES:
    case CDLField::VARBYTES:
      retVal = field.Length();
      break;
    case CDLField::STRING:
      retVal = field.Length() + 1;
      break;
    case CDLField::CHARRAY:
      retVal = field.Length();
      break;
    case CDLField::ENUM:
      retVal = 4;
      break;
    case CDLField::RECORD:
      retVal = -1;
      break;
    }
  return retVal;
}

/*
 * GetMaxSize
 *    For every node in "nodes", sum the results from "GetSize" and return
 * the result.
 */

int CDLGenericGenerator::GetMaxSize(List<FieldNode>& nodes) const

{
  Listiter<FieldNode> iter(nodes);
  int retVal = 0;
  FieldNode curNode;
  
  while (iter.next(curNode))
    retVal += GetSize(*(curNode.Field()));
  return retVal;
}

/*
 * GenWholeFieldName
 *    This is used recursively to build field names. It appends "separator",
 * then "fieldName" to the end of "baseName" to form one big string. If
 * the resulting string is bigger than "maxLen", it uses the hash value
 * generated by the USL String class, plus a few other neat characters
 * in the hopes that the resulting name will be both unique and semi-readable.
 * If "arrayElement" is 0 or greater, it also appends [arrayElement] to the
 * result. The result is placed in "out".
 */

Boolean CDLGenericGenerator::GenWholeFieldName(char separator, 
					       const String& fieldName,
					       const String& baseName,
					       long maxLen,
					       String& out,
					       int arrayElement) const
{
  if (baseName.length() > 0)
    {
      out = baseName;
      if (baseName[baseName.length() - 1] != separator)
	{
	  out += separator;
	}
    }
  else
    {
      out = "";
    }
  out += fieldName;
  if (arrayElement >= 0)
    {
      out += "[";
      out += int_to_str(arrayElement);
      out += "]";
    }
  if ((maxLen > 0) && (out.length() > maxLen))
    {
      String oldOut = out;
      out = oldOut.chunk(0, 5);			// New val: first 5 chars,
      out += int_to_str(oldOut.hashval());	//  the hash value,
      out += oldOut.chunk(oldOut.length() - 5, 5);	//  then last 5 chars
    }
//  CDLReport.VReport(ErrorLevels::TraceMsg,
//		    "field name: base = %s field = %s out = %s\n",
//		    (const char*)baseName, (const char*)fieldName,
//		    (const char*)out);
  return true;
}

/*****
 * This is the old version of GenWholeFieldName
 *

Boolean CDLGenericGenerator::GenWholeFieldName(char separator, 
					       const String& fieldName,
					       const String& baseName,
					       long maxLen,
					       String& out,
					       int arrayElement) const
{
  unsigned int subtract = 0;
  unsigned int tmpLen;
  String tmpStr;
  const int maxLoops = 4;
  int loopCount = 0;

  do
    {
      tmpLen = baseName.length();
      if (tmpLen <= 2)
	out = baseName;
      else
	{
	  out = baseName.chunk(0, tmpLen - subtract);
	  if (out.length() < 2)
	    out = baseName.chunk(0, 2);
	}
      tmpLen = fieldName.length();
      if (tmpLen <= 2)
	tmpStr = fieldName;
      else
	{
	  tmpStr = fieldName.chunk(0, tmpLen - subtract);
	  if (tmpStr.length() < 2)
	    tmpStr = fieldName.chunk(0, 2);
	}
      out += tmpStr;
      if (arrayElement >= 0)
	{
	  out += "[";
	  out += int_to_str(arrayElement);
	  out += "]";
	}
      if (out.length() > maxLen)
	{
	  if (subtract == 0)
	    subtract = (unsigned int)(((out.length() - maxLen) / 2) + 1);
	  else
	    subtract++;
	}
      loopCount++;
    } while ((maxLen > 0) && 
	     (out.length() > maxLen) && 
	     (loopCount <= maxLoops));
  if (loopCount > maxLoops)
    {
      if (baseName.length() > 0)
	{
	  tmpStr = baseName;
	  tmpStr += separator;
	  tmpStr += fieldName;
	}
      else
	tmpStr = fieldName;
      CDLReport.VReportMsg(ErrorLevels::InfoMsg, 1151, (const char*)tmpStr);
      return false;
    }
  else
    {
    return true;
    }
}

****/

CDLGenerator::CDLGenerator(const ArgList& args) 
    : CDLGenericGenerator(args)  
{
}

CDLGenerator::~CDLGenerator()

{
}

CDLDataGenerator::CDLDataGenerator(const ArgList& args) 
    : CDLGenericGenerator(args)  
{
}

CDLDataGenerator::~CDLDataGenerator()

{
}

