/*
 * CDLGenerator:
 *
 *    This class can generate code based on a list of CDLRecord objects.
 * This is an abstract class -- subclasses implement all the important stuff.
 *
 * cdlgenerator.H,v 1.9 1994/04/07 19:25:42 gjb Exp
 */

    /*********************************************************************
    Copyright (C) 1993-1994 by Citibank, N.A., all rights reserved.

    Permission to use, copy, modify and distribute this software and its 
    documentation for any purpose and without fee, except possible 
    redistribution costs, is hereby granted, provided that (1) the above 
    copyright notice appears in all copies and (2) both that copyright-
    notice and this permission-notice appear in supporting documentation and 
    (3) the name of Citibank, N.A. is not used in advertising or publicity 
    pertaining to distribution of the software without specific and written 
    prior permission of Citibank, N.A.

    The Software is provided "as is."  Citibank, N.A. disclaims all 
    warranties, express or implied, including but not limited to warranties 
    of suitability, merchantability and fitness for any purpose.  In no 
    event shall Citibank, N.A. be responsible for any loss of use, data or 
    profit or for any other special, indirect, incidental or consequential 
    damages, neither in an action of contract nor in an action of tort, 
    arising out of or in connection with the use or performance of this 
    software.
    *********************************************************************/


#ifndef CDL_GENERATOR_H
#define CDL_GENERATOR_H

extern "C"
{
#include <pthread.h>
}
#include <iostream.h>
#include <fstream.h>
#include <arglist.H>
#include <cdlrecord.H>
#include <cdlfiledef.H>
#include <namedobject.H>
#include <cdl/cdldefs.H>
#include <Path.h>
#include <List.h>

/* Constants */
static const int IndentStep = 2;		// How much to indent by

/*
 * FieldNode
 *    A node representing a single field in a list of all the fields.
 * "Name" (inherited) is the name of the field for SFS or RQS. 
 * "VariableName" is the name of the variable in the underlying IDL
 * data structure that represents it. "Field" is a pointer to the CDLField 
 * that contains the rest of the information on this field. "LengthField"
 * points to the field that handles the length of the data if this field is a 
 * byte array.
 */

class FieldNode : public NamedObject

{
public:
  FieldNode();
  FieldNode(String name, CDLField* ptr);
  FieldNode(const FieldNode&);
  virtual ~FieldNode();
  FieldNode& operator = (const FieldNode&);
  int operator == (const FieldNode&) const;
  CDLField* Field();
  void VariableName(const String& vName);
  const String& VariableName() const;
  void LengthFieldName(const String& lfName);
  const String& LengthFieldName() const;
private:
  CDLField theField;
  String lengthFieldName;
  String varName;
};

/*
 * CDLGenericGenerator
 *    This contains abstract member functions for the things that
 * all generators must support -- Start, End, and DeleteFiles. It
 * Also contains utility functions to keep track of the indent level,
 * and add a suffix to a name (with an optional maximum length).
 */

class CDLGenericGenerator

{
public:
  CDLGenericGenerator(const ArgList&);
  virtual ~CDLGenericGenerator();
  virtual Boolean Start(const String& fileName) = 0; // Start work on a file
  virtual Boolean End() = 0;				// End work
  virtual void DeleteFiles() = 0;			// Clean up output
  static void MakeFileName(const String& inName, 
			   String& outName,
			   const String& suffix,
			   Boolean keepDir = false,
			   unsigned int maxLen = 0);   // For file name appendices
protected:
  void Indent();				// Inrease indent lev.
  void Outdent();				// Decrease it
  void OpenBrace(ostream&);			// Indent and print "{"
  void CloseBrace(ostream&);			// Outdent and print "}"

  virtual int GetSize(CDLField& theField) const;   // Size of field in bytes
  virtual int GetMaxSize(List<FieldNode>& nodes) const;

  Boolean GenWholeFieldName(char separator, const String& fieldName,
			    const String& baseName, long maxLen, 
			    String& out,
			    int arrayElement = -1) const;

  ArgList origArgs;				// Arguments from main
  ofstream outFile;				// File for output
  int indentLevel;				// How much we're indenting
  String indent;				// A string for indentation
};

/*
 * CDLGenerator
 *    An abstract base class for generators that do not parse the .cfd
 * file. Besides the stuff from the CDLGenericGenerator class, it also
 * exports the abstract "Generate" routine.
 */

class CDLGenerator : public CDLGenericGenerator

{
public:
  CDLGenerator(const ArgList&);
  virtual ~CDLGenerator();
  virtual Boolean Generate(CDLRecord&) = 0;		// Process a record
};

/*
 * CDLDataGenerator
 *    Like CDLGenerator, but for generators that parse both the .cdl and
 * .cfd files. Here, the "Generate" routine takes both a CDLRecord and
 * a CDLFileDef.
 */

class CDLDataGenerator : public CDLGenericGenerator

{
public:
  CDLDataGenerator(const ArgList&);
  virtual ~CDLDataGenerator();
  virtual Boolean Generate(CDLRecord&, CDLFileDef&) = 0;
};

/* Inlines */

inline CDLField* FieldNode::Field()
     
{ return &theField; }

inline const String& FieldNode::VariableName() const

{ return varName; }

inline const String& FieldNode::LengthFieldName() const

{ return lengthFieldName; }

#endif

