/*
 * CDLIdlGenerator:
 *
 *    Read the records given us and spit out an IDL file (or files) that 
 * represents our record. 
 *
 * cdlidlgenerator.C,v 1.15 1994/04/07 19:25:45 gjb Exp 
 */

    /*********************************************************************
    Copyright (C) 1993-1994 by Citibank, N.A., all rights reserved.

    Permission to use, copy, modify and distribute this software and its 
    documentation for any purpose and without fee, except possible 
    redistribution costs, is hereby granted, provided that (1) the above 
    copyright notice appears in all copies and (2) both that copyright-
    notice and this permission-notice appear in supporting documentation and 
    (3) the name of Citibank, N.A. is not used in advertising or publicity 
    pertaining to distribution of the software without specific and written 
    prior permission of Citibank, N.A.

    The Software is provided "as is."  Citibank, N.A. disclaims all 
    warranties, express or implied, including but not limited to warranties 
    of suitability, merchantability and fitness for any purpose.  In no 
    event shall Citibank, N.A. be responsible for any loss of use, data or 
    profit or for any other special, indirect, incidental or consequential 
    damages, neither in an action of contract nor in an action of tort, 
    arising out of or in connection with the use or performance of this 
    software.
    *********************************************************************/


#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <unistd.h>
#include <objdce/defs.H>
#include <Search_path.h>
#include <cdlidlgenerator.H>
#include <cdlfield.H>
#include <cdlrecord.H>
#include <parselist.H>
#include <cdlsuffixes.H>
#include <cdlenuminfo.H>
#include <objdce/objtranparams.H>

/*
 * Handy constants
 *
 * HeapThreshold -- Above this many bytes, records have the [heap] attribute
 */

static const int HeapThreshold = 1024;

/*
 * CDLIdlGenerator
 *
 *    Not much to do here, except pass our arguments up the line.
 */

CDLIdlGenerator::CDLIdlGenerator(const ArgList& args,
				 Boolean nMode) : CDLGenerator(args)

{	
  newName = "";
  mainName = "";
  acfName = "";
  nidlMode = nMode;
}

CDLIdlGenerator::~CDLIdlGenerator()

{
}

/* 
 * Start
 *
 *    Start parsing by generating a name for the output file and
 * opening it. If that works, generate the header for our file, including
 * comments and the "interface" declaration.
 *    This function also opens the .acf file, which primarily exists so
 * we can put the "out_of_line" declaration there. Additionally, it 
 * calls GenEnums to spit out the list of "enum" declarations.
 */

Boolean CDLIdlGenerator::Start(const String& fileName)

{
  String interfaceName;
  String importFileName;

  MakeFileName(fileName, newName, CDLIdlFileSuffix);
  CDLReport.VReport(ErrorLevels::TraceMsg, 
		    "Made File name: %s\n", (const char*)newName);
  MakeFileName(fileName, interfaceName, CDLIdlInterfaceSuffix,
	       false, CDLIdlInterfaceMaxLen); 
  CDLReport.VReport(ErrorLevels::TraceMsg, "Made Interface name: %s\n", 
		    (const char*)interfaceName);
  MakeFileName(fileName, acfName, CDLAcfFileSuffix);
  CDLReport.VReport(ErrorLevels::TraceMsg, "Made file name: %s\n",
		    (const char*)acfName);
  mainName = newName;
  outFile.open(newName);
  if (outFile.fail())
    {
      CDLReport.VReportMsg(ErrorLevels::ErrorMsg, 1002, (const char*)newName);
      return false;
    }
  acfFile.open(acfName);
  if (acfFile.fail())
    {
      CDLReport.VReportMsg(ErrorLevels::ErrorMsg, 1002, (const char*)acfName);
      return false;
    }
  else
    {
      outFile << "/*" << endl;
      outFile << " * IDL file generated by Citi-DDL:" << endl;
      outFile << " * cdlidlgenerator.C,v 1.15 1994/04/07 19:25:45 gjb Exp" << endl;
      outFile << " */\n" << endl;

      outFile << "interface " << interfaceName << endl;
      outFile << "{" << endl;
      Indent();
      for (int inc = 0; inc < NumMainIncludeNames(); inc++)
	{
	  MakeFileName(GetMainIncludeName(inc), importFileName, 
		       CDLIdlFileSuffix);
	  CDLReport.VReport(ErrorLevels::TraceMsg,
			    "Made Import File Name: %s\n",
			    (const char*)importFileName);
	  outFile << indent << "import \"" << importFileName << "\";" << endl;
	}
      cout << endl;

      GenEnums();

      acfFile << "/*" << endl;
      acfFile << " * ACF file generated by Citi-DDL:" << endl;
      acfFile << " * cdlidlgenerator.C,v 1.15 1994/04/07 19:25:45 gjb Exp" << endl;
      acfFile << " */\n" << endl;

      acfFile << "interface " << interfaceName << endl;
      acfFile << "{" << endl;

      return true;
    }
}

/*
 * End
 *
 *    Finish up generating the file by writing the close brace and
 * closing the output file.
 */

Boolean CDLIdlGenerator::End()

{
  Outdent();
  outFile << "}" << endl;
  outFile << flush;
  outFile.close();
  acfFile << "}" << endl;
  acfFile << flush;
  acfFile.close();
  return true;
}

/*
 * DeleteFiles
 *
 *    Delete the output file. Called when parsing fails.
 */

void CDLIdlGenerator::DeleteFiles()

{
  outFile.close();
  unlink(mainName);
  acfFile.close();
  unlink(acfName);
}

/*
 * Generate
 *
 *    Given a CDLRecord object, generate its IDL type definition.
 * This is a pretty straightforward one-to-one mapping of fields in the
 * CDL record to fields in the IDL struct.
 *    This function also puts a "typedef" in the ACF file which specifies
 * the "out_of_line" attribute.
 */

Boolean CDLIdlGenerator::Generate(CDLRecord& rec)

{
  String typeName;
  String modName;
  CDLField* curField;
  CDLRecord* curRecord;
  String structName;
  List<FieldNode> recursiveFields;

  BuildFieldList(rec, recursiveFields);
  int maxRecSize = GetMaxSize(recursiveFields);
  
  structName = rec.Name() + CDLIdlRecordSuffix;
  CDLReport.Report(ErrorLevels::TraceMsg, "In CDLIdlGenerator:Generate\n");
  outFile << "/*" << endl;
  outFile << " * " << structName << ":" << endl;
  outFile << " * " << endl;
  outFile << " * " << rec.Documentation() << endl;
  outFile << " */\n" << endl;
  
  outFile << indent << "typedef ";
  if (nidlMode)
    {
      outFile << "[v1_struct] ";
    }
  outFile << "struct " << endl;
  outFile << indent << "{" << endl;
  Indent();
  
  for (int inc = 0; inc < rec.NumFields(); inc++)
    {
      curField = rec.Field(inc);
      switch (curField->Modifier())
	{
	case CDLField::UNSIGNED:
	  modName = "unsigned ";
	  break;
	case CDLField::SIGNED:
	case CDLField::NONE:
	  modName = "";
	  break;
	}
      switch (curField->Type())
	{
	case CDLField::REAL:
	  typeName = "double";
	  break;
	case CDLField::BYTES:
	case CDLField::VARBYTES:
	  typeName = "byte";
	  break;
	case CDLField::STRING:
	  typeName = "char";
	  break;
	case CDLField::RECORD:
	  curRecord = curField->Record();
	  if (curRecord != (CDLRecord*)0)
	    {
	      typeName = curRecord->Name();
	      typeName += CDLIdlRecordSuffix;
	    }
	  else
	    return false;
	  break;
	case CDLField::HYPER:
	  typeName = modName + "hyper int";
	  break;
	case CDLField::LONG:
	  typeName = modName + "long int";
	  break;
	case CDLField::SHORT:
	  typeName = modName + "short int";
	  break;
	case CDLField::CHAR:
	  typeName = modName + "small int";
	  break;
	case CDLField::CHARRAY:
	  typeName = "char";
	  break;
	case CDLField::ENUM:
	  typeName = curField->RealTypeName();
	  break;
	}

      if ((curField->Type() == CDLField::VARBYTES) || 
	  (curField->Type() == CDLField::BYTES))
	{
	  outFile << indent;
	  if (nidlMode && (curField->ArraySize() > 1))
	    {
	      outFile << "[v1_array] ";
	    }
	  outFile << "unsigned long " << curField->Name();
	  outFile << CDLByteArrayLengthSuffix;
	  if (curField->ArraySize() > 1)
	    outFile << "[" << curField->ArraySize() << "]";
	  outFile << ";" << endl;
	}


      outFile << indent;

      if (curField->ArraySize() == 1)
	{
	  switch(curField->Type())
	    {
	    case CDLField::VARBYTES:
	      outFile << "[length_is(";
	      outFile << curField->Name() << CDLByteArrayLengthSuffix << ")";
	      if (nidlMode)
		{
		  outFile << ", v1_array";
		}
	      outFile << "] " << typeName << " " << curField->Name() << "[";
	      outFile << curField->Length() << "];" << endl;
	      break;
	    case CDLField::BYTES:
	    case CDLField::CHARRAY:
	      if (nidlMode)
		{
		  outFile << "[v1_array] ";
		}
	      outFile << typeName << " " << curField->Name();
	      outFile << "[" << curField->Length() << "];" << endl;
	      break;
	    case CDLField::STRING:
	      if (nidlMode)
		{
		  outFile << "[v1_array, v1_string] ";
		}
	      outFile << typeName << " " << curField->Name();
	      outFile << "[" << curField->Length() + 1 << "];" << endl;
	      break;
	    default:
	      outFile << typeName << " ";
	      outFile << curField->Name() << ";" << endl;
	      break;
	    }
	}
      else
	{
	  if (nidlMode)
	    {
	      outFile << "[v1_array] ";
	    }
	  switch(curField->Type())
	    {
	    case CDLField::VARBYTES:
	    case CDLField::BYTES:
	    case CDLField::CHARRAY:
	      outFile << typeName << " " << curField->Name();
	      outFile << "[" << curField->ArraySize() << "][";
	      outFile << curField->Length() << "];" << endl;
	      break;
	    case CDLField::STRING:
	      outFile << "char " << curField->Name();
	      outFile << "[" << curField->ArraySize() << "] [";
	      outFile << curField->Length() + 1 << "];" << endl;
	      break;
	    default:
	      outFile << typeName << " ";
	      outFile << curField->Name() << "[" << curField->ArraySize();
	      outFile << "];" << endl;
	      break;
	    }
	}
    }
  Outdent();
  outFile << indent << "} " << structName << ";\n" << endl;

  // Put an entry in the ACF file as well
  if (maxRecSize > HeapThreshold)
    {
      acfFile << "  typedef [heap, out_of_line] ";
    }
  else
    {
      acfFile << "  typedef [out_of_line] ";
    }
  acfFile << structName << ";" << endl;

  // If IDL arrays were selected for this record, generate the proper
  // type definitions here.
  if (rec.MakeIdlArray() && !nidlMode)
    {
      outFile << indent << "typedef struct" << endl;
      Indent();
      outFile << indent << "{" << endl;
      Indent();
      outFile << indent << "long size;" << endl;
      outFile << indent << "[ptr, length_is(size), size_is(size)]" << endl;
      outFile << indent << structName << "* array;" << endl;
      Outdent();
      outFile << indent << "} " << structName;
      outFile << CDLIdlArraySuffix << ";" << endl;
      Outdent();

      acfFile << "  typedef [out_of_line] " << structName;
      acfFile << CDLIdlArraySuffix << ";" << endl;
    }

  return true;
}

/*
 * GenEnums
 *    Based on the map of defined "enum" types, spit out all the necessary
 * "enum" declarations. This is necessary because we want to use enums
 * in IDL declarations, so we must define them in the .idl file.
 */

Boolean CDLIdlGenerator::GenEnums()

{
  Mapiter<Symbol, CDLEnumInfo> enumIter(CDLEnums);
  Map_pair<Symbol, CDLEnumInfo>* curPair;
  String curCase;
  CDLEnumInfo curEnum;
  int inc;
  
  while ((curPair = enumIter.next()) != 0)
    {
      curEnum = curPair->value;
      if ((!curEnum.Imported()) && (curEnum.NumCases() > 0))
	{
	  outFile << indent << "typedef ";
	  if (nidlMode)
	    {
	      outFile << "[v1_enum] ";
	    }
	  outFile << "enum { ";
	  Indent();
	  for (inc = 0; inc < (curEnum.NumCases() - 1); inc++)
	    {
	      curCase = curEnum.Case(inc);
	      outFile << curCase << ", " << endl << indent;
	    }
	  curCase = curEnum.Case(curEnum.NumCases() - 1);
	  outFile << curCase << " } " << curEnum.Name() << ";" << endl;
	}
    }
  return true;
}

/*
 * FindIdl
 *    Using the Standard Components "Search_path" class, return the full
 * pathname of idl in "idlNameOut". However, if CDL_IDL is set, just
 * use that name.
 */

Boolean CDLIdlGenerator::FindIdl(String& idlNameOut)

{
  Boolean retVal = false;

  char* envVal = getenv("CDL_IDL");
  if (envVal != (char*)0)
    {
      idlNameOut = envVal;
      retVal = true;
    }
  else
    {
      Path foundIt;
      Search_path idlPath(CDLIdlSearchPath);
      Path idlName = "idl";
      if (idlPath.find(idlName, foundIt, Ksh_test::x))
	{
	  CDLReport.VReport(ErrorLevels::InfoMsg,
			    "Found idl as %s\n", (const char*)foundIt);
	  idlNameOut = (const String&)foundIt;
	  retVal = true;
	}
      else
	{
	  CDLReport.Report(ErrorLevels::ErrorMsg,
			   "Couldn't locate idl\n");
	  idlNameOut = "";
	}
    }
  return retVal;
}

/*
 * RunIdl
 *
 *    fork and exec the IDl compiler to turn the .idl file into a .h
 * file. Check the exit status and report any errors that turned up.
 */

Boolean CDLIdlGenerator::RunIdl(Boolean makeStubs)

{
  int status = 0;
#if HAVE_UNION_WAIT
  union wait wStatus;
  union wait* waiter;
  wStatus.w_status = 0;
#else
  int* statusP;
#endif
  String idlNameStr;
  if (!FindIdl(idlNameStr))
    {
      return false;
    }
  char* idlName = (char*)((const char*)idlNameStr);

  CDLReport.Report(ErrorLevels::TraceMsg, "In CDLIdlGenerator::RunIdl\n");
  ArgList idlArgs(origArgs, "IUD");
  idlArgs.PutEnd("-I.");
  idlArgs.PutEnd(CDLDCEIncludeCmd);
  idlArgs.PutEnd(newName);
  if (!makeStubs)		// Disable "caux" and "saux" if user wants to
    {
      idlArgs.PutEnd("-server");
      idlArgs.PutEnd("none");
      idlArgs.PutEnd("-client");
      idlArgs.PutEnd("stub");
    }
  idlArgs.PutEnd("-keep");
  idlArgs.PutEnd("c_source");		// Don't compile IDL output
  idlArgs.PutFront(idlName);
  Boolean retVal = true;

  if (vfork() == 0)
    {
#if CONST_EXEC_ARGV
      char* const* idlArgv = idlArgs.GetArgv();
#else
      const char ** idlArgv = (const char **)idlArgs.GetArgv();
#endif
      status = execv(idlName, idlArgv);
      if (status == -1)
	{
	  CDLReport.VReportMsg(ErrorLevels::ErrorMsg, 501, "idl");
	  retVal = false;
	}
      retVal = true;
    } 
  else 
    {
#if HAVE_UNION_WAIT
      waiter = &wStatus;
      wait(waiter);
      status = wStatus.w_status;
#else
      statusP = &status;
      wait(statusP);
#endif
      if (WIFEXITED(status) && (WEXITSTATUS(status) != 0))
        {
	  CDLReport.VReportMsg(ErrorLevels::ErrorMsg, 502, "idl", WEXITSTATUS(status));
          retVal = false;
        }
      else if (WIFSIGNALED(status))
        {
	  CDLReport.VReportMsg(ErrorLevels::ErrorMsg, 503, "idl", WTERMSIG(status));
          retVal = false;
        }
      else
        retVal = true;
    }
  return retVal;
}

/*
 * BuildFieldList
 *    Recursively build a list of FieldNode objects that represent all the
 * fields in this record. For now, we'll just use this to determine about
 * how big it is.
 */

Boolean CDLIdlGenerator::BuildFieldList(CDLRecord& rec, List<FieldNode>& out)

{
  return ExpandOneRecord(&rec, "", out);
}

Boolean CDLIdlGenerator::ExpandOneRecord(CDLRecord* rec, const String& name,
					 List<FieldNode>& out)

{
  CDLField* curField;
  CDLField* curLengthField;
  FieldNode* newField;
  FieldNode* newLengthField;
  int inc;
  int arrayPos;
  int realArrayPos;
  Boolean isArray;
  String thisFieldName;
  String lengthName;

  for (inc = 0; inc < rec->NumFields(); inc++)
    {
      curField = rec->Field(inc);
      isArray = (curField->ArraySize() > 1);
      for (arrayPos = 0; arrayPos < curField->ArraySize(); arrayPos++)
	{
	  if (isArray)
	    {
	      realArrayPos = arrayPos;
	    }
	  else
	    {
	      realArrayPos = -1;
	    }
	  GenWholeFieldName('.', curField->Name(), name, 0,
			    thisFieldName, realArrayPos);
	  if (IsRecord(curField->Type()))
	    {
	      thisFieldName += ".";
	      ExpandOneRecord(curField->Record(), thisFieldName, out);
	    }
	  else
	    {
	      newField = new FieldNode(thisFieldName, curField);
	      newField->VariableName(thisFieldName);
	      out.put(*newField);
	      delete newField;
	      if (curField->Type() == CDLField::BYTES)
		{
		  lengthName = curField->Name() + CDLByteArrayLengthSuffix;
		  curLengthField = new CDLField(CDLField::LONG, 
						CDLField::UNSIGNED, 
						lengthName, 1);
		  GenWholeFieldName('.', lengthName, name, 0,
				    thisFieldName);
		  newLengthField = new FieldNode(thisFieldName, curLengthField);
		  newLengthField->VariableName(thisFieldName);
		  out.put(*newLengthField);
		  delete newLengthField;
		  delete curLengthField;
		}
	    }
	}
    }
  return true;
}


