/*
 * CDLNidlGenerator:
 *
 *    Read the records given us and spit out an NIDL file (or files) that 
 * represents our record. 
 *
 * cdlnidlgenerator.C,v 1.1 1994/04/19 19:21:09 gjb Exp 
 */

    /*********************************************************************
    Copyright (C) 1993-1994 by Citibank, N.A., all rights reserved.

    Permission to use, copy, modify and distribute this software and its 
    documentation for any purpose and without fee, except possible 
    redistribution costs, is hereby granted, provided that (1) the above 
    copyright notice appears in all copies and (2) both that copyright-
    notice and this permission-notice appear in supporting documentation and 
    (3) the name of Citibank, N.A. is not used in advertising or publicity 
    pertaining to distribution of the software without specific and written 
    prior permission of Citibank, N.A.

    The Software is provided "as is."  Citibank, N.A. disclaims all 
    warranties, express or implied, including but not limited to warranties 
    of suitability, merchantability and fitness for any purpose.  In no 
    event shall Citibank, N.A. be responsible for any loss of use, data or 
    profit or for any other special, indirect, incidental or consequential 
    damages, neither in an action of contract nor in an action of tort, 
    arising out of or in connection with the use or performance of this 
    software.
    *********************************************************************/


#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <unistd.h>
#include <objdce/defs.H>
#include <Search_path.h>
#include <cdlnidlgenerator.H>
#include <cdlfield.H>
#include <cdlrecord.H>
#include <parselist.H>
#include <cdlsuffixes.H>
#include <cdlenuminfo.H>
#include <objdce/objtranparams.H>

/*
 * Handy constants
 *
 * HeapThreshold -- Above this many bytes, records have the [heap] attribute
 */

static const int HeapThreshold = 1024;

/*
 * CDLNidlGenerator
 *
 *    Not much to do here, except pass our arguments up the line.
 */

CDLNidlGenerator::CDLNidlGenerator(const ArgList& args) : CDLGenerator(args)

{	
  newName = "";
  mainName = "";
}

CDLNidlGenerator::~CDLNidlGenerator()

{
}

/* 
 * Start
 *
 *    Start parsing by generating a name for the output file and
 * opening it. If that works, generate the header for our file, including
 * comments and the "interface" declaration.
 *    This function also opens the .acf file, which primarily exists so
 * we can put the "out_of_line" declaration there. Additionally, it 
 * calls GenEnums to spit out the list of "enum" declarations.
 */

Boolean CDLNidlGenerator::Start(const String& fileName)

{
  String interfaceName;
  String importFileName;

  MakeFileName(fileName, newName, CDLNidlFileSuffix);
  CDLReport.VReport(ErrorLevels::TraceMsg, 
		    "Made File name: %s\n", (const char*)newName);
  MakeFileName(fileName, interfaceName, CDLIdlInterfaceSuffix,
	       false, CDLIdlInterfaceMaxLen); 
  CDLReport.VReport(ErrorLevels::TraceMsg, "Made Interface name: %s\n", 
		    (const char*)interfaceName);
  mainName = newName;
  outFile.open(newName);
  if (outFile.fail())
    {
      CDLReport.VReportMsg(ErrorLevels::ErrorMsg, 1002, (const char*)newName);
      return false;
    }
  else
    {
      outFile << "/*" << endl;
      outFile << " * NIDL file generated by Citi-DDL:" << endl;
      outFile << " * cdlnidlgenerator.C,v 1.1 1994/04/19 19:21:09 gjb Exp" << endl;
      outFile << " */\n" << endl;

      outFile << "%c" << endl;
      outFile << "interface " << interfaceName << endl;
      outFile << "{" << endl;
      Indent();
      for (int inc = 0; inc < NumMainIncludeNames(); inc++)
	{
	  MakeFileName(GetMainIncludeName(inc), importFileName, 
		       CDLNidlFileSuffix);
	  CDLReport.VReport(ErrorLevels::TraceMsg,
			    "Made Import File Name: %s\n",
			    (const char*)importFileName);
	  outFile << indent << "import \"" << importFileName << "\";" << endl;
	}
      cout << endl;

      GenEnums();

      return true;
    }
}

/*
 * End
 *
 *    Finish up generating the file by writing the close brace and
 * closing the output file.
 */

Boolean CDLNidlGenerator::End()

{
  Outdent();
  outFile << "}" << endl;
  outFile << flush;
  outFile.close();
  return true;
}

/*
 * DeleteFiles
 *
 *    Delete the output file. Called when parsing fails.
 */

void CDLNidlGenerator::DeleteFiles()

{
  outFile.close();
  unlink(mainName);
}

/*
 * Generate
 *
 *    Given a CDLRecord object, generate its NIDL type definition.
 * This is a pretty straightforward one-to-one mapping of fields in the
 * CDL record to fields in the NIDL struct.
 */

Boolean CDLNidlGenerator::Generate(CDLRecord& rec)

{
  String typeName;
  String modName;
  CDLField* curField;
  CDLRecord* curRecord;
  String structName;

  structName = rec.Name() + CDLIdlRecordSuffix;
  CDLReport.Report(ErrorLevels::TraceMsg, "In CDLNidlGenerator:Generate\n");
  outFile << "/*" << endl;
  outFile << " * " << structName << ":" << endl;
  outFile << " * " << endl;
  outFile << " * " << rec.Documentation() << endl;
  outFile << " */\n" << endl;
  
  outFile << indent << "typedef struct " << endl;
  outFile << indent << "{" << endl;
  Indent();
  
  for (int inc = 0; inc < rec.NumFields(); inc++)
    {
      curField = rec.Field(inc);
      switch (curField->Modifier())
	{
	case CDLField::UNSIGNED:
	  modName = "unsigned ";
	  break;
	case CDLField::SIGNED:
	case CDLField::NONE:
	  modName = "";
	  break;
	}
      switch (curField->Type())
	{
	case CDLField::REAL:
	  typeName = "double";
	  break;
	case CDLField::BYTES:
	case CDLField::VARBYTES:
	  typeName = "byte";
	  break;
	case CDLField::STRING:
	  typeName = "string0";
	  break;
	case CDLField::RECORD:
	  curRecord = curField->Record();
	  if (curRecord != (CDLRecord*)0)
	    {
	      typeName = curRecord->Name();
	      typeName += CDLIdlRecordSuffix;
	    }
	  else
	    return false;
	  break;
	case CDLField::HYPER:
	  typeName = modName + "hyper int";
	  break;
	case CDLField::LONG:
	  typeName = modName + "long int";
	  break;
	case CDLField::SHORT:
	  typeName = modName + "short int";
	  break;
	case CDLField::CHAR:
	  typeName = modName + "small int";
	  break;
	case CDLField::CHARRAY:
	  typeName = "char";
	  break;
	case CDLField::ENUM:
	  typeName = curField->RealTypeName();
	  break;
	}
      if ((curField->Type() == CDLField::VARBYTES) || 
	  (curField->Type() == CDLField::BYTES))
	{
	  outFile << indent << "unsigned long " << curField->Name();
	  outFile << CDLByteArrayLengthSuffix;
	  if (curField->ArraySize() > 1)
	    outFile << "[" << curField->ArraySize() << "]";
	  outFile << ";" << endl;
	}
      if (curField->ArraySize() == 1)
	{
	  switch(curField->Type())
	    {
	    case CDLField::VARBYTES:
	      outFile << indent << "[last_is(";
	      outFile << curField->Name() << CDLByteArrayLengthSuffix;
	      outFile << ")] " << typeName << " " << curField->Name() << "[";
	      outFile << curField->Length() << "];" << endl;
	      break;
	    case CDLField::BYTES:
	    case CDLField::CHARRAY:
	      outFile << indent << typeName << " " << curField->Name();
	      outFile << "[" << curField->Length() << "];" << endl;
	      break;
	    case CDLField::STRING:
	      outFile << indent << typeName << " [";
	      outFile << curField->Length() + 1 << "]" << curField->Name();
	      outFile << ";" << endl;
	      break;
	    default:
	      outFile << indent << typeName << " ";
	      outFile << curField->Name() << ";" << endl;
	      break;
	    }
	}
      else
	{
	  switch(curField->Type())
	    {
	    case CDLField::VARBYTES:
	    case CDLField::BYTES:
	    case CDLField::CHARRAY:
	      outFile << indent << typeName << " " << curField->Name();
	      outFile << "[" << curField->ArraySize() << "][";
	      outFile << curField->Length() << "];" << endl;
	      break;
	    case CDLField::STRING:
	      outFile << indent << "char " << curField->Name();
	      outFile << "[" << curField->ArraySize() << "] [";
	      outFile << curField->Length() + 1 << "];" << endl;
	      break;
	    default:
	      outFile << indent << typeName << " ";
	      outFile << curField->Name() << "[" << curField->ArraySize();
	      outFile << "];" << endl;
	      break;
	    }
	}
    }
  Outdent();
  outFile << indent << "} " << structName << ";\n" << endl;

  // Supporting IDL arrays would be a royal pain to do in NIDL
  if (rec.MakeIdlArray())
    {
      CDLPNumWarning(1004);
    }

  return true;
}

/*
 * GenEnums
 *    Based on the map of defined "enum" types, spit out all the necessary
 * "enum" declarations. This is necessary because we want to use enums
 * in IDL declarations, so we must define them in the .idl file.
 */

Boolean CDLNidlGenerator::GenEnums()

{
  Mapiter<Symbol, CDLEnumInfo> enumIter(CDLEnums);
  Map_pair<Symbol, CDLEnumInfo>* curPair;
  String curCase;
  CDLEnumInfo curEnum;
  int inc;
  
  while ((curPair = enumIter.next()) != 0)
    {
      curEnum = curPair->value;
      if ((!curEnum.Imported()) && (curEnum.NumCases() > 0))
	{
	  outFile << indent << "typedef enum { ";
	  Indent();
	  for (inc = 0; inc < (curEnum.NumCases() - 1); inc++)
	    {
	      curCase = curEnum.Case(inc);
	      outFile << curCase << ", " << endl << indent;
	    }
	  curCase = curEnum.Case(curEnum.NumCases() - 1);
	  outFile << curCase << " } " << curEnum.Name() << ";" << endl;
	}
    }
  return true;
}

/*
 * FindNidl
 *    Using the Standard Components "Search_path" class, return the full
 * pathname of nidl in "idlNameOut". However, if CDL_NIDL is set, just
 * use that name.
 */

Boolean CDLNidlGenerator::FindNidl(String& idlNameOut)

{
  Boolean retVal = false;

  char* envVal = getenv("CDL_NIDL");
  if (envVal != (char*)0)
    {
      idlNameOut = envVal;
      retVal = true;
    }
  else
    {
      Path foundIt;
      Search_path idlPath(CDLIdlSearchPath);
      Path idlName = "nidl";
      if (idlPath.find(idlName, foundIt, Ksh_test::x))
	{
	  CDLReport.VReport(ErrorLevels::InfoMsg,
			    "Found nidl as %s\n", (const char*)foundIt);
	  idlNameOut = (const String&)foundIt;
	  retVal = true;
	}
      else
	{
	  CDLReport.Report(ErrorLevels::ErrorMsg,
			   "Couldn't locate nidl\n");
	  idlNameOut = "";
	}
    }
  return retVal;
}

/*
 * RunNidl
 *
 *    fork and exec the NIDL compiler to turn the .nidl file into a .h
 * file. Check the exit status and report any errors that turned up.
 */

Boolean CDLNidlGenerator::RunNidl(Boolean makeStubs)

{
  int status = 0;
#if HAVE_UNION_WAIT
  union wait wStatus;
  union wait* waiter;
  wStatus.w_status = 0;
#else
  int* statusP;
#endif
  String idlNameStr;
  String curArg;
  int curArgNum;

  if (!FindNidl(idlNameStr))
    {
      return false;
    }
  char* idlName = (char*)((const char*)idlNameStr);

  CDLReport.Report(ErrorLevels::TraceMsg, "In CDLNidlGenerator::RunNidl\n");
  ArgList idlArgs(origArgs, "I");		// Take off all "-I" args
  ArgList realArgs;

  realArgs.PutEnd(newName);
  realArgs.PutEnd("-m");
  for (curArgNum = 0; curArgNum < idlArgs.GetArgc(); curArgNum++)
    {
      realArgs.PutEnd("-idir");
      curArg = idlArgs.Get(curArgNum).chunk(2);  // Strip off first two chars
      realArgs.PutEnd(curArg);
    }
  realArgs.PutEnd("-idir");
  realArgs.PutEnd(".");
  realArgs.PutFront(idlName);
  Boolean retVal = true;

  if (vfork() == 0)
    {
#if CONST_EXEC_ARGV
      char* const* idlArgv = realArgs.GetArgv();
#else
      const char ** idlArgv = (const char **)realArgs.GetArgv();
#endif
      status = execv(idlName, idlArgv);
      if (status == -1)
	{
	  CDLReport.VReportMsg(ErrorLevels::ErrorMsg, 501, "nidl");
	  retVal = false;
	}
      retVal = true;
    } 
  else 
    {
#if HAVE_UNION_WAIT
      waiter = &wStatus;
      wait(waiter);
      status = wStatus.w_status;
#else
      statusP = &status;
      wait(statusP);
#endif
      if (WIFEXITED(status) && (WEXITSTATUS(status) != 0))
        {
	  CDLReport.VReportMsg(ErrorLevels::ErrorMsg, 502, 
			       "nidl", WEXITSTATUS(status));
          retVal = false;
        }
      else if (WIFSIGNALED(status))
        {
	  CDLReport.VReportMsg(ErrorLevels::ErrorMsg, 503, 
			       "nidl", WTERMSIG(status));
          retVal = false;
        }
      else
        retVal = true;
    }
  return retVal;
}
