/*
 * CDLRecord:
 *
 *    This record contains lists of the fields in a record, its name
 * and the documentation string associated with it.
 *
 * cdlrecord.C,v 1.7 1994/02/01 22:33:14 gjb Exp
 */

    /*********************************************************************
    Copyright (C) 1993 by Citibank, N.A., all rights reserved.

    Permission to use, copy, modify and distribute this software and its 
    documentation for any purpose and without fee, except possible 
    redistribution costs, is hereby granted, provided that (1) the above 
    copyright notice appears in all copies and (2) both that copyright-
    notice and this permission-notice appear in supporting documentation and 
    (3) the name of Citibank, N.A. is not used in advertising or publicity 
    pertaining to distribution of the software without specific and written 
    prior permission of Citibank, N.A.

    The Software is provided "as is."  Citibank, N.A. disclaims all 
    warranties, express or implied, including but not limited to warranties 
    of suitability, merchantability and fitness for any purpose.  In no 
    event shall Citibank, N.A. be responsible for any loss of use, data or 
    profit or for any other special, indirect, incidental or consequential 
    damages, neither in an action of contract nor in an action of tort, 
    arising out of or in connection with the use or performance of this 
    software.
    *********************************************************************/


#include <cdlrecord.H>
#include <parselist.H>

#ifdef _AIX
#include <List.c>

#pragma define (Const_list_of_piter<CDLRecord>)
#pragma define (List_of_p<CDLRecord>)
#pragma define (List_of_piter<CDLRecord>)
#endif

/*
 * CDLRecord
 *
 *   Just set the members of the object to reasonable values based
 * on what the constructor was passed.
 */

CDLRecord::CDLRecord() : NamedObject()

{
  documentation = "";
  imported = false;
  makeIdlArray = false;
}

CDLRecord::CDLRecord(const String& recName, 
		     const String& recDocs,
		     Boolean fromIncludeFile,
		     Boolean idlArray) : NamedObject(recName)

{
  documentation = recDocs;
  imported = fromIncludeFile;
  makeIdlArray = idlArray;
}

CDLRecord::CDLRecord(const CDLRecord& rec)

{
  this->operator = (rec);
}

/*
 * ~CDLRecord
 *
 *    Nothing to delete.
 */

CDLRecord::~CDLRecord()

{
}

/*
 * DeleteFields
 *
 *    Delete all the fields in this record, preparing it for use to
 * represent another type. All the CDLField objects are deleted.
 */

void CDLRecord::DeleteFields()

{
  List_of_piter<CDLField> iter(fields);
  CDLField* deadField;
  iter.reset();
  while (iter.remove_next(deadField))
    delete deadField;
}  

/*
 * operator = 
 *
 *    Simply copy each field from that on the other record.
 */

CDLRecord& CDLRecord::operator = (const CDLRecord& rec)

{
  NamedObject::operator = (rec);
  documentation = rec.documentation;
  fields = rec.fields;
  imported = rec.imported;
  makeIdlArray = rec.makeIdlArray;
  return *this;
}

/*
 * AddField
 *
 *    Make sure a field with a duplicate name isn't already in the record.
 * If there's no duplicate, add the new field to the list of fields in this
 * record. The CDLField object must have been created already, which is
 * what the wrapper routines in mainparsewrap.C do.
 */

Boolean CDLRecord::AddField(CDLField* field)

{
  Boolean found = false;
  Boolean retVal;
  List_of_piter<CDLField> iter(fields);
  CDLField* curField;
  String errStr;
  
  if (field->Name() == name)		// Check for field name == record name
    {
      CDLReport.VGetMsg(1119, errStr, (const char*)(field->Name()));
      CDLPError(errStr);
      found = true;
    }
  else					// Check for duplicate field names
    {
      while (!found && iter.next(curField))
	{
	  if (*curField == *field)
	    {
	      found = true;
	      CDLReport.VGetMsg(1104, errStr, (const char*)(field->Name()), 
				(const char*)name);
	      CDLPError(errStr);
	      retVal = false;
	    }
	}
    }
  if (!found)			// No errors, so add the field to the record
    {
      fields.put(field);
      retVal = true;
    }
  return retVal;
}

/*
 * Find
 *
 *    Return a pointer to the field in this record with the specified
 * name, or NULL if there's no match. 
 */

CDLField* CDLRecord::Find(const String& findName)

{
  Boolean found = false;
  CDLField* retVal;
  Const_list_of_piter<CDLField> iter(fields);

  while (!found && iter.next(retVal))
    {
      if (retVal->Name() == findName)
	found = true;
    }
  if (!found)
    retVal = (CDLField*)0;
  return retVal;
}

