    /*********************************************************************
    Copyright (C) 1993-1994 by Citibank, N.A., all rights reserved.

    Permission to use, copy, modify and distribute this software and its 
    documentation for any purpose and without fee, except possible 
    redistribution costs, is hereby granted, provided that (1) the above 
    copyright notice appears in all copies and (2) both that copyright-
    notice and this permission-notice appear in supporting documentation and 
    (3) the name of Citibank, N.A. is not used in advertising or publicity 
    pertaining to distribution of the software without specific and written 
    prior permission of Citibank, N.A.

    The Software is provided "as is."  Citibank, N.A. disclaims all 
    warranties, express or implied, including but not limited to warranties 
    of suitability, merchantability and fitness for any purpose.  In no 
    event shall Citibank, N.A. be responsible for any loss of use, data or 
    profit or for any other special, indirect, incidental or consequential 
    damages, neither in an action of contract nor in an action of tort, 
    arising out of or in connection with the use or performance of this 
    software.
    *********************************************************************/


Punctuation 	[\"\'\[\]\;\:\{\}\,\=] 
Nonpunctuation 	[^\"\'\[\]\;\:\{\}\,\=] 
Digit		[0-9] 
Blank 		[ \t]
Blanks 		[ \t\n]* 
Nonblank 	[^ \t\n] 
TextChar	[^\"\'\[\]\;\:\{\}\,\=\t\n ] 
AllTextChar	[^\"]
AllStringChar	[^\"\n]
Hash		[\#]
Newline		[\n]

%{
/*
 * mainlex.l
 *
 *    This file is the lexical analyzer for the main .cdl file. Like all
 * lex source files, it's responsible for figuring out whether a token
 * (usually some text surrounded by whitespace) is a particular keyword,
 * an integer, a string, or whatever.
 *
 * mainlex.l,v 1.9 1994/02/01 22:33:37 gjb Exp
 */

extern char* yytext;

#include <mainparse.h>
#include <mainparsedefs.H>
#include <sys/param.h>
#include <parselist.H>
#include <sys/stat.h>

static char inputFileName[MAXPATHLEN];	/* This is the file we're reading from */
static int newlineCount;		/* Here, we count newlines

/*
 * GotToken
 *    This is a debugging function that prints out what token we got.
 */

void GotToken(n)

char* n;

{
  if (yydebug)
    printf("%s: -->%s<--, line %i + %i\n", n, yytext, CDLMainCurrentLine,
	   newlineCount);
}

/*
 * CDLMainBuffer
 *    Count the number of newlines in "yytext" -- which is the text of the
 * current token -- and add that to the current line count. For this
 * to work, the rest of the parser must call this function for every token
 * it finds, or the line counts will never come out right.
 */

void CDLMainBuffer()

{
  register char* curChar;

  for (curChar = yytext; *curChar != '\0'; curChar++)
    {
      if (*curChar == '\n')
	newlineCount++;
    }
  CDLMainCurrentLine += newlineCount;
  SetCurrentLine(CDLMainCurrentLine);
  newlineCount = 0;
}

%}

%%

 /****
 *    This section contains the definitions of the tokens. Each token has
 * a definition, followed by some C code in squiggly brackets. See the
 * "lex" and "flex" documentation for more info.
 *    lex will match the first entry in this list that matches the token
 * being parsed. So, here the order is important, and must go from the
 * most-specific to least-specific.
 ****/

 /*
 * <<EOF>>
 *     This code is executed when the end of a file is reached. Since
 * the "import" statement allows a .cdl file to include another, then
 * here we try to pop from the include file stack. When we reach
 * the bottom of the stack, we terminate the parsing by calling yyterminate.
 * Otherwise, we use yy_switch_to_buffer to go back to the file that imported
 * this one, and re-set the current line counter.
 */

<<EOF>> {
  void* bufPtr;
  YY_BUFFER_STATE oldBuf;
  YY_BUFFER_STATE curBuf;
  char* newName;

  PopIncludeFile(&bufPtr, &newName, &CDLMainCurrentLine);
  SetCurrentLine(CDLMainCurrentLine);
  if (bufPtr == (void*)0)
    {
      ClearIncludeNames();
      yyterminate();		/* The end of the top-level file */
    }
  else
    {				/* The end of an "imported" file */
      unlink(inputFileName);
      CDLMainSetName(newName);
      curBuf = YY_CURRENT_BUFFER;
      oldBuf = (YY_BUFFER_STATE)bufPtr;
      yy_delete_buffer(curBuf);
      yy_switch_to_buffer(oldBuf);
    }
}

 /*
 * #line
 *     These next three rules handle the "#line" directive produced by
 * cpp. It takes several forms, and basically lets us re-set the
 * line number of the file currently being parsed.
 */

{Hash}{Blank}+{Digit}+{Blank}+\"{TextChar}+\"{Blank}+{Digit}+ {
  int newLineNo;
  sscanf(yytext, "# %i", &newLineNo);
  GotToken("Line directive");
  CDLMainCurrentLine = newLineNo - 1;	/* Other stuff counts newlines */
  SetCurrentLine(CDLMainCurrentLine);
}

{Hash}{Blank}+{Digit}+{Blank}+\"{TextChar}+\" {
  int newLineNo;
  sscanf(yytext, "# %i", &newLineNo);
  GotToken("Line directive");
  CDLMainCurrentLine = newLineNo - 1;	/* Other stuff counts newlines */
  SetCurrentLine(CDLMainCurrentLine);
}

{Hash}{Blank}+{Digit}+ {
  int newLineNo;
  sscanf(yytext, "# %i", &newLineNo);
  GotToken("Line directive");
  CDLMainCurrentLine = newLineNo - 1;	/* Other stuff counts newlines */
  SetCurrentLine(CDLMainCurrentLine);
}

 /*
 * Punctuation
 *     These rules handle punctuation marks. lex expects tokens to be
 * separated by white space, but in YADDL we don't require punctuation to
 * be so separated. So, we have a few separate cases for punctuation.
 * If the punctuation character is followed by a character that isn't
 * blank, we have the parser backtrack a character, and just return
 * the punctuation character.
 */

{Punctuation}  {
  GotToken("Punctuation");
  CDLMainBuffer();
  return (int)(yytext[yyleng - 1]);
}

{Punctuation}{Nonblank} {
    register int inc;
    int firstChar;

    firstChar = (int)yytext[0];
    yyless(1);
    GotToken("Punctuation + Nonblank");
    CDLMainBuffer();
    return firstChar;
  }

 /*
 * Keywords
 *    These are the keywords supported by YADDL. A few are grouped
 * together while others have just one possibility, but each represents 
 * a specific thing we'll be looking for in the actual yacc parser.
 */

real {
    AddToParseList(yytext);
    GotToken("RealTypeSpec");
    CDLMainBuffer();
    return RealTypeSpec;
  }

record {
    GotToken("RecordSpec");
    CDLMainBuffer();
    return RecordSpec;
  }

enum {
    GotToken("EnumSpec");
    CDLMainBuffer();
    return EnumSpec;
  }

string {
    AddToParseList(yytext);
    GotToken("StringTypeSpec");
    CDLMainBuffer();
    return StringTypeSpec;
  }

bytes | 
varBytes {
    AddToParseList(yytext);
    GotToken("VarTypeSpec");
    CDLMainBuffer();
    return VarTypeSpec;
  }

hyper | 
long | 
short {
    AddToParseList(yytext);
    GotToken("IntTypeSpec");
    CDLMainBuffer();
    return IntTypeSpec;
  }

char {
    AddToParseList(yytext);
    GotToken("CharTypeSpec");
    CDLMainBuffer();
    return CharTypeSpec;
  }

signed | 
unsigned {
    AddToParseList(yytext);
    GotToken("IntModSpec");
    CDLMainBuffer();
    return IntModSpec;
  }

documentation {
    GotToken("DocumentationSpec");
    CDLMainBuffer();
    return DocumentationSpec;
  }

idl_array {
    GotToken("IdlArraySpec");
    CDLMainBuffer();
    return IdlArraySpec;
  }

import {
    GotToken("ImportSpec");
    CDLMainBuffer();
    return ImportSpec;
  }

const {
    GotToken("ConstSpec");
    CDLMainBuffer();
    return ConstSpec;
  }

 /*
 * Other tokens
 *    These represent constant numeric and string information that
 * might be found in the file. A Real must be a number that includes a decimal
 * point. An Integer has no decimal. A CharConst is a single character 
 * surrounded by single quotes. A QuotedCString is a string surrounded
 * by double quotation marks that contains no newlines so it can be
 * used in C or C++. A QuotedString is a QuotedCString that can include
 * newline characters. A TextString is pretty much any other type of
 * string. Finally, Blanks catches any extra whitespace that shows up
 * to prevent parsing errors. 
 */

{Digit}+\.{Digit}+ {
    AddToParseList(yytext);
    GotToken("Real");
    CDLMainBuffer();
    return Real;
  }

{Digit}+ {
    AddToParseList(yytext);
    GotToken("Integer");
    CDLMainBuffer();
    return Integer;
  }

\'{TextChar}\' {
    char tmpStr[2];

    tmpStr[0] = yytext[1];
    tmpStr[1] = '\0';
    AddToParseList(tmpStr);
    GotToken("CharConst");
    CDLMainBuffer();
    return CharConst;
  }

\"{AllStringChar}+\" {
    char tmpStr[CDL_MAIN_LONG_FIELD_LEN];

    strncpy(tmpStr, (yytext + 1), CDL_MAIN_LONG_FIELD_LEN - 1);
    if (yyleng < CDL_MAIN_LONG_FIELD_LEN)
      tmpStr[yyleng - 2] = '\0';
    else
      tmpStr[CDL_MAIN_LONG_FIELD_LEN - 1] = '\0';
    AddToParseList(tmpStr);
    GotToken("QuotedString");
    CDLMainBuffer();
    return QuotedCString;
  }

\"{AllTextChar}+\" {
    char tmpStr[CDL_MAIN_LONG_FIELD_LEN];

    strncpy(tmpStr, (yytext + 1), CDL_MAIN_LONG_FIELD_LEN - 1);
    if (yyleng < CDL_MAIN_LONG_FIELD_LEN)
      tmpStr[yyleng - 2] = '\0';
    else
      tmpStr[CDL_MAIN_LONG_FIELD_LEN - 1] = '\0';
    AddToParseList(tmpStr);
    GotToken("QuotedString");
    CDLMainBuffer();
    return QuotedString;
  }

{TextChar}+ {
    AddToParseList(yytext);
    GotToken("TextString");
    CDLMainBuffer();
    return TextString;
  }

{Blanks}  {
  CDLMainBuffer();
}

%%

 /*
 * OpenInclude
 *    Open the named include file by checking the list of include 
 * directories, pushing the file name on a list of included files,
 * and pushing the lex directory stack. This function is here rather
 * than elsewhere because it uses lots of data types and functions defined
 * by lex.
 */

int OpenInclude(fileName)

char* fileName;

{
  int retVal = 0;
  FILE* newFile;
  YY_BUFFER_STATE newBuf;
  YY_BUFFER_STATE currentBuf;
  char realFileName[MAXPATHLEN];

  newFile = OpenIncludeFile(fileName,		/* Check list of include files */
			    realFileName, inputFileName);	
  if (newFile == (FILE*)0)		/* Something went wrong */
    {
      fprintf(stderr, "Error opening file %s:\n", fileName);
      perror("");
      return retVal;
    }
  else
    {
      if (!IsIncluding())
	AddMainIncludeName(fileName);
      if (CheckIncludeName(fileName))
	fprintf(stderr, "File %s already included\n", fileName);
      else
	{
	  AddIncludeName(fileName);
	  currentBuf = YY_CURRENT_BUFFER;
	  PushIncludeFile((void*)currentBuf, CDLMainGetName(), 
			  CDLMainCurrentLine);
	  CDLMainCurrentLine = 1;
	  SetCurrentLine(CDLMainCurrentLine);
	  CDLMainSetName(realFileName);
	  newBuf = yy_create_buffer(newFile, YY_BUF_SIZE);
	  yy_switch_to_buffer(newBuf);
	}
    }
}

