%{
/*
 * mainparse.y:
 *
 *    This is the parser for the main .cdl files that specify the records.
 * It defines the syntax of the .cdl file, based on the tokens specified
 * in mainlex.l. Parsing starts at the end, at the "WholeFile" rule,
 * and basically proceeds upwards, with the most specific constructs
 * being at the top and the highest-level constructs being at the
 * bottom. To make sense of this file, start from the bottom and
 * flatten as you go up.
 *
 * mainparse.y,v 1.10 1994/04/07 19:26:28 gjb Exp
 */

    /*********************************************************************
    Copyright (C) 1993-1994 by Citibank, N.A., all rights reserved.

    Permission to use, copy, modify and distribute this software and its 
    documentation for any purpose and without fee, except possible 
    redistribution costs, is hereby granted, provided that (1) the above 
    copyright notice appears in all copies and (2) both that copyright-
    notice and this permission-notice appear in supporting documentation and 
    (3) the name of Citibank, N.A. is not used in advertising or publicity 
    pertaining to distribution of the software without specific and written 
    prior permission of Citibank, N.A.

    The Software is provided "as is."  Citibank, N.A. disclaims all 
    warranties, express or implied, including but not limited to warranties 
    of suitability, merchantability and fitness for any purpose.  In no 
    event shall Citibank, N.A. be responsible for any loss of use, data or 
    profit or for any other special, indirect, incidental or consequential 
    damages, neither in an action of contract nor in an action of tort, 
    arising out of or in connection with the use or performance of this 
    software.
    *********************************************************************/


#include <stdio.h>
#include <mainparse.h>
#include <mainparsedefs.H>
#include <parselist.H>

extern int OpenInclude();
%}

%token Integer TextString QuotedString RealTypeSpec StringTypeSpec VarTypeSpec 
%token IntTypeSpec IntModSpec DocumentationSpec RecordSpec ImportSpec
%token IdlArraySpec ConstSpec Real CharConst CharTypeSpec QuotedCString
%token EnumSpec
%start WholeFile

%%

/*
 * Basic types
 *    These are abstractions of the basic type keywords defined in
 * mainlex.l. Note especially that an ArraySize can either be an integer
 * or a text string. The routine that creates new fields is responsible
 * for evaluating this constant. 
 */

RealType	: RealTypeSpec
		;
StringType	: StringTypeSpec
		;
VarType		: VarTypeSpec
		;
IntType		: IntTypeSpec
		;
CharType	: CharTypeSpec
		;
UserType	: RecordSpec TextString
		;
EnumType	: EnumSpec TextString
  		;
IntMod		: IntModSpec
		| /* Nothing */
		;
ArraySize	: Integer
		| TextString
		;

/*
 * IntField
 *    An integer field -- it can be of type char, short, long, or hyper,
 * and include an optional signed or unsigned specifier before the type
 * name. This can also be an array of integer fields.
 */

IntField	: IntMod IntType TextString ';'
		{
		  if (ParseListLength() == 2)
		    CDLMainAddField("", GetFromParseList(0), 
				    GetFromParseList(1), "", "");
		  else
		    CDLMainAddField(GetFromParseList(0), GetFromParseList(1),
				    GetFromParseList(2), "", "");
		  ClearParseList();
		}
		| IntMod IntType TextString '[' ArraySize ']' ';'
		{
		  if (ParseListLength() == 3)
		    CDLMainAddField("", GetFromParseList(0), 
				    GetFromParseList(1), "", 
				    GetFromParseList(2));
		  else
		    CDLMainAddField(GetFromParseList(0), GetFromParseList(1),
				    GetFromParseList(2), "", 
				    GetFromParseList(3));
		  ClearParseList();
		}
		;

/*
 * CharField
 *    A char type. This can be a single character, an array of characters
 * (i.e. a non-null-terminated string), or an array of arrays of characters.
 */

CharField	: IntMod CharType TextString ';'
		{
		  if (ParseListLength() == 2)
		    CDLMainAddField("", GetFromParseList(0),
				    GetFromParseList(1), "", "");
		  else
		    CDLMainAddField(GetFromParseList(0), GetFromParseList(1),
				    GetFromParseList(2), "", "");
		  ClearParseList();
		}
		| IntMod CharType TextString '[' ArraySize ']' ';'
		{
		  if (ParseListLength() == 3)
		    CDLMainAddField("", GetFromParseList(0),  
				    GetFromParseList(1),
				    GetFromParseList(2), "");
		  else
		    CDLMainAddField(GetFromParseList(0), GetFromParseList(1),
				    GetFromParseList(2), 
				    GetFromParseList(3), "");
		  ClearParseList();
		}
		| IntMod CharType TextString '[' ArraySize ']' '[' ArraySize ']' ';'
		{
		  if (ParseListLength() == 4)
		    CDLMainAddField("", GetFromParseList(0), 
				    GetFromParseList(1), GetFromParseList(3), 
				    GetFromParseList(2));
		  else
		    CDLMainAddField(GetFromParseList(0), GetFromParseList(1),
				    GetFromParseList(2), GetFromParseList(4), 
				    GetFromParseList(3));
		  ClearParseList();
		}
		;

/*
 * StringField
 *    A field of type "string", or an array of strings.
 */

StringField	: StringType TextString '[' ArraySize ']' ';'
                {
		  CDLMainAddField("", GetFromParseList(0), GetFromParseList(1),
				  GetFromParseList(2), "");
		  ClearParseList();
		}
		| StringType TextString '[' ArraySize ']' '[' ArraySize ']' ';'
                {
		  CDLMainAddField("", GetFromParseList(0), GetFromParseList(1),
				  GetFromParseList(3), GetFromParseList(2));
		  ClearParseList();
		}
		;

/*
 * VarField
 *    A field of type "bytes" or "varBytes". Arrays of this type aren't
 * supported yet.
 */

VarField	: VarType TextString '[' ArraySize ']' ';'
                {
		  CDLMainAddField("", GetFromParseList(0), GetFromParseList(1),
				  GetFromParseList(2), "");
		  ClearParseList();
		}
		| VarType TextString '[' ArraySize ']' '[' ArraySize ']' ';'
                {
		  CDLMainAddField("", GetFromParseList(0), GetFromParseList(1),
				  GetFromParseList(3), GetFromParseList(2));
		  ClearParseList();
		}
		;

/* 
 * RealField
 *    A field of the "real" type, or an array of reals.
 */

RealField	: RealType TextString ';'
		{
 		  CDLMainAddField("", GetFromParseList(0), 
				  GetFromParseList(1), "", "");
		  ClearParseList();
		}
		| RealType TextString '[' ArraySize ']' ';'
		{
 		  CDLMainAddField("", GetFromParseList(0), 
				  GetFromParseList(1), "", GetFromParseList(2));
		  ClearParseList();
		}
		;

/*
 * UserField
 *    This is a user-defined field, which must start with the "record"
 * keyword, or an array of such records. The function we call to add the
 * field will generate an error message if the record hasn't been defined
 * yet.
 */

UserField	: UserType TextString ';'
		{
		  CDLMainAddUserField(GetFromParseList(0), 
				      GetFromParseList(1), "");
		  ClearParseList();
		}
		| UserType TextString '[' ArraySize ']' ';'
		{
		  CDLMainAddUserField(GetFromParseList(0), GetFromParseList(1),
				      GetFromParseList(2));
		  ClearParseList();
		}
		;

/*
 * EnumField
 *    An enum is also a user-defined type of field, which starts with the
 * "enum" keyword inside a record definition. Here, we call a function
 * that looks up our enum name before continuing.
 */

EnumField	: EnumType TextString ';'
		{
		  CDLMainAddEnumField(GetFromParseList(0),
				      GetFromParseList(1), "");
		  ClearParseList();
		}
		| EnumType TextString '[' ArraySize ']' ';'
		{
		  CDLMainAddEnumField(GetFromParseList(0),
				      GetFromParseList(1), GetFromParseList(2));
		  ClearParseList();
		}
		;

/*
 * AmbiguousField
 *    The "enum" and "record" keywords are a pain, so this function takes
 * a simple TextString as a field name, then compares it to the existing
 * record and enum names before creating the field.
 */

AmbiguousField	: TextString TextString ';'
  		{
		  CDLMainAddAmbiguousField(GetFromParseList(0),
					   GetFromParseList(1), "");
		  ClearParseList();
		}
		| TextString TextString '[' ArraySize ']' ';'
		{
		  CDLMainAddAmbiguousField(GetFromParseList(0),
					   GetFromParseList(1),
					   GetFromParseList(2));
		  ClearParseList();
		}
		;

/*
 * WrongField
 *    This defines several common errors programmers might make when
 * defining a field. They attempt to print out a nice error message. 
 */

WrongField	: TextString TextString TextString '[' ArraySize ']' ';'
		{
		  CDLPNumError(1111);
		  CDLMainFailed = 1;
		  ClearParseList();
		}		
		| TextString TextString TextString ';'
		{
		  CDLPNumError(1111);
		  CDLMainFailed = 1;
		  ClearParseList();
		}		
		;

/*
 * Field, FieldList
 *    The body of a record specification constists of a FieldList, which is
 * list of the individual field types.
 */

Field		: UserField
		| IntField
		| CharField
		| VarField
		| StringField
		| RealField
		| EnumField
		| AmbiguousField
  		| WrongField
		;
FieldList	: FieldList Field
		| Field
		{}
		| error ';'
  		{
		  CDLMainFailed = 1;
		  CDLPNumError(1111);
		  ClearParseList();
		  yyerrok;
		}
		;

/*
 * RecordMods, RecordModList, RecordMod, DocString
 *    These are the optional modifiers at the start of a "record"
 * specification that go before the list of fields. The list is
 * either a comma-separated list of modifiers surrounded by
 * square brackets, or nothing at all. If it isn't nothing, there
 * are two types of modifiers -- the "idl_array" keyword, or the
 * "documentation" specification. The format of "documentation"
 * is defined by DocString.
 */

DocString	: DocumentationSpec ':' QuotedString
  		{
		  CDLMainSetDocumentation(GetParseListEnd());
		}
		| DocumentationSpec ':' QuotedCString
  		{
		  CDLMainSetDocumentation(GetParseListEnd());
		}
		;
RecordMod	: DocString
		{}
                | IdlArraySpec
    		{
		  CDLMainMakeIdlArray = 1;
  		}
		| TextString
		{
		  CDLPNumError(1112);
		  CDLMainFailed = 1;
		  ClearParseList();
		}
		;
RecordModList	: RecordModList ',' RecordMod
                | RecordMod
                ;
RecordMods	: '[' RecordModList ']'
		| /* Nothing */
		;

/*
 * RecordStart
 *    This is how a record must start. It must have the "record"
 * keyword, a name, and the RecordMods modifiers.
 */

RecordStart	: RecordSpec TextString RecordMods
		{
		  CDLMainStartRecord(GetFromParseList(0),
				     IsIncluding());
		  ClearParseList();
		}
		;

/*
 * RecordDef
 *    This is a record definition. Here, it's defined as the 
 * start of a record, and a list of fields.
 */

RecordDef	: RecordStart '{' FieldList '}' ';'
		{
		  CDLMainEndRecord();
		}
		| error ';'
                {
		  CDLPNumError(1113);
		  ClearParseList();
		  CDLMainFailed = 1;
		  yyerrok;
		}
		;

/*
 * ImportDef
 *    This is an "import" definition. The user must just specify the
 * name of the file to import in double quotes, and that's about it.
 * When we get this, we call "OpenInclude", which starts parsing from the
 * imported file, and raises an error if the file can't be opened.
 * The lex code takes care of bringing us back to this file when EOF
 * is reached.
 */

ImportDef	: ImportSpec QuotedCString ';'
		{
		  OpenInclude(GetFromParseList(0));
		  ClearParseList();
		}
		| error ';'
		{
		  CDLPNumError(1125);
		  ClearParseList();
		  CDLMainFailed = 1;
		  yyerrok;
 		}
		;

/*
 * EnumDef, EnumCaseList
 *    This is an "enum" definition. It starts with the "enum" keyword,
 * and contains a comma-separated list of possible values in squiggly
 * brackets. EnumDef defines the whole structure, and EnumCaseList
 * defines the list of cases.
 */

EnumCaseList	: EnumCaseList ',' TextString
		| TextString
		;

EnumDef		: EnumSpec TextString '{' EnumCaseList '}' ';'
		{
		  CDLMainAddEnum(GetFromParseList(0), 0);
		  ClearParseList();
		}
		| error ';'
		{
		  CDLPNumError(1128);
		  ClearParseList();
		  CDLMainFailed = 1;
		  yyerrok;
		}
		;

/* 
 * ConstDef
 *    This is a "const" definition. A const must begin with the "const"
 * keyword, and can have an integer type, a real, a single character, or
 * a string.
 */

ConstDef	: ConstSpec IntMod IntTypeSpec TextString '=' Integer ';'
	/* An integer constant */
  		{
		  if (ParseListLength() == 4)
		    {
			  CDLMainAddConst(GetFromParseList(0), 
					  GetFromParseList(1),
	   				  GetFromParseList(2), 
					  GetFromParseList(3));
		    }
		  else
		    {
			  CDLMainAddConst("", GetFromParseList(0),
	   				  GetFromParseList(1), 
					  GetFromParseList(2));
		    }
		  ClearParseList();
		}
		| ConstSpec RealType TextString '=' Real ';'
	/* A "real" constant */
		{
		  CDLMainAddConst("", GetFromParseList(0), GetFromParseList(1),
				  GetFromParseList(2));
		  ClearParseList();
		}
		| ConstSpec CharTypeSpec TextString '=' CharConst ';'
	/* A "char" constant */
		{
		  CDLMainAddConst("", GetFromParseList(0), GetFromParseList(1),
				  GetFromParseList(2));
		  ClearParseList();
		}
		| ConstSpec StringType TextString '=' QuotedCString ';'
	/* A "string" constant */
		{
		  CDLMainAddConst("", GetFromParseList(0), GetFromParseList(1),
			  	  GetFromParseList(2));			
		  ClearParseList();
		}
		| ConstSpec TextString TextString '=' TextString ';'
 	/* A constant with an illegal type. This causes an error. */
		{
		  CDLPNumError(1120);
		  CDLMainFailed = 1;
		  ClearParseList();
		}
		| error ';'
 	/* Error recovery for an illegal "const" specification */
		{
		  CDLPNumError(1120);
		  ClearParseList();
		  CDLMainFailed = 1;
		  yyerrok;
		}
		;

/*
 * WrongDef
 *    This is a definition that looks like several common things the
 * programmer might do wrong. It tries to print out a friendlier error
 * message than "syntax error".
 */

WrongDef	: TextString QuotedCString ';'
		{
		  CDLPNumError(1113);
		  CDLMainFailed = 1;
		  ClearParseList();
		}
		| TextString TextString ';'
		{
		  CDLPNumError(1113);
		  CDLMainFailed = 1;
		  ClearParseList();
		}	
		| TextString TextString '[' TextString ']'
		{
		  CDLPNumError(1113);
		  CDLMainFailed = 1;
		  ClearParseList();
		}
		| TextString TextString TextString
		{
		  CDLPNumError(1113);
		  CDLMainFailed = 1;
		  ClearParseList();
		}
		| TextString TextString TextString ';'
		{
		  CDLPNumError(1113);
		  CDLMainFailed = 1;
		  ClearParseList();
		}
		;

/*
 * WholeFile, ListOfItems, OneRecord
 *    Parsing starts at WholeFile. A Whole file is a ListOfItems, which is
 * a list of OneRecord construct. Each OneRecord may be an "import",
 * "record", "const", or "enum" definition. There's also the WrongDef
 * construct, which catches common errors a programmer might make
 * and tries to produce a friendlier error message than "syntax error".
 */

OneRecord	: ImportDef 
		| ConstDef
		| EnumDef
  		| RecordDef
		| WrongDef
		;
ListOfItems	: ListOfItems OneRecord
		| OneRecord
		;
WholeFile	: ListOfItems
		;

%%
