/*
 * mainparsewrap.C
 *
 *    These simple functions are called by the parser, and do the 
 * appropriate things to the appropriate objects. When a record definition
 * is begun, a CDLRecord object is dynamically allocated. Later, fields are
 * added when functions are called by the parser. When the record ends, 
 * it is added to the list of records in the CDLMainParser object, accessable
 * via a static member function. Since the functions below are called from 
 * the yacc-generated parser, we are guaranteed that CDLMainAddField will
 * not be called unless we are in the context of a record, and CDLMainStartRecord
 * was called first.
 *
 * mainparsewrap.C,v 1.9 1994/04/07 19:26:34 gjb Exp
 */

    /*********************************************************************
    Copyright (C) 1993-1994 by Citibank, N.A., all rights reserved.

    Permission to use, copy, modify and distribute this software and its 
    documentation for any purpose and without fee, except possible 
    redistribution costs, is hereby granted, provided that (1) the above 
    copyright notice appears in all copies and (2) both that copyright-
    notice and this permission-notice appear in supporting documentation and 
    (3) the name of Citibank, N.A. is not used in advertising or publicity 
    pertaining to distribution of the software without specific and written 
    prior permission of Citibank, N.A.

    The Software is provided "as is."  Citibank, N.A. disclaims all 
    warranties, express or implied, including but not limited to warranties 
    of suitability, merchantability and fitness for any purpose.  In no 
    event shall Citibank, N.A. be responsible for any loss of use, data or 
    profit or for any other special, indirect, incidental or consequential 
    damages, neither in an action of contract nor in an action of tort, 
    arising out of or in connection with the use or performance of this 
    software.
    *********************************************************************/


#include <iostream.h>
#include <cdlrecord.H>
#include <cdlfield.H>
#include <mainparsedefs.H>
#include <cdlmainparser.H>	
#include <cdlconstinfo.H>
#include <cdlenuminfo.H>
#include <Map.h>
#include <parselist.H>
#include <sys/param.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>

/* Globals */

static const unsigned int DocumentationLen = 1024;

int CDLMainCurrentLine;			// Current line, as counted by mainlex.l
int CDLMainFailed;			// If parsing failed

static char CDLMainFileName[MAXPATHLEN + 1]; // Name of file being parsed
static CDLRecord* curRecord;		// The record we\'re working on now
static CDLMainParser* theParser;	// Where to put the records
int CDLMainMakeIdlArray;		// Whether to generate IDL array code
static char CDLMainDocumentation[DocumentationLen];   // Documentation string

void CDLMainerror(char* errStr)

{
  CDLPError(errStr);
}

void CDLMainSetParser(CDLMainParser* p)

{
  theParser = p;
}

void CDLMainSetName(const char* fName)

{
  strncpy(CDLMainFileName, fName, MAXPATHLEN);
  CDLMainFileName[MAXPATHLEN] = '\0';
  SetCurrentFile(fName);
}

char* CDLMainGetName()

{
  return CDLMainFileName;
}

CDLMainParser* CDLMainGetParser()

{ 
  return theParser;
}

void CDLMainSetDocumentation(const char* newDocString)

{
  strncpy(CDLMainDocumentation, newDocString, DocumentationLen);
  CDLMainDocumentation[DocumentationLen - 1] = '\0';
}

/*
 * LookupArrayLen
 *    This is a utility function that compares the string passed to a
 * legal array length. An array length must be either a numeric constant
 * or a symbolic constant that's already been defined. If that's the
 * case, the actual value of the array length is returned. If the array
 * length the null string, it returns 1, for "not an array". If the
 * length parameter is invalid, it returns 0.
 */

static int LookupArrayLen(const char* lengthString)

{
  int retVal = 0;
  String errString;

  if ((lengthString == (char*)0) || (lengthString[0] == '\0'))
    {
      retVal = 1;
    }
  else
    {
      String tmpLength = lengthString;
      if (tmpLength.strspn("0123456789") == tmpLength.length())
	{					// Value is an integer
	  retVal = atoi(lengthString);
	}
      else
	{
	  CDLConstInfo testConst = CDLConstants[lengthString];
	  if (testConst.Name() == "")
	    {					// Couldn't find constant
	      CDLReport.VGetMsg(1123, errString, lengthString);
	      CDLPError(errString);
	      CDLMainFailed = 1;
	    }
	  else
	    {
	      tmpLength = testConst.Value();
	      if (tmpLength.strspn("0123456789") == tmpLength.length())
		{
		  retVal = atoi(tmpLength);
		}
	      else
		{
		  CDLReport.VGetMsg(1124, errString, lengthString);
		  CDLPError(errString);
		  CDLMainFailed = 1;
		}
	    }
	}
    }
  return retVal;
}

/*
 * CDLMainAddField
 *    Figure out the proper enumerated field types based on the
 * strings passed, and report an error if they're wrong. Then, add
 * the field to the record.
 */

void CDLMainAddField(const char* fieldMod, const char* fieldType, 
		     const char* fieldName, const char* fieldLen,
                     const char* arrayLen)
{
  CDLField::FieldType fType;
  CDLField::ModifierType mType;
  int len;
  int aSize;
  CDLField* newField;
  String errStr;

  aSize = LookupArrayLen(arrayLen);
  len = LookupArrayLen(fieldLen);

  if (!strcmp(fieldType, "real"))
    fType = CDLField::REAL;
  else if (!strcmp(fieldType, "bytes"))
    fType = CDLField::BYTES;
  else if (!strcmp(fieldType, "varBytes"))
    fType = CDLField::VARBYTES;
  else if (!strcmp(fieldType, "string"))
    fType = CDLField::STRING;
  else if (!strcmp(fieldType, "hyper"))
    fType = CDLField::HYPER;
  else if (!strcmp(fieldType, "long"))
    fType = CDLField::LONG;
  else if (!strcmp(fieldType, "short"))
    fType = CDLField::SHORT;
  else if (!strcmp(fieldType, "char"))
    {
      if (len > 1)
	fType = CDLField::CHARRAY;
      else
	fType = CDLField::CHAR;
    }
  else
    {
      CDLReport.VGetMsg(1108, errStr, fieldType);
      CDLPError(errStr);
      return;
    }

  switch(fType)
    {
    case CDLField::HYPER:
    case CDLField::LONG:
    case CDLField::SHORT:
    case CDLField::CHAR:
    case CDLField::CHARRAY:
      if (!strcmp(fieldMod, "signed"))
	mType = CDLField::SIGNED;
      else if (!strcmp(fieldMod, "unsigned"))
	mType = CDLField::UNSIGNED;
      else if (!strcmp(fieldMod, ""))
	mType = CDLField::SIGNED;
      else
	{
	  CDLReport.VGetMsg(1109, errStr, fieldMod);
	  CDLPError(errStr);
	  CDLMainFailed = 1;
	  return;
	}
      if (fType == CDLField::CHARRAY)	// Treat char array like a string
	newField = new CDLField(fType, mType, fieldName, len, aSize);
      else
	newField = new CDLField(fType, mType, fieldName, 0, aSize);
      break;
    case CDLField::BYTES:
    case CDLField::VARBYTES:
    case CDLField::STRING:
      newField = new CDLField(fType, fieldName, len, aSize);
      break;
    case CDLField::REAL:
      newField = new CDLField(fType, fieldName, 0, aSize);
      break;
    case CDLField::RECORD:
      CDLReport.GetMsg(1110, errStr);
      CDLPError(errStr);
      CDLMainFailed = 1;
      break;
    case CDLField::ENUM:
      CDLReport.GetMsg(1129, errStr);
      CDLPError(errStr);
      CDLMainFailed = 1;
      break;
    }
  if (!curRecord->AddField(newField))
    CDLMainFailed = 1;
}

/*
 * CDLMainAddUserField
 *    Add a user-defined field (a record field) to the record.
 */

void CDLMainAddUserField(const char* recName, const char* fieldName,
		         const char* arrayLen)

{
  String errStr;
  int aSize;
  
  aSize = LookupArrayLen(arrayLen);

  CDLRecordField* newField = new CDLRecordField(fieldName, recName, aSize);
  if (curRecord->Name() == recName)
    {
      CDLReport.VGetMsg(1107, errStr, (const char*)(curRecord->Name()), 
			recName);
      CDLPError(errStr);
      CDLMainFailed = 1;
    }
  else if (!curRecord->AddField(newField))
    CDLMainFailed = 1;
}

/*
 * CDLMainAddEnumField
 *    Add a field of type "enum" to the record. This entails a tiny bit
 * of additional checking, so we put it here in a separate function.
 */

void CDLMainAddEnumField(const char* en, const char* fieldName,
			 const char* arrayLen)

{
  String errStr;
  int aSize;
  CDLEnumInfo testEnum;
  Symbol enumName = en;
  
  aSize = LookupArrayLen(arrayLen);

  testEnum = CDLEnums[enumName];
  if (testEnum.Name() != "")		// It exists
    {
      CDLField* newField = new CDLField(CDLField::ENUM, fieldName,
					en, aSize);
      if (!curRecord->AddField(newField))
	CDLMainFailed = 1;
    }
  else
    {
      CDLReport.VGetMsg(1127, errStr, en);
      CDLPError(errStr);
      CDLMainFailed = 1;
    }
}

/*
 * CDLMainAddAmbiguousField
 *    The field name that the parser found wasn't one of our pre-defined
 * keywords, so it might be the name of a record or enum. Search the lists
 * of record and enumerated types and find out.
 */

void CDLMainAddAmbiguousField(const char* typeName, const char* fieldName,
                              const char* arrayLen)

{
  Symbol typeNameSymbol = typeName;
  Mapiter<Symbol, CDLEnumInfo> enumIter(CDLEnums);

  enumIter = CDLEnums.element(typeNameSymbol);
  if (enumIter.curr() != (Map_pair<Symbol, CDLEnumInfo>*)0)
    {			// The name is an enum
      CDLMainAddEnumField(typeName, fieldName, arrayLen);
    }
  else if (theParser->Find(typeName) != (CDLRecord*)0)
    {
      CDLMainAddUserField(typeName, fieldName, arrayLen);
    }
  else
    {	
      String errMsg;
      CDLReport.VGetMsg(1108, errMsg, typeName);
      CDLPError(errMsg);
      CDLMainFailed = 1;
    }
}

/*
 * CDLMainAddConst
 *    Add a constant to the global list of constants. Later, we'll
 * reference this list when looking up the sizes of array elements,
 * and some generators will stick this info in generated code.
 */

void CDLMainAddConst(const char* fieldMod, const char* fieldType,
                     const char* cName, const char* fieldValue)

{
  String errStr;
  CDLConstInfo testConst;
  Symbol constName = cName;

  testConst = CDLConstants[constName];
  if (testConst.Name() == "")		// The const isn't there already
    {
      Boolean importedConst = IsIncluding();
      CDLField::FieldType fType;
      CDLField::ModifierType mType;

      if (!strcmp(fieldType, "real"))
	fType = CDLField::REAL;
      else if (!strcmp(fieldType, "long"))
	fType = CDLField::LONG;
      else if (!strcmp(fieldType, "short"))
	fType = CDLField::SHORT;
      else if (!strcmp(fieldType, "string"))
	fType = CDLField::STRING;
      else if (!strcmp(fieldType, "char"))
	fType = CDLField::CHAR;
      else
	{
	  CDLReport.VGetMsg(1121, errStr, fieldType);
	  CDLPError(errStr);
	  CDLMainFailed = 1;
	  return;
	}
      
      if ((fType == CDLField::LONG) || (fType == CDLField::SHORT))
	{
	  if (!strcmp(fieldMod, "signed"))
	    mType = CDLField::SIGNED;
	  else if (!strcmp(fieldMod, "unsigned"))
	    mType = CDLField::UNSIGNED;
	  else if (!strcmp(fieldMod, ""))
	    mType = CDLField::SIGNED;
	  else
	    {
	      CDLReport.VGetMsg(1109, errStr, fieldMod);
	      CDLPError(errStr);
	      CDLMainFailed = 1;
	      return;
	    }
	}
      else
	{
	  mType = CDLField::NONE;
	}

      CDLConstInfo newConst(fType, mType, cName, 
			    fieldValue, importedConst);
      CDLConstants[constName] = newConst;
      CDLReport.VReport(ErrorLevels::TraceMsg,
			"Added a new constant \"%s\" with the value %s\n",
			cName, fieldValue);
    }
  else
    {
      CDLReport.VGetMsg(1122, errStr, cName);
      CDLPError(errStr);
      CDLMainFailed = 1;
    }
}

/*
 * CDLMainAddEnum
 *    Add an enum to our global list of defined enums. The parser passes
 * the index of the next entry on the parse list (after the last one it
 * read), and we must read the remaining entries and interpret them as
 * cases for our enum.
 */

void CDLMainAddEnum(const char* n, unsigned int lastPos)

{
  String errStr;
  CDLEnumInfo testEnum;
  Symbol enumName = n;
  String curCase;
  List<String> caseList;
  int inc;

  testEnum = CDLEnums[enumName];
  if (testEnum.Name() == "")		// It's not already defined
    {
      Boolean imported = IsIncluding();
      for (inc = lastPos + 1; inc < ParseListLength(); inc++)
	{
	  curCase = GetFromParseList(inc);
	  caseList.put(curCase);
	}
      
      CDLEnumInfo newEnum(n, caseList, imported);
      CDLEnums[enumName] = newEnum;
      CDLReport.VReport(ErrorLevels::TraceMsg,
			"Added a new enum \"%s\" with %i cases\n",
			(const char*)(newEnum.Name()), caseList.length());
    }
  else
    {
      CDLReport.VGetMsg(1126, errStr, n);
      CDLPError(errStr);
      CDLMainFailed = 1;
    }
}

/*
 * CDLMainStartRecord
 *    Create a new CDLRecord object. This is the first step in adding a
 * record.
 */

void CDLMainStartRecord(const char* recName,
		        Boolean imported)

{
  CDLReport.VReport(ErrorLevels::TraceMsg,
		    "Starting to parse record \"%s\"\n", recName);
  curRecord = new CDLRecord(recName, CDLMainDocumentation, 
			    imported, CDLMainMakeIdlArray);
}

/*
 * CDLMainEndRecord
 *    We're done parsing a "record" definition, so add the record to
 * the list of records maintained by the current parser object.
 * Also, clear out the various global, record-specific flags that must
 * be set by the yacc code -- in this case, the documentation string, and
 * the "idl_array" flag. 
 */

void CDLMainEndRecord()

{
  if (!theParser->AddRecord(curRecord))
    {
      CDLMainFailed = 1;
    }
  else
    {
      CDLReport.VReport(ErrorLevels::TraceMsg,
			"Done with record \"%s\"\n", 
			(const char*)(curRecord->Name()));
    }
  CDLMainMakeIdlArray = 0;
  strcpy(CDLMainDocumentation, "");
}
