/*
 * parselist.C:
 *
 *    Interfaces to maintain a list of strings. The parsing code uses
 * this interface to manipulate the list, which is more flexible than
 * using a bunch of global variables.
 *    We also have a stack of yacc input buffers, which we push on to and
 * pop off of here.
 *
 * parselist.C,v 1.6 1994/02/01 22:33:46 gjb Exp
 */

    /*********************************************************************
    Copyright (C) 1993 by Citibank, N.A., all rights reserved.

    Permission to use, copy, modify and distribute this software and its 
    documentation for any purpose and without fee, except possible 
    redistribution costs, is hereby granted, provided that (1) the above 
    copyright notice appears in all copies and (2) both that copyright-
    notice and this permission-notice appear in supporting documentation and 
    (3) the name of Citibank, N.A. is not used in advertising or publicity 
    pertaining to distribution of the software without specific and written 
    prior permission of Citibank, N.A.

    The Software is provided "as is."  Citibank, N.A. disclaims all 
    warranties, express or implied, including but not limited to warranties 
    of suitability, merchantability and fitness for any purpose.  In no 
    event shall Citibank, N.A. be responsible for any loss of use, data or 
    profit or for any other special, indirect, incidental or consequential 
    damages, neither in an action of contract nor in an action of tort, 
    arising out of or in connection with the use or performance of this 
    software.
    *********************************************************************/


#include <parselist.H>
#include <String.h>
#include <Path.h>
#include <List.h>
#include <mainparsedefs.H>
#include <cdlsuffixes.H>

/* 
 * IncludeInfo
 *    This class is used for a stack of include file information.
 */

class IncludeInfo

{
public:
  IncludeInfo();
  IncludeInfo(const String& fileName, void* buffer, int lineNo);
  IncludeInfo(const IncludeInfo&);
  IncludeInfo& operator = (const IncludeInfo&);
  int operator == (const IncludeInfo&) const;
  const String& FileName() const;
  void* Buffer();
  const int LineNo() const;
private:
  String file;
  void* buf;
  int no;
};

/* Variables used in this file */

static List<String> TheStringList;
static List<IncludeInfo> IncludeStack;
static List<String> IncludeList;
static List<String> MainIncludeList;
static List<String> IncludeDirList;
static String curFile;
static int curLine;

#ifdef _AIX
#include <List.c>

#pragma define (Const_listiter<IncludeInfo>)
#pragma define (List<IncludeInfo>)
#pragma define (Listiter<IncludeInfo>)
#endif

/*
 * IncludeInfo definitions
 */

IncludeInfo::IncludeInfo()

{
  buf = (void*)0;
  no = 0;
}

IncludeInfo::IncludeInfo(const String& fileName, void* buffer, int lineNo)

{
  file = fileName;
  buf = buffer;
  no = lineNo;
}

IncludeInfo::IncludeInfo(const IncludeInfo& info)

{
  this->operator = (info);
}

IncludeInfo& IncludeInfo::operator = (const IncludeInfo& info)

{
  file = info.file;
  buf = info.buf;
  no = info.no;
  return *this;
}

int IncludeInfo::operator == (const IncludeInfo& info) const

{
  return ((file == info.file) && (no == info.no));
}

inline const String& IncludeInfo::FileName() const

{
  return file;
}

inline void* IncludeInfo::Buffer()

{
  return buf;
}

inline const int IncludeInfo::LineNo() const

{
  return no;
}

/*
 * Parse list:
 *    This is a list of strings. The parser uses this to store an 
 * arbitrarily long list of strings and read them back later.
 */

void AddToParseList(const char* str)

{
  String tmpStr(str);
  TheStringList.put(tmpStr);
}

const char* GetFromParseList(const int index)

{
  return (const char*)(TheStringList[index]);
}

const char* GetParseListEnd()

{
  return (const char*)(TheStringList[TheStringList.length() - 1]);
}

void ClearParseList()

{
  Listiter<String> iter(TheStringList);
  while (TheStringList)		// Not empty
    iter.remove_next();
}

int ParseListLength()

{	
  return TheStringList.length();
}

/*
 * Input buffer stack
 *    This is a stack of pointers to lex input buffers. The parser uses
 * this to keep a stack of included files and return to a previous level
 * when the end of a file is reached.
 */

void PushIncludeFile(void* buffer, const char* inName, int lineNo)

{
  IncludeInfo newInfo(inName, buffer, lineNo);
  IncludeStack.put(newInfo);
}

void PopIncludeFile(void** buffer, char** outName, int* lineNo)

{
  IncludeInfo tmpInfo;

  if (IncludeStack.length() == 0)
    *buffer = (void*)0;
  else
    {
      IncludeStack.unput(tmpInfo);
      *buffer = tmpInfo.Buffer();
      *outName = (char*)((const char*)tmpInfo.FileName());
      *lineNo = tmpInfo.LineNo();
    }
}

Boolean IsIncluding()

{
  return (IncludeStack.length() > 0);
}

void ClearIncludeFiles()

{
  Listiter<IncludeInfo> iter(IncludeStack);
  while (IncludeStack)		// While not empty
    iter.remove_next();
}

/*
 * Include name list
 *    This is a list of files included by the file currently being parsed.
 * It is used to avoid including the same file more than once.
 */

void AddIncludeName(const char* newName)

{	
  String tmpStr(newName);
  IncludeList.put(newName);
}

int CheckIncludeName(const char* name)

{
  String tmpStr(name);
  Listiter<String> iter(IncludeList);
  return (iter.find_next(tmpStr));
}

const char* GetIncludeName(const int index)

{
  return IncludeList[index];
}

void ClearIncludeNames()

{
  Listiter<String> iter(IncludeList);
  while (IncludeList)
    iter.remove_next();
}

int NumIncludeNames()

{
  return IncludeList.length();
}

/*
 * Main include name list
 *    This is like the regular include name list, but is only used to
 * keep track of which files were included in the main file (the one the
 * user ran cdl on) so a proper list can be built in the generated C++ code.
 */

void AddMainIncludeName(const char* newName)

{
  String tmpStr(newName);
  MainIncludeList.put(newName);
}

const char* GetMainIncludeName(const int index)

{
  return (const char*)MainIncludeList[index];
}

void ClearMainIncludeNames()

{
  Listiter<String> iter(MainIncludeList);
  while (MainIncludeList)
    iter.remove_next();
}

int NumMainIncludeNames()

{
  return MainIncludeList.length();
}

/*
 * OpenIncludeFile
 *    Open the named include file by searching the list of include 
 * directories passed on the command line. If the file is found, it is
 * preprocessed and a descriptor to the preprocessed file is returned.
 */

FILE* OpenIncludeFile(const char* name, char* realOutName, char* outName)

{
  String testName;
  String tmpName;
  FILE* retVal;
  Listiter<String> iter(IncludeDirList);

  outName[0] = '\0';
  retVal = NULL;
  while ((retVal == NULL) && iter.next(testName))
    {					// For each directory, build the name,
      testName += name;			// and return a FILE* when you get
      retVal = fopen(testName, "r");	// one that works
    }
  if (retVal != NULL)
    {
      fclose(retVal);
      if (CDLMainGetParser()->PreProcess(testName, tmpName))
	{
	  strcpy(realOutName, testName);
	  strcpy(outName, tmpName);
	  retVal = fopen(tmpName, "r");
	}
      else
	{
	  retVal = NULL;
	}
    }
  return retVal;
}

/* 
 * The calling function (main) makes sure that the elements in argv all begin
 * with "-I"
 */



void MakeIncludeList(int argc, const char* const* argv)

{
  String curStr;
  register int inc;
  Listiter<String> iter(IncludeDirList);

  while (IncludeDirList)
    iter.remove_next();
  for (inc = 0; inc < argc; inc++)
    {
      curStr = argv[inc];
      curStr.get();			// Remove two first chars ("-I")
      curStr.get();
      if (curStr[curStr.length() - 1] != '/')
	curStr += "/";
      IncludeDirList.put(curStr);
    }
  IncludeDirList.put("./");			// Add defaults: current dir
  IncludeDirList.put(CDLDCEIncludes);		//  and DCE include dir
}

/*
 * These next functions let us globally store information on the
 * current file being parsed and the current line.
 */

void SetCurrentFile(const char* fileName)

{
  curFile = fileName;
}

void SetCurrentLine(int lineNo)

{
  curLine = lineNo;
}

const char* GetCurrentFile()

{
  return curFile;
}

int GetCurrentLine()

{
  return curLine;
}

/*
 * CDLPError, CDLPWarning
 *    Print an error message in the appropriate format. These messages
 * must be in a format that emacs and other types of tools can handle.
 * This format is defined in the message catalog under message 400 (for error
 * messages) and 401 (for warnings). In both cases, the first parameter is
 * the file name, the second is the line number, and the third is the error
 * message.
 *    We output these messages using printf rather than the "CDLReport"
 * object because we don't want the extra crud the CDLReport object puts
 * into formatted messages. The extra crud means standard tools can't 
 * handle the output.
 */

void CDLPError(const char* errStr)

{
  String msg;

  CDLReport.VGetMsg(400, msg, (const char*)curFile, curLine, errStr);
  fprintf(stderr, "%s", (const char*)msg);
}

void CDLPNumError(const int msgNum)

{
  String errMsg;
  String msg;

  CDLReport.GetMsg(msgNum, errMsg);
  CDLReport.VGetMsg(400, msg, (const char*)curFile, curLine, 
		    (const char*)errMsg);
  fprintf(stderr, "%s", (const char*)msg);
}

void CDLPWarning(const char* errStr)

{
  String msg;

  CDLReport.VGetMsg(401, msg, (const char*)curFile, curLine, errStr);
  fprintf(stderr, "%s", (const char*)msg);
}

void CDLPNumWarning(const int msgNum)

{
  String errMsg;
  String msg;

  CDLReport.GetMsg(msgNum, errMsg);
  CDLReport.VGetMsg(401, msg, (const char*)curFile, curLine, 
		    (const char*)errMsg);
  fprintf(stderr, "%s", (const char*)msg);
}

